create function instantiate(schema regnamespace default 'omni_email') returns void
    language plpgsql
as
$$
declare
    old_search_path text := current_setting('search_path');
begin
    -- Set the search path to target schema and public
    perform
        set_config('search_path', schema::text || ',public,omni_polyfill,pg_catalog', true);

    -- Core

    create domain email_address as text check ( position('@' in value) > 0 );

    perform identity_type('email_message_id', type => 'uuid', nextval => 'uuidv7()');

    create table email_message
    (
        id email_message_id primary key default email_message_id_nextval()
    );

    create table email_message_sender
    (
        id             email_message_id not null unique references email_message (id),
        sender_address email_address    not null,
        unique (id, sender_address)
    );

    create table email_message_sender_display_name
    (
        id                  email_message_id not null,
        sender_address      email_address    not null,
        sender_display_name text             not null,
        foreign key (id, sender_address) references email_message_sender (id, sender_address)
    );

    create table email_message_subject
    (
        id      email_message_id not null references email_message (id),
        subject text             not null
    );

    create type email_message_recipient_type as enum ('To','Cc','Bcc');

    create table email_message_recipient
    (
        id                email_message_id             not null references email_message (id),
        recipient_address email_address                not null,
        recipient_type    email_message_recipient_type not null default 'To',
        unique (id, recipient_address)
    );

    create table email_message_recipient_display_name
    (
        id                     email_message_id not null,
        recipient_address      email_address    not null,
        recipient_display_name text             not null,
        foreign key (id, recipient_address) references email_message_recipient (id, recipient_address)
    );

    create table email_message_text_body
    (
        id        email_message_id not null unique references email_message (id),
        text_body text             not null
    );

    create table email_message_html_body
    (
        id        email_message_id not null unique references email_message (id),
        html_body text             not null
    );

    create type email_attachment_disposition as enum ('attachment','inline');

    create table email_message_attachment
    (
        id                     email_message_id             not null references email_message (id),
        attachment_file_name   text                         not null,
        attachment_mime_type   text                         not null,
        attachment_file        bytea                        not null,
        attachment_disposition email_attachment_disposition not null default 'attachment'
    );

    create table email_message_header
    (
        id           email_message_id not null references email_message (id),
        header_name  text             not null,
        header_value text             not null,
        unique (id, header_name)
    );

    --- View
    create view email_message_view as
        select
            id,
            sender_display_name,
            sender_address,
            recipients,
            subject,
            text_body,
            html_body
        from
            email_message
            natural left join email_message_sender
            natural left join email_message_sender_display_name
            natural left join email_message_subject
            natural left join email_message_recipient
            natural left join email_message_text_body
            natural left join email_message_html_body
            natural left join email_message_attachment
            natural left join email_message_header
            left join         lateral ( select
                                                    array_agg(array [email_message_recipient.recipient_type, recipient_display_name, recipient_address]::text[])
                                                    over (partition by id) as recipients
                                        from
                                            email_message_recipient emr
                                            natural left join email_message_recipient_display_name where emr.id = email_message.id) recipients on true;

    -- Outbox
    create table email_message_outbox
    (
        id email_message_id not null unique references email_message (id)
    );
    create view email_message_outbox_available as
        select * from email_message_outbox for update skip locked;

    create table email_message_sent
    (
        id email_message_id not null unique references email_message (id)
    );

    create function email_message_sent() returns trigger
        language plpgsql as
    $email_message_sent$
    begin
        delete from email_message_outbox where id = new.id;
        return new;
    end;
    $email_message_sent$;

    execute format('alter function email_message_sent set search_path to %I,public', schema);

    create trigger email_message_sent
        after insert
        on email_message_sent
        for each row
    execute function email_message_sent();

    perform set_config('search_path', old_search_path, true);
end;
$$;

select instantiate(schema => 'omni_email');
create function instantiate_resend(schema regnamespace default 'omni_email',
                                   credentials name default 'omni_credentials.credentials') returns void
    language plpgsql
as
$$
declare
    old_search_path text := current_setting('search_path');
begin
    perform from pg_extension where extname = 'omni_httpc';
    if not found then
        raise exception 'omni_httpc extension is required';
    end if;

    begin
        perform credentials::regclass;
    exception
        when others then
            raise exception 'omni_credentials table % must be available', credentials;
    end;

    -- Set the search path to target schema and public
    perform
        set_config('search_path', schema::text || ',public,omni_polyfill,pg_catalog', true);

    create table email_message_sent_via_resend
    (
        id        email_message_id not null references email_message (id),
        resend_id uuid             not null unique
    );

    create function send_email_resend(id email_message_id) returns bool
        language plpgsql as
    $send_email_resend$
    declare
        api_key text;
        res     omni_httpc.http_response;
    begin
        execute format('select value from %s where name = %L and kind = %L',
                       current_setting('omni_credentials.credentials_table'), 'resend.com', 'api_secret') into api_key;
        if api_key is null then
            raise exception 'credential api:resend.com is not defined';
        end if;
        select
            (omni_httpc.http_execute(omni_httpc.http_request('https://api.resend.com/emails', method => 'POST',
                                                             headers => array [omni_http.http_header('Authorization', 'Bearer ' || api_key),
                                                                 omni_http.http_header('Content-Type', 'application/json')
                                                                 ],
                                                             body => convert_to(
                                                                     jsonb_build_object('from', sender_address, 'to',
                                                                                        recipient_addresses, 'subject',
                                                                                        subject, 'text', text_body,
                                                                                        'html', html_body)::text,
                                                                     'utf-8')))).*
        into res
        from
            email_message
            natural left join email_message_subject
            natural left join email_message_text_body
            natural left join email_message_html_body
            natural left join email_message_sender
            left join         lateral ( select
                                            jsonb_agg(emr.recipient_address) as recipient_addresses
                                        from
                                            email_message_recipient                                emr
                                            natural left join email_message_recipient_display_name emrdn
                                        where
                                            emr.id = email_message.id) raddr on true
        where
            email_message.id = send_email_resend.id;
        if res.status != 200 then
            raise exception '%', convert_from(res.body, 'utf8')::json;
        end if;
        insert
        into
            email_message_sent
        values (id);
        insert
        into
            email_message_sent_via_resend
        values (id, ((convert_from(res.body, 'utf8')::jsonb) ->> 'id')::uuid);
        return res.status = 200;
    end;
    $send_email_resend$;
    execute format('alter function send_email_resend set omni_credentials.credentials_table = %I', credentials);
    execute format('alter function send_email_resend set search_path = %s,public', schema);


    perform set_config('search_path', old_search_path, true);
end;
$$;

;
