create domain port integer check (value >= 0 and value <= 65535);

create type http_request as
(
    method       omni_http.http_method,
    path         text,
    query_string text,
    body         bytea,
    headers      http_headers
);

create function http_request(path text, method omni_http.http_method default 'GET', query_string text default null,
                             body bytea default null,
                             headers http_headers default array []::http_headers)
    returns http_request
    language sql
    immutable
as
$$
select row (method, path, query_string, body, headers)
$$;

create type http_response as
(
    body    bytea,
    status  smallint,
    headers http_headers
);

create type http_proxy as
(
    url           text,
    preserve_host boolean
);

create domain abort as omni_types.unit;

select omni_types.sum_type('http_outcome', 'http_response', 'abort', 'http_proxy');

create function abort() returns http_outcome as
$$
select omni_httpd.http_outcome_from_abort(omni_types.unit())
$$ language sql immutable;
create function http_proxy(url text, preserve_host boolean default true) returns http_outcome as
$$
select omni_httpd.http_outcome_from_http_proxy(row (url, preserve_host))
$$ language sql immutable;

create function http_response(
    body anycompatible default null,
    status int default 200,
    headers http_headers default null
)
    returns http_outcome
as
'MODULE_PATHNAME',
'http_response'
    language c immutable;

create type http_protocol as enum ('http', 'https');;
create table listeners
(
    id             integer primary key generated always as identity,
    address        inet          not null default '127.0.0.1',
    port           port          not null default 80,
    effective_port port          not null default 0,
    protocol       http_protocol not null default 'http'
);

create function check_if_role_accessible_to_current_user(role name) returns boolean
as
$$
declare
    result boolean;
begin
    with
        recursive
        role_members as (select
                             roleid,
                             member
                         from
                             pg_auth_members
                         where
                             roleid = (select oid from pg_roles where rolname = role)

                         union all

                         select
                             am.roleid,
                             am.member
                         from
                             pg_auth_members  am
                             join
                                 role_members rm on am.roleid = rm.member)
    select
        true
    into result
    from
        pg_roles r
    where
        (r.rolname = current_user and r.rolsuper) or
        (r.oid in (select member from role_members)) or
        (r.rolname = role and role = current_user)
    limit 1;
    if not found then
        return false;
    else
        return true;
    end if;
end;

$$ language plpgsql;


create table handlers
(
    id        integer primary key generated always as identity,
    query     text not null,
    role_name name not null default current_user check (check_if_role_accessible_to_current_user(role_name))
);

create function handlers_query_validity_trigger() returns trigger
as
'MODULE_PATHNAME',
'handlers_query_validity_trigger' language c;

create constraint trigger handlers_query_validity_trigger
    after insert or update
    on handlers
    deferrable initially deferred
    for each row
execute function handlers_query_validity_trigger();

create table listeners_handlers
(
    listener_id integer not null references listeners (id),
    handler_id  integer not null references handlers (id)
);
create index listeners_handlers_index on listeners_handlers (listener_id, handler_id);

create table configuration_reloads
(
    id          integer primary key generated always as identity,
    happened_at timestamp not null default now()
);

-- Wait for the number of configuration reloads to be `n` or greater
-- Useful for testing
create procedure wait_for_configuration_reloads(n int) as
$$
declare
    c  int = 0;
    n_ int = n;
begin
    loop
        with
            reloads as (select
                            id
                        from
                            omni_httpd.configuration_reloads
                        order by happened_at asc
                        limit n_)
        delete
        from
            omni_httpd.configuration_reloads
        where
            id in (select id from reloads);
        declare
            rowc int;
        begin
            get diagnostics rowc = row_count;
            n_ = n_ - rowc;
            c = c + rowc;
        end;
        exit when c >= n;
    end loop;
end;
$$ language plpgsql;



create function reload_configuration_trigger() returns trigger
as
'MODULE_PATHNAME',
'reload_configuration'
    language c;

create function reload_configuration() returns bool
as
'MODULE_PATHNAME',
'reload_configuration'
    language c;

create trigger listeners_updated
    after update or delete or insert
    on listeners
execute function reload_configuration_trigger();

create trigger handlers_updated
    after update or delete or insert
    on handlers
execute function reload_configuration_trigger();

create trigger listeners_handlers_updated
    after update or delete or insert
    on listeners_handlers
execute function reload_configuration_trigger();;
create function cascading_query_reduce(internal, name text, query text) returns internal
as
'MODULE_PATHNAME',
'cascading_query_reduce' language c;

create function cascading_query_final(internal) returns text
as
'MODULE_PATHNAME',
'cascading_query_final' language c;

create aggregate cascading_query (name text, query text) (
    sfunc = cascading_query_reduce,
    finalfunc = cascading_query_final,
    stype = internal
    );
;
create function default_page(request http_request) returns setof http_outcome as
$$
begin
    return query (with
                      stats as (select * from pg_catalog.pg_stat_database where datname = current_database())
                  select *
                  from
                      omni_httpd.http_response(headers => array [omni_http.http_header('content-type', 'text/html')],
                                               body => $html$
       <!DOCTYPE html>
       <html>
         <head>
           <title>Omnigres</title>
           <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bulma@0.9.4/css/bulma.min.css">
           <meta name="viewport" content="width=device-width, initial-scale=1">
         </head>
         <body class="container">
         <section class="section">
           <div class="container">
             <h1 class="title">Omnigres</h1>

             <div class="tile is-ancestor">
                <div class="tile is-parent is-8">
                 <article class="tile is-child notification is-primary">
                   <div class="content">
                     <p class="title">Welcome!</p>
                     <p class="subtitle">What's next?</p>
                     <div class="content">
                     <p>You can update the query in the <code>omni_httpd.handlers</code> table to change this default page.</p>

                     <p><a href="https://docs.omnigres.org">Documentation</a></p>
                     </div>
                   </div>
                 </article>
               </div>
               <div class="tile is-vertical">
                 <div class="tile">
                   <div class="tile is-parent is-vertical">
                     <article class="tile is-child notification is-grey-lighter">
                       <p class="title">Database</p>
                       <p class="subtitle"><strong>$html$ || current_database() || $html$</strong></p>
                       <p> <strong>Backends</strong>: $html$ || (select numbackends from stats) || $html$ </p>
                       <p> <strong>Transactions committed</strong>: $html$ || (select xact_commit from stats) || $html$ </p>
                     </article>
                   </div>
                 </div>
               </div>
             </div>

             <p class="is-size-7">
               Running on <strong> $html$ || version() || $html$ </strong>
             </p>
           </div>
         </section>
         </body>
       </html>
       $html$));
end;
$$ language plpgsql;;
create function static_file_handlers(fs regproc, handler_priority int, listing bool default false)
    returns table
            (
                name     text,
                query    text,
                priority int
            )
    language plpgsql
as
$plpgsql$
begin
    perform from pg_extension where extname = 'omni_vfs';
    if not found then
        raise exception 'omni_vfs required';
    end if;
    perform from pg_extension where extname = 'omni_mimetypes';
    if not found then
        raise exception 'omni_mimetypes required';
    end if;
    perform
    from
        pg_proc
    where
            oid = fs::oid and
            (array_length(proargnames, 1) = 0 or proargnames is null) and
            omni_vfs_types_v1.is_valid_fs(prorettype);
    if not found then
        raise exception 'must have %() function with no arguments returning a valid omni_vfs filesystem', fs;
    end if;
    return query
        select
            handler_name,
            handler_query,
            handler_priority
        from
            (values
                 ('file',
                  format($$select omni_httpd.http_response(
            omni_vfs.read(%1$s(), request.path),
            headers => array [omni_http.http_header('content-type', coalesce (mime_types.name, 'application/octet-stream'))]::omni_http.http_header[])
            from request
            left join omni_mimetypes.file_extensions on request.path like '%%.' || file_extensions.extension
            left join omni_mimetypes.mime_types_file_extensions mtfe on mtfe.file_extension_id = file_extensions.id
            left join omni_mimetypes.mime_types on mtfe.mime_type_id = mime_types.id
            where request.method = 'GET' and (omni_vfs.file_info(%1$s(), request.path)).kind = 'file'
        $$, fs)),
                 ('directory',
                  format($$
        select
            omni_httpd.http_response(
                    omni_vfs.read(%1$s(), request.path || '/index.html'),
                    headers => array [omni_http.http_header('content-type', 'text/html')]::omni_http.http_header[])
        from
            request
        where
            request.method = 'GET' and
            (omni_vfs.file_info(%1$s(), request.path)).kind = 'dir' and
            (omni_vfs.file_info(%1$s(), request.path || '/index.html')).kind = 'file'
            $$, fs)),
                 ('directory_listing',
                  format($$
        select
        omni_httpd.http_response(
           (select string_agg('<a href="' || case when request.path = '/' then '/' else request.path || '/' end  || name || '">' || name || '</a>', '<br>') from omni_vfs.list(%1$s(), request.path)),
           headers =>  array [omni_http.http_header('content-type', 'text/html')]::omni_http.http_header[])
        from
            request
        where
            request.method = 'GET' and
            (omni_vfs.file_info(%1$s(), request.path)).kind = 'dir' and
            (omni_vfs.file_info(%1$s(), request.path || '/index.html')) is not distinct from null
            $$, fs))) handlers(handler_name, handler_query)
        where
            (handler_name = 'directory_listing' and listing) or
            (handler_name != 'directory_listing');
end;
$plpgsql$;;
create function unload() returns void
    language c as
'MODULE_PATHNAME';;
-- Empty implementation
create function handler(int, http_request) returns http_outcome
    language plpgsql as
$$
begin
    return omni_httpd.http_response(status => 404);
end;
$$;

comment on function handler(int, http_request) is $$
This function responds to all HTTP requests. Can't be deleted but can be replaced.
$$;

-- Effectively, soft-deprecate omni_httpd.handlers by wrapping them into the `handler`
create function regenerate_handler_trigger() returns trigger
    security definer
    language plpgsql as
$$
declare
    rec      record;
    body     text;
    proc_oid oid;
    different_roles bool;
    _role_name      name;
begin
    -- Safety check
    select 'omni_httpd.handler'::regproc::oid into proc_oid;

    select distinct on (role_name) count(*) > 1, role_name
    from omni_httpd.handlers
    group by role_name
    into different_roles, _role_name;

    if different_roles then
        for rec in select listener_id, handler_id, query, role_name
                   from omni_httpd.listeners_handlers
                            inner join omni_httpd.handlers on handlers.id = listeners_handlers.handler_id
            loop
                declare
                    ident text;
                begin
                    ident := 'handler_' || rec.listener_id || '_' || rec.handler_id;
                    execute format(
                            'create or replace function omni_httpd.%I(omni_httpd.http_request) returns omni_httpd.http_outcome language plpgsql security definer as $__sql__$ begin return (with request as (select ($1).*) %s limit 1); end;$__sql__$',
                            ident, rec.query);
                    execute format('alter function omni_httpd.%I(omni_httpd.http_request) owner to %s', ident,
                                   rec.role_name::regrole);
                end;
            end loop;
    end if;

    body :=
            format('create or replace procedure omni_httpd.handler(int, omni_httpd.http_request, out omni_httpd.http_outcome) language plpgsql as $___sql___$ begin $3 := (');

    if not different_roles then
    body := body || 'with request as (select ($2).*) select * from (';
    end if;

    select body || omni_httpd.cascading_query('handler_' || listener_id || '_' || handler_id,
                                              case
                                                  when different_roles then
                                                      'select * from omni_httpd.handler_' || listener_id || '_' ||
                                                      handler_id ||
                                                      '($2) where $1 = ' || listener_id
                                                  else 'select * from (' || query || ') q where $1 = ' || listener_id end
                   )
    from omni_httpd.listeners_handlers
             inner join omni_httpd.handlers on handlers.id = listeners_handlers.handler_id
    into body;

    if not different_roles then
        body := body || ') t';
    end if;

    body := body || format(' limit 1); end; $___sql___$ ');

    -- Enact the regeneration
    execute body;

    if not different_roles then
        execute format(
                'alter procedure omni_httpd.handler(int, omni_httpd.http_request, out omni_httpd.http_outcome) owner to %s',
                       _role_name::regrole);
        execute format(
                'alter procedure omni_httpd.handler(int, omni_httpd.http_request, out omni_httpd.http_outcome) security definer',
                       _role_name::regrole);
        execute format(
                'alter procedure omni_httpd.handler(int, omni_httpd.http_request, out omni_httpd.http_outcome) owner to %s',
                       _role_name::regrole);
    end if;

    -- Perform safety check
    perform where proc_oid = 'omni_httpd.handler'::regproc::oid;
    if not found then
        raise exception 'omni_httpd.handler changed OID, invariant violation';
    end if;
    return new;
end;
$$;

create trigger deprecated_handlers_to_handler_generator
    after insert or update or delete or truncate
    on handlers
    for each statement
execute function regenerate_handler_trigger();

create trigger deprecated_listeners_handlers_to_handler_generator
    after insert or update or delete or truncate
    on listeners_handlers
    for each statement
execute function regenerate_handler_trigger();

drop trigger handlers_updated on handlers;
drop trigger listeners_handlers_updated on listeners_handlers;
;
-- Basic implementation
create function websocket_handler(int, uuid, http_request) returns bool
    language plpgsql as
$$
    begin
        return false;
    end;
$$;

comment on function websocket_handler(int, uuid, http_request) is $$
This function responds to WebSocket upgrade attempts. If it returns anything but true, the connection is rejected.

Can't be deleted but can be replaced.
$$;


create function websocket_on_open(uuid) returns void
    language plpgsql as
$$
begin
end;
$$;

comment on function websocket_on_open(uuid) is $$
This function is called when WebSocket connection is open.

Can't be deleted but can be replaced.
$$;

comment on function websocket_handler(int, uuid, http_request) is $$
This function responds to WebSocket upgrade attempts. If it returns anything but true, the connection is rejected.

Can't be deleted but can be replaced.
$$;


create function websocket_on_close(uuid) returns void
    language plpgsql as
$$
begin
end;
$$;

comment on function websocket_on_close(uuid) is $$
This function is called when WebSocket connection is close.

Can't be deleted but can be replaced.
$$;

create function websocket_on_message(uuid, bytea) returns void
    language plpgsql as
$$
begin
end;
$$;

create function websocket_on_message(uuid, text) returns void
    language plpgsql as
$$
begin
end;
$$;


comment on function websocket_on_message(uuid, bytea) is $$
This function handles WebSocket binary messages.

Can't be deleted but can be replaced.
$$;

comment on function websocket_on_message(uuid, text) is $$
This function handles WebSocket text messages.

Can't be deleted but can be replaced.
$$;

create function websocket_send(uuid, text) returns void
    language c as
'MODULE_PATHNAME',
'websocket_send_text';

create function websocket_send(uuid, bytea) returns void
    language c as
'MODULE_PATHNAME',
'websocket_send_bytea';;
create procedure stop(immediate boolean default false)
as
'MODULE_PATHNAME'language c;

create procedure start(immediate boolean default false)
as
'MODULE_PATHNAME'language c;;
-- Replace function with procedure
drop function handler(int, http_request);

-- Empty implementation of the handler
-- NOTE: we use `inout` here mostly to support Postgres 13
-- (Postgres 13 does not support `out` in procedures)
-- and we're currently passing `null` into it
-- TODO: review this policy once we stop supporting Postgres 13
create procedure handler(int, http_request, inout http_outcome)
    language plpgsql as
$$
begin
    $3 := omni_httpd.http_response(status => 404);
end;
$$;

comment on procedure handler(int, http_request, inout http_outcome) is $$
This procedure handlers incoming HTTP requests. Can't be deleted but can be replaced.
$$;
;
drop procedure handler(int, http_request, inout http_outcome);

create procedure handler(int, http_request, out http_outcome)
    language plpgsql as
$$
begin
    $3 := omni_httpd.http_response(status => 404);
end;
$$;

create type urlpattern as
(
    protocol text,
    username text,
    password text,
    hostname text,
    port     int,
    pathname text,
    search   text,
    hash     text,
    method   omni_http.http_method
);

create function urlpattern(pathname text default null, search text default null, hash text default null,
                           method omni_http.http_method default null,
                           hostname text default null, port int default null,
                           protocol text default null, username text default null,
                           password text default null) returns urlpattern
    language sql
    immutable
as
$$
select row (protocol, username, password, hostname, port, pathname, search, hash, method)::omni_httpd.urlpattern;
$$;

create domain route_priority as int;

create materialized view available_routers as
    select
        c.oid::regclass                                                                as router_relation,
        min(case when a.atttypid = 'omni_httpd.urlpattern'::regtype then a.attnum end) as match_col_idx,
        min(case when a.atttypid = 'regprocedure'::regtype then a.attnum end)          as handler_col_idx,
        min(case when a.atttypid = 'route_priority'::regtype then a.attnum end)        as priority_col_idx
    from
        pg_class          c
        join pg_attribute a
             on a.attrelid = c.oid
    where
        a.attnum > 0       -- Skip system columns
            and
        not a.attisdropped -- Skip dropped columns
    group by
        c.oid
    having
                count(*) filter (where a.atttypid = 'omni_httpd.urlpattern'::regtype) = 1 and
                count(*) filter (where a.atttypid = 'regprocedure'::regtype) = 1 and
                count(*) filter (where a.atttypid = 'route_priority'::regtype) = any (array [0, 1]);

create view default_router as
    select *
    from
        (values
             (urlpattern(),
              'handler(int,omni_httpd.http_request)'::regprocedure,
              (-2147483648)::route_priority)) t
    where
        coalesce(not current_setting('omni_httpd.no_init', true)::bool, true); -- should init

refresh materialized view available_routers;

create or replace function refresh_available_routers() returns event_trigger as
$$
begin
    refresh materialized view omni_httpd.available_routers;
end;
$$ language plpgsql;

create event trigger refresh_available_routers
    on ddl_command_end
    when tag in ('CREATE TABLE', 'CREATE TABLE AS', 'CREATE VIEW', 'DROP TABLE', 'DROP VIEW', 'CREATE MATERIALIZED VIEW', 'DROP MATERIALIZED VIEW')
execute function refresh_available_routers();

create function routes()
    returns table
            (
                router   regclass,
                matcher  regtype,
                match    jsonb,
                priority int,
                handler  regprocedure,
                type     text
            )
    language plpgsql
as
$$
declare
    rec       record;
    route_rec record;
begin
    for rec in select * from omni_httpd.available_routers
        loop
            router := rec.router_relation;
            matcher := 'omni_httpd.urlpattern';
            for route_rec in execute format('select to_json(t) as row from %1$s t', router)
                loop
                    declare
                        keys text[];
                        row  json := route_rec.row;
                    begin
                        select array_agg(key) into keys from (select * from json_object_keys(row) as t(key)) as t;
                        match := jsonb_strip_nulls((row ->> keys[rec.match_col_idx])::jsonb);
                        handler := row ->> keys[rec.handler_col_idx];
                        with
                            function_details as (select
                                                     unnest(p.proallargtypes) as arg_type,
                                                     unnest(p.proargmodes)    as arg_mode
                                                 from
                                                     pg_proc p
                                                 where
                                                     p.oid = handler)
                        select
                            case
                                when arg_mode is not null then
                                    (case
                                         when arg_mode = 'o' then 'handler'
                                         when arg_mode = 'b' then 'middleware'
                                         else 'unsupported'
                                        end)
                                else 'unsuppported' -- Defaults to IN if proargmodes is NULL
                                end as type
                        into type
                        from
                            function_details fd
                            join pg_type     t on fd.
                                                      arg_type = t.oid
                        where
                            t.typname = 'http_outcome' and
                            t.typnamespace = 'omni_httpd'::regnamespace;
                        type := coalesce(type, 'handler');
                        if rec.priority_col_idx is not null then
                            priority := (row ->> keys[rec.priority_col_idx])::int;
                        else
                            priority := 0;
                        end if;
                        return next;
                    end;
                end loop;
        end loop;
    return;
end;
$$ set search_path = '';

create view routes as
    select *
    from
        routes()
    order by
        priority desc nulls last;

create table urlpattern_router (
    match urlpattern unique,
    handler regprocedure
);

comment on table urlpattern_router is 'Basic default priority router';

create table router_priority (
   priority route_priority
);

comment on table router_priority is 'Router priority specialization';;
drop function static_file_handlers;
create function instantiate_static_file_handler(schema name default '@extschema@',
                                                name name default 'static_file_handler') returns void
    language plpgsql as
$$
declare
    old_search_path text := current_setting('search_path');
begin
    -- check for required extensions
    perform from pg_extension where extname = 'omni_vfs';
    if not found then
        raise exception 'omni_vfs extension is required to be installed';
    end if;

    perform from pg_extension where extname = 'omni_mimetypes';
    if not found then
        raise exception 'omni_mimetypes extension is required to be installed';
    end if;

    perform set_config('search_path', schema::text, true);

    create
    or
    replace function static_file_handler(request omni_httpd.http_request,
                                         fs anyelement, fs_path text default '', path text default null,
                                         listing boolean default false)
    returns omni_httpd.http_outcome
    language plpgsql as
$static_file_handler$
declare
    outcome omni_httpd.http_outcome;
begin
    if path is null then
        path := request.path;
    end if;
    --- File
    if request.method = 'GET' and
       (omni_vfs.file_info(fs, fs_path || path)).kind = 'file' then
        select
            omni_httpd.http_response(
                    omni_vfs.read(fs, fs_path || path),
                    headers => array [omni_http.http_header('content-type',
                                                            coalesce(mime_types.name, 'application/octet-stream'))]::omni_http.http_header[])
        from
                (select)                                            _
                    left join omni_mimetypes.file_extensions on path like '%%.' || file_extensions.extension
                left join omni_mimetypes.mime_types_file_extensions mtfe on mtfe.file_extension_id = file_extensions.id
                left join omni_mimetypes.mime_types on mtfe.mime_type_id = mime_types.id
        into outcome;
    end if;

    if outcome is not null then
        return outcome;
    end if;

    --- Directory
    if request.method = 'GET' and
       ((path = '/') or
        (omni_vfs.file_info(fs, fs_path || path)).kind = 'dir') and
       (omni_vfs.file_info(fs, fs_path || path || '/index.html')).kind = 'file' then
        select
            omni_httpd.http_response(
                    omni_vfs.read(fs, fs_path || path || '/index.html'),
                    headers => array [omni_http.http_header('content-type', 'text/html')]::omni_http.http_header[])
        into outcome;
    end if;

    if outcome is not null then
        return outcome;
    end if;

    --- File listing

    if listing and request.method = 'GET' and
       ((path = '/') or
        (omni_vfs.file_info(fs, fs_path || path)).kind = 'dir') and
       (omni_vfs.file_info(fs, fs_path || path || '/index.html')) is not distinct from null then
        select
            omni_httpd.http_response(
                    (select string_agg('<a href="' || case when path = '/' then '/' else path || '/' end ||
                                    name ||
                                    '">' || name || '</a>', '<br>')
                     from
                         omni_vfs.list(fs, fs_path || path)),
                    headers => array [omni_http.http_header('content-type', 'text/html')]::omni_http.http_header[])

        into outcome;
    end if;

    if outcome is not null then
        return outcome;
    end if;

    return omni_httpd.http_response(status => 404);
end;
$static_file_handler$;

    if name != 'static_file_handler' then
        execute format('alter procedure static_file_handler rename to %I', name);
    end if;

    create table static_file_handler_router (

    );

    perform set_config('search_path', old_search_path, true);
end;
$$;
;
alter event trigger refresh_available_routers disable;
drop materialized view available_routers cascade;

create materialized view available_routers as
select c.oid::regclass                                                                as router_relation,
       min(case when a.atttypid = 'omni_httpd.urlpattern'::regtype then a.attnum end) as match_col_idx,
       min(case when a.atttypid = 'regprocedure'::regtype then a.attnum end)          as handler_col_idx,
       min(case when a.atttypid = 'route_priority'::regtype then a.attnum end)        as priority_col_idx,
       min(case when a.atttypid = 'regprocedure'::regtype then a.attname end)::name   as handler_col_name
from pg_class c
         join pg_attribute a
              on a.attrelid = c.oid
where a.attnum > 0       -- Skip system columns
  and not a.attisdropped -- Skip dropped columns
group by c.oid
having count(*) filter (where a.atttypid = 'omni_httpd.urlpattern'::regtype) = 1
   and count(*) filter (where a.atttypid = 'regprocedure'::regtype) = 1
   and count(*) filter (where a.atttypid = 'route_priority'::regtype) = any (array [0, 1]);

refresh materialized view available_routers;

alter event trigger refresh_available_routers enable;

create table handler_role
(
    handler regprocedure not null primary key,
    role    regrole      not null
);
;
create function stop_handling() returns boolean
    language 'c' as
'MODULE_PATHNAME';

;
-- Initialization
with
    config as
        (select
             coalesce(not current_setting('omni_httpd.no_init', true)::bool, true)     as should_init,
             coalesce(current_setting('omni_httpd.init_listen_address', true),
                      '0.0.0.0')::inet                                                 as init_listen_address,
             coalesce(current_setting('omni_httpd.init_port', true)::port, 8080::port) as init_port),
    listener
        as (insert into listeners (address, port) select init_listen_address, init_port from config where should_init returning id),
    handler as (insert into handlers (query) (select
                                                  query
                                              from
                                                  (values
                                                       ($$
       SELECT omni_httpd.default_page(request.*::omni_httpd.http_request) FROM request $$)) v(query),
                                                                                            config
                                              where
                                                  should_init) returning id)
insert
into
    listeners_handlers (listener_id, handler_id)
select
    listener.id,
    handler.id
from
    config,
    listener,
    handler
where
    config.should_init;;
