create function api(path text,
                    server text default coalesce(current_setting('omni_kube.server', true),
                                                 'https://kubernetes.default.svc'),
                    cacert text default current_setting('omni_kube.cacert', true),
                    clientcert omni_httpc.client_certificate default row (current_setting('omni_kube.clientcert', true), current_setting('omni_kube.client_private_key', true))::omni_httpc.client_certificate,
                    token text default current_setting('omni_kube.token', true),
                    method omni_http.http_method default 'GET',
                    body jsonb default null) returns jsonb
    language plpgsql
    stable
as
$$
declare
    response omni_httpc.http_response;
    body jsonb;
    request_digest  text;
    cached_response jsonb;
begin
    if cacert is null then
        select p.cacert into cacert from omni_kube.pod_credentials() p;
    end if;
    if token is null then
        select p.token into token from omni_kube.pod_credentials() p;
    end if;
    if substring(path, 1, 1) != '/' then
        raise exception 'path must start with a leading slash';
    end if;

    request_digest := encode(digest(method || ' ' || server || path || coalesce(cacert, 'NULL_CACERT') ||
                                    coalesce(token, 'NULL_TOKEN') || coalesce(body, 'null'), 'sha256'), 'hex');
    cached_response := omni_var.get_statement('omni_kube.request_' || request_digest, null::jsonb);
    if cached_response is not null then
        return cached_response;
    end if;
    select *
    into response
    from
        omni_httpc.http_execute_with_options(
                omni_httpc.http_execute_options(allow_self_signed_cert => cacert is null,
                                                cacerts => case when cacert is null then null else array [cacert] end,
                                                clientcert => case when clientcert is null then null else clientcert end),
                omni_httpc.http_request(
                        format('%1$s%2$s', server, path),
                        method,
                        array [
                            omni_http.http_header('Content-Type', 'application/json'),
                            omni_http.http_header('Accept', 'application/json'),
                            omni_http.http_header('Authorization', 'Bearer ' || token)
                            ],
                        convert_to(body::text, 'UTF8')
                ));
    if response.error is not null then
        raise exception 'error: %', response.error;
    end if;
    body := convert_from(response.body, 'utf-8')::jsonb;
    if response.status >= 400 then
        raise exception '%', body ->> 'reason' using detail = body ->> 'message';
    end if;
    return omni_var.set_statement('omni_kube.request_' || request_digest, body);
end;
$$;
;
create function pod_credentials()
    returns table
            (
                token  text,
                cacert text
            )
    security definer
    language plpgsql
as
$$
begin
    if pg_stat_file('/var/run/secrets/kubernetes.io/serviceaccount/token', true) is distinct from null then
        token := pg_read_file('/var/run/secrets/kubernetes.io/serviceaccount/token');
    end if;
    if pg_stat_file('/var/run/secrets/kubernetes.io/serviceaccount/ca.crt', true) is distinct from null then
        cacert := pg_read_file('/var/run/secrets/kubernetes.io/serviceaccount/ca.crt');
    end if;
    if not (token is null and cacert is null) then
        return next;
    end if;
    return;
end;
$$;
;
-- Namespaces
create view namespaces_with_data as
select (data -> 'metadata' ->> 'uid')::uuid       as uid,
       data -> 'metadata' ->> 'name'              as name,
       (data -> 'metadata' ->> 'creationTimestamp')::timestamptz as creation_timestamp,
       data -> 'status' ->> 'phase'               as phase,
       data
from (select jsonb_array_elements((api('/api/v1/namespaces'))::jsonb -> 'items') as data)
         as namespaces;

create view namespaces as
select uid, name, creation_timestamp, phase
from namespaces_with_data;

-- Pods
create view pods_with_data as
select (data -> 'metadata' ->> 'uid')::uuid                                                                   as uid,
       data -> 'metadata' ->> 'name'                                                                          as name,
       data -> 'metadata' ->> 'namespace'                                                                     as namespace,
       (data -> 'metadata' ->> 'creationTimestamp')::timestamptz as creation_timestamp,
       data -> 'status' ->> 'phase'                                                                           as phase,
       data -> 'spec' ->> 'nodeName'                                                                          as node,
       data
from (select jsonb_array_elements((api('/api/v1/pods'))::jsonb -> 'items') as data)
         as pods;

create view pods as
select uid, name, namespace, creation_timestamp, phase, node
from pods_with_data;

create view pod_labels as
select uid, name, key, value
from pods_with_data,
     lateral (select key, value from jsonb_each(data -> 'metadata' -> 'labels') as label(key, value)) as labels;

create view pod_ip_addresses as
select uid, name, ip_address
from pods_with_data,
     lateral (select jsonb_array_elements_text(jsonb_path_query_array(data, '$.status.podIPs[*].ip'))::inet as ip_address) as ip_addresses(ip_address);

-- Nodes

create view nodes_with_data as
select (data -> 'metadata' ->> 'uid')::uuid          as uid,
       data -> 'metadata' ->> 'name'                 as name,
       (data -> 'metadata' ->> 'creationTimestamp')::timestamptz as creation_timestamp,
       (data -> 'spec' ->> 'unschedulable')::boolean as unschedulable,
       data
from (select jsonb_array_elements((api('/api/v1/nodes'))::jsonb -> 'items') as data)
         as codes;

create view nodes as
select uid, name, creation_timestamp, unschedulable
from nodes_with_data;

create view node_cidrs as
select uid, name, cidr
from nodes_with_data,
     lateral (select jsonb_array_elements_text(jsonb_path_query_array(data, '$.spec.podCIDRs[*]'))::cidr as cidr) as cidrs(cidr);

create view node_images as
select data -> 'metadata' ->> 'name'                            as node,
       jsonb_array_elements(case
                                when jsonb_typeof(image_data -> 'names') = 'null' then '[
                                  null
                                ]'::jsonb
                                else image_data -> 'names' end) as image,
       image_data ->> 'sizeBytes' as size_bytes
from (select jsonb_array_elements((api('/api/v1/nodes'))::jsonb -> 'items') as data)
         as codes,
     lateral (select jsonb_path_query(data, '$.status.images[*]') as image_data ) as images;

create view node_labels as
select uid, name, key, value
from nodes_with_data,
     lateral (select key, value from jsonb_each(data -> 'metadata' -> 'labels') as label(key, value)) as labels;

-- Services

create view services_with_data as
select (data -> 'metadata' ->> 'uid')::uuid       as uid,
       data -> 'metadata' ->> 'name'              as name,
       data -> 'metadata' ->> 'namespace'         as namespace,
       (data -> 'metadata' ->> 'creationTimestamp')::timestamptz as creation_timestamp,
       data -> 'spec' ->> 'type'                  as type,
       data -> 'spec' ->> 'externalName'          as external_name,
       data -> 'spec' ->> 'internalTrafficPolicy' as internal_traffic_policy,
       data -> 'spec' ->> 'externalTrafficPolicy' as external_traffic_policy,
       data -> 'spec' ->> 'ipFamilyPolicy'        as ip_family_policy,
       data
from (select jsonb_array_elements((api('/api/v1/services'))::jsonb -> 'items') as data)
         as services;

create view services as
select uid,
       name,
       namespace,
       creation_timestamp,
       type,
       external_name,
       internal_traffic_policy,
       external_traffic_policy,
       ip_family_policy
from services_with_data;

create view service_ports as
select uid,
       name,
       port ->> 'name'              as port_name,
       (port ->> 'port')::int       as port,
       (port ->> 'targetPort')::int as target_port,
       port ->> 'protocol'          as protocol
from services_with_data,
     lateral (select jsonb_path_query(data, '$.spec.ports[*]') as port) as ports(port);

create view service_cluster_ips as
select uid,
       name,
       ip_address
from services_with_data,
     lateral (select (jsonb_path_query(data, '$.spec.clusterIPs[*]') #>> '{}')::inet as ip_address) as ips(ip_address);

create view service_external_ips as
select uid,
       name,
       ip_address
from services_with_data,
     lateral (select (jsonb_path_query(data, '$.spec.externalIPs[*]') #>> '{}')::inet as ip_address) as ips(ip_address);

create view service_ip_families as
select uid,
       name,
       ip_family
from services_with_data,
     lateral (select (jsonb_path_query(data, '$.spec.ipFamilies[*]') #>> '{}') as ip_family) as ip_families(ip_family);

create view service_labels as
select uid, name, key, value
from services_with_data,
     lateral (select key, value from jsonb_each(data -> 'metadata' -> 'labels') as label(key, value)) as labels;

--- Jobs

create view jobs_with_data as
select (data -> 'metadata' ->> 'uid')::uuid       as uid,
       data -> 'metadata' ->> 'name'              as name,
       data -> 'metadata' ->> 'namespace' as namespace,
       (data -> 'metadata' ->> 'creationTimestamp')::timestamptz as creation_timestamp,
       (data -> 'status' ->> 'startTime')::timestamptz           as start_time,
       (data -> 'status' ->> 'completionTime')::timestamptz      as completion_time,
       data -> 'spec' ->> 'completionMode'        as completion_mode,
       (data -> 'spec' ->> 'completions')::int    as completions,
       (data -> 'spec' ->> 'parallelism')::int    as parallelism,
       (data -> 'status' ->> 'active')::int       as active_pods,
       (data -> 'status' ->> 'failed')::int       as failed_pods,
       (data -> 'status' ->> 'ready')::int        as ready_pods,
       (data -> 'status' ->> 'succeeded')::int    as succeeded_pods,
       data
from (select jsonb_array_elements((api('/apis/batch/v1/jobs'))::jsonb -> 'items') as data)
         as jobs;

create view jobs as
select uid,
       name,
       namespace,
       creation_timestamp,
       start_time,
       completion_time,
       completion_mode,
       completions,
       parallelism,
       active_pods,
       failed_pods,
       ready_pods,
       succeeded_pods
from jobs_with_data;

create view job_labels as
select uid, name, key, value
from jobs_with_data,
     lateral (select key, value from jsonb_each(data -> 'metadata' -> 'labels') as label(key, value)) as labels;;
