create type requirement as
(
    name    text,
    version text
);

create function requirement(name text, version text) returns requirement
    immutable
    language sql
as
$$
select row (name, version)::omni_manifest.requirement
$$;

create function requirement_to_text(requirement requirement) returns text
    language sql
    immutable as
$$
select
        requirement.name || case
                                when left(requirement.version, 1) = '='
                                    then requirement.version
                                else '=' || requirement.version end;
$$;
create function requirement_to_json(requirement requirement) returns json
    language sql
    immutable as
$$
select json_build_object(requirement.name, requirement.version)
$$;
create function requirement_to_jsonb(requirement requirement) returns jsonb
    language sql
    immutable as
$$
select jsonb_build_object(requirement.name, requirement.version)
$$;
create function requirements_to_text(requirements requirement[]) returns text
    language sql
    immutable as
$$
with
    reqs as (select
                 omni_manifest.requirement_to_text(row (name, version)) as requirement
             from
                 unnest(requirements))
select
    string_agg(requirement, ',')
from
    reqs
$$;
create function requirements_to_json(requirements requirement[]) returns json
    language sql
    immutable as
$$
select
    json_object_agg(name, version)
from
    unnest(requirements) r(name, version)
$$;
create function requirements_to_jsonb(requirements requirement[]) returns jsonb
    language sql
    immutable as
$$
select
    jsonb_object_agg(name, version)
from
    unnest(requirements) r(name, version)
$$;
create function text_to_requirement(requirement text) returns requirement
    language sql
    immutable as
$$
with
    arr as (select regexp_split_to_array(requirement, '=') val)
select
    row (arr.val[1], arr.val[2])::omni_manifest.requirement
from
    arr
$$;
create function json_to_requirement(requirement json) returns requirement
    language sql
    immutable as
$$
select
    row (key, value)::omni_manifest.requirement
from
    json_each_text(requirement)
limit 1
$$;
create function jsonb_to_requirement(requirement jsonb) returns requirement
    language sql
    immutable as
$$
select
    row (key, value)::omni_manifest.requirement
from
    jsonb_each_text(requirement)
limit 1
$$;
create function text_to_requirements(requirements text) returns requirement[]
    language sql
    immutable as
$$
select
    array_agg(omni_manifest.text_to_requirement(requirement))
from
    regexp_split_to_table(requirements, '\s*,\s*') r(requirement)
$$;
create function json_to_requirements(requirements json) returns requirement[]
    language sql
    immutable as
$$
select
    array_agg(row (key, value)::omni_manifest.requirement)
from
    json_each_text(requirements)
$$;
create function jsonb_to_requirements(requirements jsonb) returns requirement[]
    language sql
    immutable as
$$
select
    array_agg(row (key, value)::omni_manifest.requirement)
from
    jsonb_each_text(requirements)
$$;

create cast (requirement as text) with function requirement_to_text as implicit;
create cast (requirement as json) with function requirement_to_json as implicit;
create cast (requirement as jsonb) with function requirement_to_jsonb as implicit;
create cast (requirement[] as text) with function requirements_to_text as implicit;
create cast (requirement[] as json) with function requirements_to_json as implicit;
create cast (requirement[] as jsonb) with function requirements_to_jsonb as implicit;
create cast (text as requirement) with function text_to_requirement as implicit;
create cast (json as requirement) with function json_to_requirement as implicit;
create cast (jsonb as requirement) with function jsonb_to_requirement as implicit;
create cast (text as requirement[]) with function text_to_requirements as implicit;
create cast (json as requirement[]) with function json_to_requirements as implicit;
create cast (jsonb as requirement[]) with function jsonb_to_requirements as implicit;

create type artifact as
(
    self requirement,
    requirements requirement[]
);

create function artifact(self requirement, requirements requirement[]) returns artifact
    immutable
    language sql
as
$$
select row (self, requirements)::omni_manifest.artifact
$$;
;
create function install_plan(artifacts artifact[])
    returns requirement[]
    language plpgsql
as
$$
declare
    plan omni_manifest.requirement[];
    artifact             omni_manifest.artifact;
    req                  omni_manifest.requirement;
    rec  record;
    _id  integer;
    _id_ integer;
    c    int;
begin
    create temporary table omni_manifest__deps__
    (
        id                       integer primary key generated always as identity,
        requirement              omni_manifest.requirement unique,
        outstanding_dependencies int not null
    ) on
          commit drop;

    create temporary table omni_manifest__parents__
    (
        child  int,
        parent int,
        constraint u unique (child, parent)
    )
        on commit drop;
    foreach artifact in array artifacts
        loop
            insert
            into
                omni_manifest__deps__ (requirement, outstanding_dependencies)
            values ((artifact).self, coalesce(cardinality((artifact).requirements), 0))
            on conflict (requirement) do update set outstanding_dependencies = coalesce(cardinality((artifact).requirements), 0)
            returning id into _id;

            if cardinality((artifact).requirements) > 0 then
                foreach req in array (artifact).requirements
                    loop
                        insert
                        into
                            omni_manifest__deps__ (requirement, outstanding_dependencies)
                        values (req, 0)
                        on conflict (requirement) do update set requirement = excluded.requirement
                        returning id
                            into _id_;
                        insert
                        into
                            omni_manifest__parents__
                        values (_id_, _id)
                        on conflict (child, parent) do nothing;
                    end loop;
            end if;

        end loop;

    c := 1;
    while c > 0
        loop
            select count(*) from omni_manifest__deps__ where omni_manifest__deps__.outstanding_dependencies > 0 into c;
            for rec in select * from omni_manifest__deps__ order by (requirement).name, (requirement).version asc
        loop
                    if rec.outstanding_dependencies = 0 then
                        plan = array_append(plan, rec.requirement);
                        delete from omni_manifest__deps__ where omni_manifest__deps__.id = rec.id;
                        update omni_manifest__deps__
                        set
                            outstanding_dependencies = outstanding_dependencies - 1
                        from
                            omni_manifest__parents__
                        where
                            parent = id and
                            child = rec.id;
                    end if;
                end loop;
        end loop;

    drop table omni_manifest__deps__;
    drop table omni_manifest__parents__;

    return plan;
end;
$$;
;
create type requirement_status as enum ('installed', 'missing', 'updated');

create type install_report as
(
    requirement requirement,
    status      requirement_status
);
create function install(artifacts artifact[]) returns setof install_report
    language plpgsql
as
$$
declare
    rec record;
    missing_requirements bool = false;
    plan omni_manifest.requirement[];
begin
    plan = omni_manifest.install_plan(artifacts);
    -- Check if we're missing some artifacts
    for rec in
        select
            row (r.name, r.version)::omni_manifest.requirement as req,
            e.name                                                      as ext_name,
            case when r.version = '*' then e.version else r.version end as target_version,
            pg_extension.extversion                                     as installed_version
        from
            unnest(plan) with ordinality r(name, version, ord)
            left outer join pg_extension on pg_extension.extname = r.name
            left outer join lateral
                (select name, version from pg_available_extension_versions where r.version != '*'
                 union select name, default_version as version from pg_available_extensions where r.version = '*') e
                on e.name = r.name and (e.version = r.version or r.version = '*')
        order by ord
        loop
            if rec.ext_name is null then
                missing_requirements = true;
                return next row (rec.req, 'missing')::omni_manifest.install_report;
            else
                if rec.target_version = rec.installed_version then
                    return next row (rec.req, 'kept')::omni_manifest.install_report;
                    continue;
                end if;
                if not missing_requirements then

                    if rec.installed_version is null then
                        execute format('create extension %s version %L', (rec.req).name, rec.target_version);
                        return next row (rec.req, 'installed')::omni_manifest.install_report;
                    else
                        execute format('alter extension %s update to %L', (rec.req).name, rec.target_version);
                        return next row (rec.req, 'updated')::omni_manifest.install_report;
                    end if;
                end if;
            end if;
        end loop;
    return;
end;
$$;
;
create function text_to_artifact(artifact text) returns artifact
    language sql
    immutable as
$$
with
    arr as (select regexp_split_to_array(artifact, '#') val)
select
    row (arr.val[1]::omni_manifest.requirement, arr.val[2]::omni_manifest.requirement[])::omni_manifest.artifact
from
    arr
$$;
create function text_to_artifacts(artifacts text) returns artifact[]
    language sql
    immutable as
$$
select
            array_agg(omni_manifest.text_to_artifact(artifact)) filter (where length(artifact) > 0)
from
    regexp_split_to_table(artifacts, '\s*(\n+|;)\s*') r(artifact)
$$;
create function json_to_artifact(artifact json) returns artifact
    language sql
    immutable as
$$
select
    row ((artifact -> 'target')::omni_manifest.requirement, (artifact -> 'requirements')::omni_manifest.requirement[])::omni_manifest.artifact
$$;
create function jsonb_to_artifact(artifact jsonb) returns artifact
    language sql
    immutable as
$$
select
    row ((artifact -> 'target')::omni_manifest.requirement, (artifact -> 'requirements')::omni_manifest.requirement[])::omni_manifest.artifact
$$;
create function json_to_artifacts(artifacts json) returns artifact[]
    language sql
    immutable as
$$
with
    artifacts_ as (select value::omni_manifest.artifact from json_array_elements(artifacts) t(value))
select
    array_agg(value)
from
    artifacts_
$$;
create function jsonb_to_artifacts(artifacts jsonb) returns artifact[]
    language sql
    immutable as
$$
with
    artifacts_ as (select value::omni_manifest.artifact from jsonb_array_elements(artifacts) t(value))
select
    array_agg(value)
from
    artifacts_
$$;
create function artifact_to_json(artifact artifact) returns json
    language sql
    immutable as
$$
select
    json_build_object('target', artifact.self::json,
                      'requirements', artifact.requirements::json)
$$;
create function artifacts_to_json(artifacts artifact[]) returns json
    language sql
    immutable as
$$
select
    json_agg(a::json)
from
    unnest(artifacts) a
$$;
create function artifact_to_jsonb(artifact artifact) returns jsonb
    language sql
    immutable as
$$
select
    jsonb_build_object('target', artifact.self::jsonb,
                       'requirements', artifact.requirements::jsonb)
$$;
create function artifacts_to_jsonb(artifacts artifact[]) returns jsonb
    language sql
    immutable as
$$
select
    jsonb_agg(a::jsonb)
from
    unnest(artifacts) a
$$;
create function artifact_to_text(artifact artifact) returns text
    language sql
    immutable as
$$
select
    artifact.self::text || case
                               when artifact.requirements is null then ''
                               else '#' || artifact.requirements::text end
$$;
create function artifacts_to_text(artifacts artifact[]) returns text
    language sql
    immutable as
$$
with
    arts as (select
                 a::text
             from
                 unnest(artifacts) a)
select
    string_agg(a, ';')
from
    arts
$$;

create cast (text as artifact) with function text_to_artifact as implicit;
create cast (text as artifact[]) with function text_to_artifacts as implicit;
create cast (json as artifact) with function json_to_artifact as implicit;
create cast (jsonb as artifact) with function jsonb_to_artifact as implicit;
create cast (json as artifact[]) with function json_to_artifacts as implicit;
create cast (jsonb as artifact[]) with function jsonb_to_artifacts as implicit;
create cast (artifact as json) with function artifact_to_json as implicit;
create cast (artifact[] as json) with function artifacts_to_json as implicit;
create cast (artifact as jsonb) with function artifact_to_jsonb as implicit;
create cast (artifact[] as jsonb) with function artifacts_to_jsonb as implicit;
create cast (artifact as text) with function artifact_to_text as implicit;
create cast (artifact[] as text) with function artifacts_to_text as implicit;;
alter type omni_manifest.requirement_status add value 'kept';;
