create or replace function instantiate(schema name default 'omni_rest')
    returns void
    language plpgsql
as
$instantiate$
begin
    perform set_config('search_path', schema::text || ',public', true);

    create type postgrest_settings as
    (
        -- The only schemas `postgrest` is allowed to access
        schemas name[]
    );

    create function postgrest_settings(schemas name[] default array []::name[])
    returns postgrest_settings
    language sql
    immutable
as
$$
select row (schemas)::omni_rest.postgrest_settings;
$$;

    create procedure _postgrest_relation(request omni_httpd.http_request,
                                     relation inout regclass,
                                     namespace inout text,
                                     settings postgrest_settings default postgrest_settings())
    language plpgsql
as
$$
declare
    saved_search_path text := current_setting('search_path');
begin

    if request.method = 'GET' then
        namespace := omni_http.http_header_get(request.headers, 'accept-profile');
    end if;

    if request.method in ('POST', 'PATCH') then
        namespace := omni_http.http_header_get(request.headers, 'content-profile');
    end if;

    if namespace is null and cardinality(settings.schemas) > 0 then
        namespace := settings.schemas[1]::text;
    end if;

    if namespace is null then
        return;
    end if;

    if not namespace::name = any (settings.schemas) then
        relation := null;
        namespace := null;
        return;
    end if;

    perform set_config('search_path', namespace, true);

    relation := to_regclass(split_part(request.path, '/', 2));
    if relation is null or relation::text like 'pg_%' then
        relation := null;
        return;
    end if;

    perform set_config('search_path', saved_search_path, false);

end;
$$;

    /*
 * Given a schema name, a function name and an array of argument names,
 * it finds all functions that could be called taking into account required and optional arguments.
 * For example, if you have 2 function
 *  1. public.say_hello(message text default 'hello') 
 *  2. public.say_hello() 
 * The call  _postgrest_function_by_arguments('public', 'say_hello', '{}') would return an array with a reference to each.
 * Meanwhile a call _postgrest_function_by_arguments('public', 'say_hello', '{message}') would return an array with only one reference to function 1.
 */
create or replace function _postgrest_function_by_arguments(namespace text, function_name text, function_arguments text[])
    returns regproc[]
    language sql
    stable
as
$$
with arguments as (select oid,
                          array_agg(coalesce(name, ''))
                          filter (where idx <= (pronargs - pronargdefaults)) as required_parameters,
                          array_agg(coalesce(name, ''))                      as parameters
                   from pg_proc,
                        unnest(proargnames, proargtypes, proargmodes)
                            with ordinality as _ (name, type, mode, idx)
                   where type is not null -- only input arguments
                   group by oid
                   -- we can only work with named arguments
                   having count(*) - count(name) = 0)
select array_agg(p.oid::regproc)
from pg_proc p
         left join arguments a on a.oid = p.oid
         left join pg_catalog.pg_namespace n on n.oid = p.pronamespace
where n.nspname = namespace
  and p.proname = function_name
  and coalesce(required_parameters, '{}') <@ function_arguments
  and (function_arguments <@ coalesce(parameters, '{}')
    or function_arguments = coalesce(parameters, '{}'))
$$;

create procedure _postgrest_function(request omni_httpd.http_request, function_reference inout regproc,
                                     namespace inout text, settings postgrest_settings default postgrest_settings())
    language plpgsql
as
$$
declare
    saved_search_path text := current_setting('search_path');
    function_name     text;
    function_arguments text[];
    all_references    regproc[];
begin
    if request.method = 'GET' then
        namespace := omni_http.http_header_get(request.headers, 'accept-profile');
    else
        namespace := omni_http.http_header_get(request.headers, 'content-profile');
    end if;
    if namespace is null and cardinality(settings.schemas) > 0 then
        namespace := settings.schemas[1]::text;
    end if;
    if namespace is null then
        return;
    end if;
    if not namespace::name = any (settings.schemas) then
        function_reference := null;
        namespace := null;
        return;
    end if;
    perform
        set_config('search_path', namespace, true);
    function_name := split_part(request.path, '/', 3);
    select case request.method
               when 'GET' then
                   (select array_agg(name)
                    from unnest(omni_web.parse_query_string(request.query_string))
                             with ordinality as _ (name, i)
                    where i % 2 = 1)
               when 'POST' then
                   (select array_agg(jsonb_object_keys)
                    from
                        jsonb_object_keys(convert_from(request.body, 'utf-8')::jsonb))
               end
    into function_arguments;
    all_references :=
            omni_rest._postgrest_function_by_arguments(namespace, function_name, coalesce(function_arguments, '{}'));
    select all_references[1]
    into function_reference;
    if function_reference is null or function_reference::text like 'pg_%' then
        function_reference := null;
        return;
    end if;
    perform
        set_config('search_path', saved_search_path, false);
end;
$$;



    create function __eval(code text)
    returns text
    language plpgsql as
$$
declare
    result text;
begin
    execute 'select ' || code into result;
    return result;
end;
$$;

    create function postgrest_convert_operator(operator_name text)
    returns text
    immutable
    language sql
as
$$
select case operator_name
           when 'eq' then
               '='
           when 'neq' then
               '<>'
           when 'isdistinct' then
               'is distinct from '
           when 'gt' then
               '>'
           when 'lt' then
               '<'
           when 'gte' then
               '>='
           when 'lte' then
               '<='
           when 'like' then
               'like'
           when 'ilike' then
               'ilike'
           when 'match' then
               '~'
           when 'imatch' then
               '~*'
           when 'not' then
               'not'
           end;
$$;

create or replace function postgrest_parse_logical(input text)
    returns jsonb
    language plpgsql
as
$$
declare
    result             jsonb := '[]';
    -- initialize an empty json array
    current            text  := '';
    current_production text  := '';
    depth              int   := 0;
    token              text;
begin
    -- no nesting, just split and return
    if input !~ '[()]' then
        select jsonb_agg(jsonb_build_object('operator', omni_rest.postgrest_convert_operator(split_part(term, '.', 2)),
                                            'operands',
                                            jsonb_build_array(split_part(term, '.', 1), split_part(term, '.', 3))))
        from unnest(string_to_array(input, ',')) term
        into result;
        return result;
    end if;
    -- iterate through the string character by character
    foreach token in array regexp_split_to_array(input, '')
        loop
            if token = '(' then
                if depth > 0 then
                    current := current || token;
                end if;
                depth := depth + 1;
            elsif token = ')' then
                depth := depth - 1;
                if depth > 0 then
                    current := current || token;
                end if;
                -- if we finish a nested expression, process it
                if depth = 0 then
                    result := result || case lower(current_production)
                                            when '' then
                                                omni_rest.postgrest_parse_logical(current)
                                            when 'or' then
                                                jsonb_build_object('operator', 'or', 'operands',
                                                                   omni_rest.postgrest_parse_logical(current))
                                            when 'and' then
                                                jsonb_build_object('operator', 'and', 'operands',
                                                                   omni_rest.postgrest_parse_logical(current))
                                            when 'not.or' then
                                                jsonb_build_object('operator', 'not', 'operands',
                                                                   jsonb_build_array(jsonb_build_object('operator',
                                                                                                        'or',
                                                                                                        'operands',
                                                                                                        omni_rest.postgrest_parse_logical(current))))
                                            when 'not.and' then
                                                jsonb_build_object('operator', 'not', 'operands',
                                                                   jsonb_build_array(jsonb_build_object('operator',
                                                                                                        'and',
                                                                                                        'operands',
                                                                                                        omni_rest.postgrest_parse_logical(current))))
                                            when 'not' then
                                                jsonb_build_object('operator', 'not', 'operands',
                                                                   omni_rest.postgrest_parse_logical(current))
                        end;
                    current := '';
                end if;
            elsif depth > 0 then
                current := current || token;
            end if;
            if depth = 0 then
                if token = ',' then
                    if current_production !~ '[()]' then
                        result := result || jsonb_build_array(jsonb_build_object('operator',
                                                                                 omni_rest.postgrest_convert_operator(split_part(current_production, '.', 2)),
                                                                                 'operands', jsonb_build_array(
                                                                                         split_part(current_production, '.', 1),
                                                                                         split_part(current_production, '.', 3))));
                    end if;
                    current_production := '';
                else
                    current_production := current_production || token;
                end if;
            end if;
        end loop;
    if current_production !~ '[()]' then
        result := result || jsonb_build_array(jsonb_build_object('operator',
                                                                 omni_rest.postgrest_convert_operator(split_part(current_production, '.', 2)),
                                                                 'operands', jsonb_build_array(
                                                                         split_part(current_production, '.', 1),
                                                                         split_part(current_production, '.', 3))));
    end if;
    return result;
end;
$$;

create function postgrest_parse_get_param(params text[])
    returns jsonb
    immutable
    language sql
as
$$
with indexed_params as (select *
                        from unnest(params)
                                 with ordinality r (param, index)),
     comparisons as (select lower(columns.param)                                                   as key,
                            values.param                                                           as value,
                            omni_rest.postgrest_convert_operator(split_part(values.param, '.', 1)) as first_operator,
                            omni_rest.postgrest_convert_operator(split_part(values.param, '.', 2)) as second_operator
                     from indexed_params columns
                              join indexed_params values
                                   on columns.index % 2 = 1
                                       and values.index % 2 = 0
                                       and columns.index = values.index - 1)
select jsonb_build_object('operator', 'and', 'operands', coalesce(jsonb_agg(
                                                                          case
                                                                              when key = 'not.and' then
                                                                                  jsonb_build_object('operator', 'not',
                                                                                                     'operands',
                                                                                                     jsonb_build_array(jsonb_build_object(
                                                                                                             'operator',
                                                                                                             'and',
                                                                                                             'operands',
                                                                                                             omni_rest.postgrest_parse_logical(value))))
                                                                              when key = 'and' then
                                                                                  jsonb_build_object('operator', 'and',
                                                                                                     'operands',
                                                                                                     omni_rest.postgrest_parse_logical(value))
                                                                              when key = 'not.or' then
                                                                                  jsonb_build_object('operator', 'not',
                                                                                                     'operands',
                                                                                                     jsonb_build_array(jsonb_build_object(
                                                                                                             'operator',
                                                                                                             'or',
                                                                                                             'operands',
                                                                                                             omni_rest.postgrest_parse_logical(value))))
                                                                              when key = 'or' then
                                                                                  jsonb_build_object('operator', 'or',
                                                                                                     'operands',
                                                                                                     omni_rest.postgrest_parse_logical(value))
                                                                              when first_operator = 'not' then
                                                                                  jsonb_build_object('operator', 'not',
                                                                                                     'operands',
                                                                                                     jsonb_build_array(jsonb_build_object(
                                                                                                             'operator',
                                                                                                             second_operator,
                                                                                                             'operands',
                                                                                                             jsonb_build_array(key, split_part(value, '.', 3)))))
                                                                              when first_operator is not null then
                                                                                  jsonb_build_object('operator',
                                                                                                     first_operator,
                                                                                                     'operands',
                                                                                                     jsonb_build_array(key, (regexp_match(value, '\.(.*)'))[1]))
                                                                              else
                                                                                  jsonb_build_object('operator',
                                                                                                     'invalid_operator',
                                                                                                     'operands',
                                                                                                     '[]'::jsonb)
                                                                              end), '[]'::jsonb))
from comparisons
where key not in ('select', 'order')
$$;

create function postgrest_format_get_param(where_ast jsonb)
    returns text
    immutable
    language plpgsql
as
$$
declare
    output text;
begin
    select case where_ast ->> 'operator'
               when 'invalid_operator' then
                   'invalid_operator invalid_operator'
               when 'not' then
                   format('not (%1$s)', omni_rest.postgrest_format_get_param(where_ast -> 'operands' -> 0))
               when 'and' then
                   (select string_agg(omni_rest.postgrest_format_get_param(op), ' and ')
                    from jsonb_array_elements(where_ast -> 'operands') op)
               when 'or' then
                   (select string_agg(omni_rest.postgrest_format_get_param(op), ' or ')
                    from jsonb_array_elements(where_ast -> 'operands') op)
               else
                   format('%1$I %2$s %3$L', where_ast -> 'operands' ->> 0, where_ast ->> 'operator',
                          where_ast -> 'operands' ->> 1)
               end
    into output;
    return output;
end;
$$;

/*
 * Given a namespace, reference to a relation and a text of comma separated field names
 * It returns an array of expressions to be used for vertical filtering.
 */
create function postgrest_select_columns(namespace text, relation regclass, _select text)
    returns text[]
    immutable
    language plpgsql
as
$$
declare
    query_columns text[] := array ['*'];
    col           text;
begin
    if _select is not null and _select <> '' then
        query_columns := array []::text[];


        foreach col in array string_to_array(_select, ',')
            loop

                declare
                    col_name    text;
                    col_expr    text;
                    col_alias   text;
                    _match      text[];
                    is_col_name bool;
                begin
                    col_expr := split_part(col, ':', 1);
                    col_alias := split_part(col, ':', 2);

                    _match := regexp_match(col_expr, '^([[:alpha:] _][[:alnum:] _]*)(.*)');
                    col_name := _match[1];
                    if col_name is null then
                        raise exception 'no column specified in %', col_expr;
                    end if;

                    -- Check if there is such an attribute
                    perform from pg_attribute where attname = col_name and attrelid = relation and attnum > 0;
                    is_col_name := found;
                    -- If not...
                    if not is_col_name then
                        -- Is it a function?
                        perform
                        from pg_proc
                        where proargtypes[0] = relation
                          and proname = col_name
                          and pronamespace::text = namespace;
                        if found then
                            -- It is a function
                            col_name := col_name || '((' || relation || '))';
                        end if;
                    end if;
                    col_expr := col_name || _match[2];

                    if col_alias = '' then
                        col_alias := col_expr;
                    end if;
                    if col_expr ~ '->' then
                        -- Handle JSON expression
                        if col_alias = col_expr then
                            -- Update the alias if it was not set
                            with arr as (select regexp_split_to_array(col_expr, '->>?') as data)
                            select data[cardinality(data)]
                            from arr
                            into col_alias;
                        end if;
                        -- Rewrite the expression to match its actual syntax
                        col_expr :=
                                regexp_replace(col_expr, '(->>?)([[:alpha:]_][[:alnum:]_]*)(?=(->|$))', '\1''\2''',
                                               'g');
                        -- TODO: record access using ->
                    end if;
                    -- TODO: sanitize col_expr
                    query_columns := query_columns || (relation || '.' || col_expr || ' as "' || col_alias || '"');


                end;
            end loop;

    end if;
    return query_columns;
end;
$$;

create procedure postgrest_get(request omni_httpd.http_request, outcome inout omni_httpd.http_outcome,
                               settings postgrest_settings default postgrest_settings())
    language plpgsql as
$$
declare
    namespace         text;
    query_columns     text[];
    new_query_columns text[];
    query             text;
    result            jsonb;
    params            text[];
    col               text;
    _select           text;
    _where            text;
    _order            text;
    _offset           numeric;
    relation          regclass;
    request_cache     text;
begin
    if outcome is distinct from null then
        return;
    end if;
    if request.method = 'GET' then
        call omni_rest._postgrest_relation(request, relation, namespace, settings);
        if relation is null then
            return; -- terminate
        end if;
    else
        return; -- terminate;
    end if;

    request_cache := 'omni_rest.query_' ||
                     encode(digest(relation || ' ' || coalesce(request.query_string, ''), 'sha256'), 'hex');
    query := omni_var.get_session(request_cache, null::text);
    if query is null then
        params := omni_web.parse_query_string(request.query_string);

        -- Columns (vertical filtering)
        _select = omni_web.param_get(params, 'select');
        query_columns := omni_rest.postgrest_select_columns(namespace, relation, _select);

        _where := omni_rest.postgrest_format_get_param(omni_rest.postgrest_parse_get_param(params));

        _offset := 0;
        select 'ORDER BY ' || string_agg(format('%1$I %2$s', split_part(o, '.', 1),
                                             -- FIXME: we should error instead of using a default on invalid ordering clauses
                                                case lower(split_part(o, '.', 2))
                                                    when 'desc' then
                                                        'desc'
                                                    else
                                                        'asc'
                                                    end), ',')
        from unnest(regexp_split_to_array(omni_web.param_get(params, 'order'), ',')) o
        into _order;
        -- Finalize the query
        query := format('select %3$s from %1$I.%2$I where %4$s %5$s', namespace, (select relname
                                                                                  from pg_class
                                                                                  where oid = relation),
                        concat_ws(', ', variadic query_columns), coalesce(_where, 'true'), _order);
        perform omni_var.set_session(request_cache, query);
    end if;
    -- Run it
    declare
        message text;
        detail  text;
        hint    text;
    begin
        select coalesce(jsonb_agg(stmt_row), '[]'::jsonb)
        into result
        from
            omni_sql.execute(query);
        outcome := omni_httpd.http_response(result, headers => array [omni_http.http_header('Content-Range',
                                                                                            _offset || '-' ||
                                                                                            jsonb_array_length(result) ||
                                                                                            '/*')]);
    exception
        when others then
            get stacked diagnostics message = message_text,
                detail = pg_exception_detail,
                hint = pg_exception_hint;
            outcome := omni_httpd.http_response(jsonb_build_object('message', message, 'detail', detail, 'hint', hint),
                                                status => 400);
    end;

end;
$$;

    /*
 * Given a function reference and an array of argument names,
 * it builds a string with placeholders for each argument cast to its type.
 * The placeholders are in the order of the function argument list.
 * This can be used with _postgrest_function_ordered_argument_values to build a SQL command that will call the referenced function.
 */
create or replace function _postgrest_function_call_arguments(fn regproc, passed_arguments text[])
    returns text
    immutable
    language sql
as
$$
select string_agg(format('%1$I => $%2$s :: %3$I', name, idx, t.typname), ', ' order by idx)
from pg_proc p,
     unnest(proargnames, proargtypes, proargmodes)
         with ordinality as _ (name, type, mode, idx)
         join pg_type t on t.oid = type
where type is not null
  and p.oid = fn
  and name::text = any (passed_arguments)
$$;

/*
 * Given a function reference and an array of argument names and a jsonb object in the shape of {[argument_name]:argument_value},
 * it returns a jsonb array with the argument values in the same order as the array inn passed_arguments.
 * This can be used with postgrest_function_call_arguments to build a SQL command that will call the referenced function.
 */
create or replace function _postgrest_function_ordered_argument_values(fn regproc, passed_arguments text[], passed_values jsonb)
    returns jsonb
    immutable
    language sql
as
$$
select jsonb_agg(passed_values ->> (name::text)
                 order by idx)
from pg_proc p,
     unnest(proargnames, proargtypes, proargmodes)
         with ordinality as _ (name, type, mode, idx)
         join pg_type t on t.oid = type
where type is not null
  and p.oid = fn
  and name::text = any (passed_arguments)
$$;

create procedure postgrest_rpc(request omni_httpd.http_request, outcome inout omni_httpd.http_outcome,
                               settings postgrest_settings default postgrest_settings())
    language plpgsql
as
$$
declare
    function_reference regproc;
    query              text;
    namespace          text;
    result             jsonb;
    arguments_definition text;
    argument_values    jsonb;
    passed_arguments   text[];
begin
    if outcome is distinct from null then
        return;
    end if;
    if request.method in ('GET', 'POST') then
        if split_part(request.path, '/', 2) <> 'rpc' then
            return;
        end if;
        call omni_rest._postgrest_function(request, function_reference, namespace, settings);
        if function_reference is null then
            return;
            -- terminate
        end if;
    else
        return;
        -- terminate;
    end if;
    select case request.method
               when 'GET' then
                   -- get the list of passed argument names from the query string
                   (select array_agg(name)
                    from unnest(omni_web.parse_query_string(request.query_string))
                             with ordinality as _ (name, i)
                    where i % 2 = 1)
               when 'POST' then
                   -- get the list of passed argument names from the request body (assume it's a json object)
                   (select array_agg(jsonb_object_keys)
                    from
                        jsonb_object_keys(convert_from(request.body, 'utf-8')::jsonb))
               end
    into passed_arguments;
    arguments_definition :=
            coalesce(omni_rest._postgrest_function_call_arguments(function_reference, passed_arguments), '');
    if (select proretset from pg_proc where oid = function_reference) then
        query := format('select jsonb_agg(row_to_json(r.*)::jsonb) as result from %1$s(%2$s) r', function_reference,
                        arguments_definition);
    else
        query := format('select %1$s(%2$s) as result', function_reference, arguments_definition);
    end if;
    if request.method = 'GET' then
        query := format('set transaction read only; %1$s', query);
        select (select jsonb_object_agg(keys.key, values.value)
                from unnest(omni_web.parse_query_string(request.query_string))
                         with ordinality as keys (key, ikey)
                         join unnest(omni_web.parse_query_string(request.query_string))
                    with ordinality as values (value,
                                               ivalue) on ikey % 2 = 1
                    and ivalue % 2 = 0
                    and ikey = ivalue - 1)
        into argument_values;
    else
        select (select convert_from(request.body, 'utf-8')::jsonb)
        into argument_values;
    end if;
    -- Run it
    declare
        message text;
        detail  text;
        hint    text;
    begin
        select omni_sql.execute(query, coalesce(
                omni_rest._postgrest_function_ordered_argument_values(function_reference, passed_arguments,
                                                                      argument_values), '[]'::jsonb)) -> 'result'
        into result;
        if jsonb_typeof(result) = 'array' then
            outcome := omni_httpd.http_response(result, headers => array [omni_http.http_header('Content-Range', '0-' ||
                                                                                                                 jsonb_array_length(result) ||
                                                                                                                 '/*')]);
        else
            outcome := omni_httpd.http_response(result);
        end if;
    exception
        when others then
            get stacked diagnostics message = message_text,
                detail = pg_exception_detail,
                hint = pg_exception_hint;
            outcome := omni_httpd.http_response(jsonb_build_object('message', message, 'detail', detail, 'hint', hint),
                                                status => 400);
    end;
end;
$$;


    create procedure postgrest_insert(request omni_httpd.http_request, outcome inout omni_httpd.http_outcome,
                                  settings postgrest_settings default postgrest_settings())
    language plpgsql as
$$
declare
    namespace text;
    query     text;
    result    jsonb;
    relation  regclass;
    preference text;
    _missing   text;
    _return    text := 'minimal';
    _tx        text;
begin
    if outcome is distinct from null then
        return;
    end if;

    if request.method = 'POST' then
        call omni_rest._postgrest_relation(request, relation, namespace, settings);
        if relation is null then
            return; -- terminate
        end if;
    else
        return; -- terminate;
    end if;

    for preference in select regexp_split_to_table(omni_http.http_header_get(request.headers, 'prefer'), ',\s+')
        loop
            declare
                preference_name  text := split_part(preference, '=', 1);
                preference_value text := split_part(preference, '=', 2);
            begin
                case
                    when preference_name = 'missing'
                        then _missing := preference_value;
                    when preference_name = 'return'
                        then _return := preference_value;
                    when preference_name = 'tx'
                        then _tx := preference_value;
                    end case;
            end;
        end loop;

    if omni_http.http_header_get(request.headers, 'content-type') = 'application/json' then
        declare
            payload  jsonb := convert_from(request.body, 'utf8')::jsonb;
            defaults jsonb;
        begin
            if jsonb_typeof(payload) = 'object' then
                payload := jsonb_build_array(payload);
            end if;
            if jsonb_typeof(payload) != 'array' then
                outcome := omni_httpd.http_response(status => 422, body => 'JSON must be object or array');
                return;
            end if;

            -- Handle defaults

            if _missing = 'default' then
                -- Prepare default expressions
                select jsonb_object_agg(attname,
                                        coalesce(pg_get_expr(adbin, adrelid), 'null'))
                into defaults
                from pg_attribute a
                         left join pg_attrdef d
                                   on a.
                                          attrelid = d.adrelid and a.attnum = d.adnum
                where a.attrelid = relation
                  and a.attnum > 0
                  and not a.attisdropped;

                -- Apply them to payload rows
                payload := jsonb_agg(
                                   p || (select jsonb_object_agg(
                                                        _defaults.key,
                                                        omni_rest.__eval(_defaults.value #>> '{}')::jsonb
                                                ) - array(
                                                        select jsonb_object_keys(p)
                                                    )
                                         from jsonb_each(defaults) _defaults)
                           )
                           from jsonb_array_elements(payload) p;
            end if;
            -- Query
            query :=
                    format(
                            'insert into %1$I.%2$I select * from jsonb_populate_recordset(null::%1$I.%2$I, %3$L::jsonb) %4$s',
                            namespace,
                            (select relname from pg_class where oid = relation),
                            payload::json,
                            case when _return = 'representation' then 'returning *' else '' end);
        end;
    end if;

    select jsonb_agg(stmt_row)
    into result
    from omni_sql.execute(query);

    if lower(_tx) = 'rollback' then
        rollback and chain;
    end if;

    outcome := omni_httpd.http_response(status => 201,
                                        body => case when _return = 'representation' then result end);
end;
$$;
    /*
 * Given a relation reference and an array of argument names,
 * it builds a string with placeholders for each argument cast to its type.
 * The placeholders are in the order of the relation definition.
 * This can be used with _postgrest_update_ordered_field_values to build a SQL command that will update the referenced relation.
 */
create or replace function _postgrest_update_fields(relation regclass, field_names text[])
    returns text
    immutable
    language sql
as
$$
select string_agg(format('%1$I = $%2$s :: %3$I', a.attname, pa.idx, t.typname), ', ' order by pa.idx)
from pg_attribute a
         join pg_type t on a.atttypid = t.oid
         join unnest(field_names) with ordinality as pa (name, idx) on pa.name = a.attname
where a.attrelid = relation
  and a.attname = any (field_names)
  and a.attnum > 0
  and not a.attisdropped;
$$;

/*
 * Given an array of argument names and a jsonb object in the shape of {[argument_name]:argument_value},
 * it returns a jsonb array with the argument values in the same order as the array inn field_names.
 * This can be used with postgrest_update_fields to build a SQL update command
 */
create or replace function _postgrest_update_ordered_field_values(field_names text[], passed_values jsonb)
    returns jsonb
    immutable
    language sql
as
$$
select jsonb_agg(passed_values ->> (name::text) order by idx)
from unnest(field_names) with ordinality as _ (name, idx);
$$;


create procedure postgrest_update(request omni_httpd.http_request, outcome inout omni_httpd.http_outcome,
                                  settings postgrest_settings default postgrest_settings())
    language plpgsql
as
$$
declare
    namespace text;
    query     text;
    result    jsonb;
    relation  regclass;
    preference text;
    _missing  text;
    _return   text := 'minimal';
    _tx       text;
    payload   jsonb;
begin
    if outcome is distinct from null then
        return;
    end if;
    if request.method = 'PATCH' then
        call omni_rest._postgrest_relation(request, relation, namespace, settings);
        if relation is null then
            return;
            -- terminate
        end if;
    else
        return;
        -- terminate;
    end if;

    -- change this when we implement other content types
    if omni_http.http_header_get(request.headers, 'content-type') <> 'application/json' then
        outcome := omni_httpd.http_response(status => 501, body => 'Only JSON content type is currently implemented');
        return;
    end if;

    -- while we support only json content type we can check its shape here
    payload := convert_from(request.body, 'utf8')::jsonb;
    if jsonb_typeof(payload) <> 'object' then
        outcome := omni_httpd.http_response(status => 422, body => 'Body must be object');
        return;
    end if;

    for preference in
        select regexp_split_to_table(omni_http.http_header_get(request.headers, 'prefer'), ',\s+')
        loop
            declare
                preference_name  text := split_part(preference, '=', 1);
                preference_value text := split_part(preference, '=', 2);
            begin
                case
                    when preference_name = 'missing' then _missing := preference_value;
                    when preference_name = 'return' then _return := preference_value;
                    when preference_name = 'tx' then _tx := preference_value;
                    end case;
            end;
        end loop;
    declare
        update_fields_definition text;
        field_values             jsonb;
        field_names              text[];
        params                   text[];
        _where                   text;
        _select       text;
        query_columns text[];
    begin
        select array_agg(jsonb_object_keys)
        from
            jsonb_object_keys(payload)
        into field_names;
        update_fields_definition :=
                coalesce(omni_rest._postgrest_update_fields(relation, field_names), '');
        field_values := omni_rest._postgrest_update_ordered_field_values(field_names, payload);
        params := omni_web.parse_query_string(request.query_string);
        _where := omni_rest.postgrest_format_get_param(omni_rest.postgrest_parse_get_param(params));
        -- Columns (vertical filtering)
        _select = omni_web.param_get(params, 'select');
        query_columns := omni_rest.postgrest_select_columns(namespace, relation, _select);

        query :=
                format('update %1$I.%2$I set %3$s where %4$s %5$s',
                       namespace,
                       (select relname
                        from pg_class
                        where oid = relation),
                       update_fields_definition,
                       coalesce(_where, 'true'),
                       case
                           when _return = 'representation' then
                               format('returning %1$s', concat_ws(', ', variadic query_columns))
                           else
                               ''
                           end
                );

        select jsonb_agg(stmt_row)
        from
            omni_sql.execute(query, coalesce(field_values, '[]'::jsonb))
        into result;
        if lower(_tx) = 'rollback' then
            rollback and chain;
        end if;
        outcome :=
                omni_httpd.http_response(status => 201, body => case when _return = 'representation' then result end);
    end;
end;
$$;



    create procedure postgrest_cors(request omni_httpd.http_request, response inout omni_httpd.http_outcome,
                                settings postgrest_settings default postgrest_settings())
    language plpgsql as
$$
declare
    origin text;
begin
    if request.method = 'OPTIONS' then
        origin := omni_http.http_header_get(request.headers, 'origin');
        response := omni_httpd.http_response(null::text,
                                             headers => array [omni_http.http_header('Access-Control-Allow-Origin', origin),
                                                 omni_http.http_header('Access-Control-Allow-Credentials', 'true'),
                                                 omni_http.http_header('Access-Control-Allow-Methods',
                                                                       'GET, POST, PATCH, PUT, DELETE, OPTIONS, HEAD'),
                                                 omni_http.http_header('Access-Control-Allow-Headers',
                                                                       'Authorization, Content-Type, Accept, Accept-Language, Content-Language'),
                                                 omni_http.http_header('Access-Control-Max-Age', '86400')
                                                 ]);
    end if;
end;
$$;
    create procedure postgrest(request omni_httpd.http_request, response inout omni_httpd.http_outcome,
                           settings postgrest_settings default postgrest_settings())
    language plpgsql
as
$$
declare
begin
    call omni_rest.postgrest_rpc(request, response, settings);
    call omni_rest.postgrest_get(request, response, settings);
    call omni_rest.postgrest_insert(request, response, settings);
    call omni_rest.postgrest_update(request, response, settings);
    call omni_rest.postgrest_cors(request, response, settings);
end;
$$;

end;
$instantiate$;
select instantiate(schema => '@extschema@');;
