create table languages
(
    id                       integer primary key generated always as identity,
    file_extension           text not null unique,
    language                 name not null,
    extension                name,
    file_processor           name,
    file_processor_extension name,
    priority                 integer not null default 0
);

insert
into
    languages (file_extension, language, extension, file_processor, file_processor_extension)
values
    ('sql', 'sql', null, null, null),
    ('pl', 'plperlu', 'plperlu', null, null),
    ('trusted.pl', 'plperl', 'plperl', null, null),
    ('py', 'plpython3u', 'plpython3u', 'omni_python.create_functions', 'omni_python'),
    ('tcl', 'pltclu', 'pltclu', null, null),
    ('trusted.tcl', 'pltcl', 'pltcl', null, null),
    ('rs', 'plrust', 'plrust', null, null);


select pg_catalog.pg_extension_config_dump('languages', '');;
create table auxiliary_tools
(
    id                  integer primary key generated always as identity,
    filename_stem       text,
    filename_extension  text,
    processor           name not null,
    processor_extension name,
    priority            integer not null default 0
);

insert
into
    auxiliary_tools (filename_stem, filename_extension, processor, processor_extension, priority)
values
    ('requirements', 'txt', 'omni_python.install_requirements', 'omni_python', 1);

select pg_catalog.pg_extension_config_dump('auxiliary_tools', '');
;
create table procs as
    select *
    from
        pg_proc
    limit 0;

create table policies as
    select *
    from
        pg_policy
    limit 0;

create table class as
    select *
    from
        pg_class
    limit 0;;
create function load_from_fs(fs anyelement, path text default '') returns setof text
    language plpgsql
as
$$
declare
    rec record;
begin
    -- Procs
    if not exists(select
                  from
                      pg_class
                  where
                      relname = '_omni_schema_pg_proc' and
                      relkind = 'r' and
                      relpersistence = 't') then
        create temporary table if not exists _omni_schema_pg_proc on commit drop as
            select * from pg_proc;
    end if;
    for rec in select * from omni_schema.procs
        loop
            execute format('drop function if exists %s', rec.oid::regprocedure);
        end loop;
    delete from omni_schema.procs;
    -- Policies
    if not exists(select
                  from
                      pg_class
                  where
                      relname = '_omni_schema_pg_policy' and
                      relkind = 'r' and
                      relpersistence = 't') then
        create temporary table if not exists _omni_schema_pg_policy on commit drop as
            select * from pg_policy;
    end if;
    for rec in select
                   policies.*,
                   pg_class.relname
               from
                   omni_schema.policies
                   inner join pg_class on pg_class.oid = policies.polrelid
        loop
            execute format('drop policy if exists %I on %I', rec.polname, rec.relname);
        end loop;
    delete from omni_schema.policies;
    -- Supported relations
    if not exists(select
                  from
                      pg_class
                  where
                      relname = '_omni_schema_pg_class' and
                      relkind = 'r' and
                      relpersistence = 't') then
        create temporary table if not exists _omni_schema_pg_class on commit drop as
            select * from pg_class;
    end if;
    for rec in select
                   class.*,
                   pg_namespace.nspname
               from
                   omni_schema.class
                   inner join pg_namespace on pg_namespace.oid = class.relnamespace
        loop
            if rec.relkind = 'v' then
                execute format('drop view if exists %I.%I cascade', rec.nspname, rec.relname);
            end if;
        end loop;
    delete from omni_schema.class;
    -- Execute
    for rec in select
                   case when path = '' then '' else path || '/' end || name as name,
                   omni_vfs.read(fs, path || '/' || name)                   as code,
                   language,
                   extension,
                   file_processor,
                   file_processor_extension,
                   processor,
                   processor_extension
               from
                   omni_vfs.list_recursively(fs, path, max => 10000) files
                   left join omni_schema.languages on files.name like concat('%', languages.file_extension)
                   left join omni_schema.auxiliary_tools
                             on files.name like concat(coalesce(auxiliary_tools.filename_stem, '%'), '.',
                                                       auxiliary_tools.filename_extension)
               where
                   auxiliary_tools.id is not null or
                   languages.id is not null
               order by coalesce(auxiliary_tools.priority, languages.priority) desc, name
        loop
            declare
                rec_code text;
            begin
                rec_code := convert_from(rec.code, 'utf-8');
                if rec.language is null and rec.processor is not null then
                    if not exists(select from pg_extension where extname = rec.processor_extension) then
                        raise notice 'Extension % required for auxiliary tool (required for %) is not installed', rec.processor_extension, rec.name;
                    else
                        execute format('select %s(%L::text)', rec.processor, rec_code);
                    end if;
                    continue;
                end if;
                if rec_code ~ 'omni_schema\[\[ignore\]\]' then
                    -- Ignore the file
                    continue;
                end if;
                if rec.language = 'sql' then
                    execute rec_code;
                else
                    if rec.language is null then
                        -- Ignore file
                        continue;
                    end if;
                    -- Check if the language is available
                    if not exists(select from pg_extension where extname = rec.extension) then
                        raise notice 'Extension % required for language % (required for %) is not installed', rec.extension, rec.language, rec.name;
                        -- Don't include it in the list of loaded files
                        continue;
                    else
                        -- Prepare and execute the SQL create function construct
                        declare
                            sql_snippet text;
                            regprocs    regprocedure[];
                        begin
                            if rec.file_processor is not null then
                                if (rec.file_processor_extension is not null and
                                    exists(select from pg_extension where extname = rec.file_processor_extension)) or
                                   rec.file_processor_extension is null then
                                    -- Can use the file processor
                                    execute format(
                                            'select array_agg(processor) from %s(%L::text, filename => %L::text, replace => true, fs => %L::%s) processor',
                                            rec.file_processor,
                                            rec_code, rec.name, fs::text, pg_typeof(fs)) into regprocs;
                                    if array_length(regprocs, 1) > 0 then
                                        return next rec.name;
                                        continue;
                                    end if;
                                end if;
                            end if;
                            if rec_code ~ 'SQL\[\[.*\]\]' then
                                sql_snippet := format('%s language %I as %L',
                                                      substring(rec_code from 'SQL\[\[(.*?)\]\]'), rec.language,
                                                      rec_code);
                                execute sql_snippet;
                            end if;
                        end;
                    end if;
                end if;
                return next rec.name;
            end;
        end loop;
    -- New procs
    insert
    into
        omni_schema.procs (select * from pg_proc except select * from _omni_schema_pg_proc);
    -- New policies
    insert
    into
        omni_schema.policies (select * from pg_policy except select * from _omni_schema_pg_policy);
    -- New supported relations
    insert
    into
        omni_schema.class (select * from pg_class except select * from _omni_schema_pg_class);
    return;
end ;
$$;;
create table migrations
(
    id         integer primary key generated always as identity,
    name       text      not null,
    migration  text      not null,
    applied_at timestamp not null default now()
);

select pg_catalog.pg_extension_config_dump('migrations', '');
select pg_catalog.pg_extension_config_dump('migrations_id_seq', '');;
create function migrate_from_fs(fs anyelement, path text default '') returns setof text
    language plpgsql
as
$$
declare
    rec record;
begin
    for rec in select
                   case when path = '' then '' else path || '/' end || files.name      as name,
                   convert_from(omni_vfs.read(fs, path || '/' || files.name), 'utf-8') as code
               from
                   omni_vfs.list_recursively(fs, path, max => 10000) as files
                   left join omni_schema.migrations
                             on migrations.name = (case when path = '' then '' else path || '/' end || files.name)
               where
                   files.name like '%.sql' and
                   migrations.name is null
               order by substring(files.name from '[0-9]+')::int asc
        loop
            execute rec.code;
            return next rec.name;
            insert
            into
                omni_schema.migrations (name, migration)
            values (rec.name, rec.code);
        end loop;
    return;
end;
$$;;
create function assemble_schema(conn_info text, fs anyelement, path text default '')
    returns table
            (
                execution_position  int,
                migration_filename  text,
                migration_statement text,
                execution_error     text
            )
    language plpgsql
as
$$
declare
    rec                             record;
    remote_fs                       omni_vfs.remote_fs;
    remote_table_fs_name            text    = 'omni_schema_migration_fs';
    current_filename                text;
    successful_statement_executions int     = 0;
    error_message                   text;
    error_detail text;
    try_again                       boolean = false;
begin
    -- statement execution status
    create temp table if not exists omni_schema_execution_status
    (
        id                       bigserial primary key,
        filepath                 text    not null,
        language                 text,
        extension                text,
        file_processor           text,
        file_processor_extension text,
        processor                text,
        processor_extension      text,
        code                     text    not null,
        -- order of successful execution
        execution_number         int,
        -- whether statement is executed successfully
        execution_successful     boolean not null default false,
        -- error for last execution of the statement
        last_execution_error     text
    ) on commit drop;
    truncate omni_schema_execution_status;

    <<top>>
    begin
        -- conn for setting up table_fs in remote database
        perform dblink_connect(remote_table_fs_name, conn_info);

        -- prepare necessary extensions in the remote database
        for rec in select distinct extension
                   from omni_schema.languages
                   union
                   select distinct file_processor_extension as extension
                   from omni_schema.languages
                   union
                   select distinct processor_extension as extension
                   from omni_schema.auxiliary_tools
            loop
                if exists (select from pg_extension where extname = rec.extension) then
                    perform dblink(remote_table_fs_name,
                                   format('create extension if not exists %s cascade', rec.extension));
                end if;
            end loop;

        -- Ensure omni_vfs is ready in the target
        perform dblink(remote_table_fs_name, 'create extension if not exists omni_vfs cascade');

        -- Remote table_fs filesystem in the target
        remote_fs := omni_vfs.remote_fs(conn_info, format('select omni_vfs.table_fs(%L)', remote_table_fs_name));

        -- read all files recognised by omni_schema.languages and omni_schema.auxiliary_tools
        for rec in
            select path || '/' || name                                           as name,
                   convert_from(omni_vfs.read(fs, path || '/' || name), 'UTF-8') as code,
                   language,
                   extension,
                   file_processor,
                   file_processor_extension,
                   processor,
                   processor_extension
            from omni_vfs.list_recursively(fs, path) files
                     left join omni_schema.languages on files.name like concat('%', languages.file_extension)
                     left join omni_schema.auxiliary_tools
                               on files.name like concat(coalesce(auxiliary_tools.filename_stem, '%'), '.',
                                                         auxiliary_tools.filename_extension)
            where (auxiliary_tools.id is not null or
                   languages.id is not null)
              and kind = 'file'
            order by coalesce(auxiliary_tools.priority, languages.priority) desc, name
            loop
                -- write file to the remote filesystem
                perform omni_vfs.write(remote_fs, rec.name, convert_to(rec.code, 'utf8'), create_file => true);

                if rec.code ~ 'omni_schema\[\[ignore\]\]' then
                    -- Ignore the file
                    continue;
                end if;
                if rec.language = 'sql' then
                    -- process sql language files statement by statement
                    declare
                        code text := convert_from(omni_vfs.read(fs, rec.name), 'UTF-8');
                        code_rec record;
                    begin
                        for code_rec in select source as current_statement, line, col
                                                 from
                                                     omni_sql.raw_statements(code::cstring)
                            loop
                                raise notice '%', json_build_object('type', 'info', 'message', 'Executing', 'code',
                                                                    code_rec.current_statement,
                                                                    'file', rec.name, 'line', code_rec.line, 'col',
                                                                    code_rec.col);
                                insert
                                into omni_schema_execution_status(filepath, language, code)
                                values (rec.name, rec.language, code_rec.current_statement);
                            end loop;
                    exception
                        when syntax_error then
                            raise notice '%', json_build_object('type', 'error', 'message', sqlerrm,
                                                                'file', rec.name);
                            insert
                            into omni_schema_execution_status(filepath, language, code, last_execution_error)
                            values (rec.name, rec.language, code, sqlerrm);

                            exit top;
                    end;
                else
                    -- process the whole file at once for other languages
                    insert
                    into omni_schema_execution_status(filepath, language, extension, file_processor,
                                                      file_processor_extension, processor, processor_extension, code)
                    values (rec.name, rec.language, rec.extension, rec.file_processor, rec.file_processor_extension,
                            rec.processor, rec.processor_extension, rec.code);
                end if;
            end loop;
        for rec in
            select distinct filepath from omni_schema_execution_status
            loop
                /*
                    create separate connections for each file to restrict any 'SET' statement
                    effects to the file which contains it
                */
                perform dblink_connect(rec.filepath, conn_info);
            end loop;
        -- execute until all statements are successfully executed or no more executions are successful
        while true
            loop
                <<file>>
                for current_filename in select distinct filepath from omni_schema_execution_status order by filepath
                    loop
                        <<statement>>
                        for rec in select *
                                   from omni_schema_execution_status
                                   where execution_successful = false
                                     and filepath = current_filename
                                   order by id
                            loop
                                declare
                                    _filepath text;
                                    _code     text;
                                begin
                                    if rec.language is null and rec.processor is not null then
                                        if not exists(select *
                                                      from dblink(rec.filepath,
                                                                  format(
                                                                          'select true from pg_extension where extname = %L',
                                                                          rec.processor_extension)) as t(b boolean)) then
                                            raise exception 'Extension % required for auxiliary tool (required for %) is not installed', rec.processor_extension, rec.filepath;
                                        else
                                            perform *
                                            from dblink(rec.filepath,
                                                        format('with cte as (%s) select true from cte',
                                                               format('select %s(%L::text)', rec.processor, rec.code))) as t(b boolean);
                                        end if;
                                    elsif rec.language = 'sql' then
                                        -- normalize sql statement like removing whitespace
                                        rec.code = omni_sql.statement_in(rec.code::cstring);
                                        if omni_sql.statement_type(rec.code::omni_sql.statement) = 'SelectStmt' then
                                            /*
                                                SELECT statement result set can be of any shape(number of columns, their types)
                                                so wrap select statements in cte to give it consistent shape and avoid the
                                                ERROR:  function returning record called in context that cannot accept type record
                                            */
                                            perform *
                                            from dblink(rec.filepath,
                                                        format('with cte as (%s) select true from cte', rec.code)) as t(b boolean);
                                        else
                                            perform dblink(rec.filepath, rec.code);
                                        end if;
                                    else
                                        if not exists(select *
                                                      from dblink(rec.filepath,
                                                                  format(
                                                                          'select true from pg_extension where extname = %L',
                                                                          rec.extension)) as t(b boolean)) then
                                            raise exception 'Extension % required for language % (required for %) is not installed', rec.extension, rec.language, rec.filepath;
                                        else
                                            -- Prepare and execute the SQL create function construct
                                            if rec.file_processor is not null then
                                                if (rec.file_processor_extension is not null and
                                                    exists(select *
                                                           from dblink(rec.filepath,
                                                                       format(
                                                                               'select true from pg_extension where extname = %L',
                                                                               rec.file_processor_extension)) as t(b boolean))) or
                                                   rec.file_processor_extension is null then
                                                    -- Can use the file processor
                                                    perform *
                                                    from dblink(rec.filepath, format(
                                                            'with cte as (select %s(%L::text, filename => %L::text, replace => true, fs => omni_vfs.table_fs(%L))) select true from cte',
                                                            rec.file_processor,
                                                            rec.code, rec.filepath, remote_table_fs_name)) as t(b boolean);
                                                end if;
                                            end if;
                                            if rec.code ~ 'SQL\[\[.*\]\]' then
                                                perform dblink(rec.filepath, format('%s language %I as %L',
                                                                                    substring(rec.code from 'SQL\[\[(.*?)\]\]'),
                                                                                    rec.language,
                                                                                    rec.code));
                                            end if;
                                        end if;
                                    end if;
                                    -- update statement execution status
                                    successful_statement_executions = successful_statement_executions + 1;
                                    update omni_schema_execution_status
                                    set execution_number     = successful_statement_executions,
                                        execution_successful = true,
                                        last_execution_error = null
                                    where id = rec.id
                                    returning filepath, code into _filepath, _code;
                                    raise notice '%', json_build_object('type', 'info', 'message',
                                                                        'Completed', 'code', _code,
                                                                        'file', _filepath);
                                    -- at least one succeeded, worth trying again
                                    try_again = true;
                                exception
                                    when others then
                                        get stacked diagnostics error_message = message_text, error_detail = pg_exception_detail;
                                        update omni_schema_execution_status
                                        set last_execution_error = error_message
                                        where id = rec.id
                                        returning filepath, code into _filepath, _code;
                                        raise notice '%', json_build_object('type', 'error', 'message',
                                                                            sqlerrm, 'detail', error_detail, 'code',
                                                                            _code,
                                                                            'file', _filepath);
                                        -- go to next file if statement execution fails to preserve serial execution of statements in a file
                                        continue file;
                                end;
                            end loop statement;
                    end loop file;
                if try_again = true then
                    try_again = false;
                else
                    exit;
                end if;
            end loop;

        for rec in
            select distinct filepath from omni_schema_execution_status
            loop
                perform dblink_disconnect(rec.filepath);
            end loop;
    end;
    -- close the connections set up at the beginning
    perform dblink_disconnect(remote_table_fs_name);

    return query select execution_number,
                        filepath,
                        code,
                        last_execution_error
                 from omni_schema_execution_status
                 order by execution_number;
end;
$$;;



create function instantiate_meta(schema name) returns void
    language plpgsql
as
$instantiate_meta$
declare
    rec record;
begin
    perform set_config('search_path', schema::text, true);

    /*
This code is partially based on meta from the Aquameta project (https://github.com/aquameta/meta),
licensed under the terms of BSD-2 Clause License. It started off as an integration but became a
significant re-working.
*/
create type cast_id as (source_type_schema_name name,source_type_name name,target_type_schema_name name,target_type_name name);
create function cast_id(source_type_schema_name name,source_type_name name,target_type_schema_name name,target_type_name name) returns cast_id immutable return row(source_type_schema_name,source_type_name,target_type_schema_name,target_type_name)::cast_id;

create type column_id as (schema_name name,relation_name name,name name);
create function column_id(schema_name name,relation_name name,name name) returns column_id immutable return row(schema_name,relation_name,name)::column_id;

create type constraint_id as (schema_name name,relation_name name,name name);
create function constraint_id(schema_name name,relation_name name,name name) returns constraint_id immutable return row(schema_name,relation_name,name)::constraint_id;

create type constraint_check_id as (schema_name name,table_name name,name name,column_names name);
create function constraint_check_id(schema_name name,table_name name,name name,column_names name) returns constraint_check_id immutable return row(schema_name,table_name,name,column_names)::constraint_check_id;

create type constraint_unique_id as (schema_name name,table_name name,name name,column_names name);
create function constraint_unique_id(schema_name name,table_name name,name name,column_names name) returns constraint_unique_id immutable return row(schema_name,table_name,name,column_names)::constraint_unique_id;

create type extension_id as (name name);
create function extension_id(name name) returns extension_id immutable return row(name)::extension_id;

create type field_id as (schema_name name,relation_name name,pk_column_names name[],pk_values text[],column_name name);
create function field_id(schema_name name,relation_name name,pk_column_names name[],pk_values text[],column_name name) returns field_id immutable return row(schema_name,relation_name,pk_column_names,pk_values,column_name)::field_id;

create type foreign_column_id as (schema_name name,name name);
create function foreign_column_id(schema_name name,name name) returns foreign_column_id immutable return row(schema_name,name)::foreign_column_id;

create type foreign_data_wrapper_id as (name name);
create function foreign_data_wrapper_id(name name) returns foreign_data_wrapper_id immutable return row(name)::foreign_data_wrapper_id;

create type foreign_key_id as (schema_name name,relation_name name,name name);
create function foreign_key_id(schema_name name,relation_name name,name name) returns foreign_key_id immutable return row(schema_name,relation_name,name)::foreign_key_id;

create type foreign_server_id as (name name);
create function foreign_server_id(name name) returns foreign_server_id immutable return row(name)::foreign_server_id;

create type foreign_table_id as (schema_name name,name name);
create function foreign_table_id(schema_name name,name name) returns foreign_table_id immutable return row(schema_name,name)::foreign_table_id;

create type function_id as (schema_name name,name name,parameters name[]);
create function function_id(schema_name name,name name,parameters name[]) returns function_id immutable return row(schema_name,name,parameters)::function_id;

create type procedure_id as (schema_name name,name name,parameters name[]);
create function procedure_id(schema_name name,name name,parameters name[]) returns procedure_id immutable return row(schema_name,name,parameters)::procedure_id;

create type operator_id as (schema_name name,name name,left_arg_type_schema_name name,left_arg_type_name name,right_arg_type_schema_name name,right_arg_type_name name);
create function operator_id(schema_name name,name name,left_arg_type_schema_name name,left_arg_type_name name,right_arg_type_schema_name name,right_arg_type_name name) returns operator_id immutable return row(schema_name,name,left_arg_type_schema_name,left_arg_type_name,right_arg_type_schema_name,right_arg_type_name)::operator_id;

create type policy_id as (schema_name name,relation_name name,name name);
create function policy_id(schema_name name,relation_name name,name name) returns policy_id immutable return row(schema_name,relation_name,name)::policy_id;

create type relation_id as (schema_name name,name name);
create function relation_id(schema_name name,name name) returns relation_id immutable return row(schema_name,name)::relation_id;

create type role_id as (name name);
create function role_id(name name) returns role_id immutable return row(name)::role_id;

create type role_setting_id as (role_name name, database_name name, setting_name name);
create function role_setting_id(role_name name, database_name name, setting_name name) returns role_setting_id immutable return row(role_name,database_name,setting_name)::role_setting_id;

create type row_id as (schema_name name,relation_name name,pk_column_names name[],pk_values text[]);
create function row_id(schema_name name,relation_name name,pk_column_names name[],pk_values text[]) returns row_id immutable return row(schema_name,relation_name,pk_column_names,pk_values)::row_id;

create type schema_id as (name name);
create function schema_id(name name) returns schema_id immutable return row(name)::schema_id;

create type sequence_id as (schema_name name,name name);
create function sequence_id(schema_name name,name name) returns sequence_id immutable return row(schema_name,name)::sequence_id;

create type table_id as (schema_name name,name name);
create function table_id(schema_name name,name name) returns table_id immutable return row(schema_name,name)::table_id;

create type table_privilege_id as (schema_name name,relation_name name,role name,type name);
create function table_privilege_id(schema_name name,relation_name name,role name,type name) returns table_privilege_id immutable return row(schema_name,relation_name,role,type)::table_privilege_id;

create type trigger_id as (schema_name name,relation_name name,name name);
create function trigger_id(schema_name name,relation_name name,name name) returns trigger_id immutable return row(schema_name,relation_name,name)::trigger_id;

create type type_id as (schema_name name,name name);
create function type_id(schema_name name,name name) returns type_id immutable return row(schema_name,name)::type_id;

create type view_id as (schema_name name,name name);
create function view_id(schema_name name,name name) returns view_id immutable return row(schema_name,name)::view_id;

create type index_id as
(
    schema_name name,
    name        name
);
create function index_id(schema_name name, name name) returns index_id immutable return row(schema_name,name)::index_id;

create type language_id as (name name);
create function language_id(name name) returns language_id immutable return row(name)::language_id;

create function column_id_to_schema_id(column_id column_id) returns schema_id immutable return schema_id((column_id).schema_name);
create cast (column_id as schema_id) with function column_id_to_schema_id(column_id) as assignment;

create function constraint_id_to_schema_id(constraint_id constraint_id) returns schema_id immutable return schema_id((constraint_id).schema_name);
create cast (constraint_id as schema_id) with function constraint_id_to_schema_id(constraint_id) as assignment;

create function constraint_check_id_to_schema_id(constraint_check_id constraint_check_id) returns schema_id immutable return schema_id((constraint_check_id).schema_name);
create cast (constraint_check_id as schema_id) with function constraint_check_id_to_schema_id(constraint_check_id) as assignment;

create function constraint_unique_id_to_schema_id(constraint_unique_id constraint_unique_id) returns schema_id immutable return schema_id((constraint_unique_id).schema_name);
create cast (constraint_unique_id as schema_id) with function constraint_unique_id_to_schema_id(constraint_unique_id) as assignment;

create function field_id_to_schema_id(field_id field_id) returns schema_id immutable return schema_id((field_id).schema_name);
create cast (field_id as schema_id) with function field_id_to_schema_id(field_id) as assignment;

create function foreign_column_id_to_schema_id(foreign_column_id foreign_column_id) returns schema_id immutable return schema_id((foreign_column_id).schema_name);
create cast (foreign_column_id as schema_id) with function foreign_column_id_to_schema_id(foreign_column_id) as assignment;

create function foreign_key_id_to_schema_id(foreign_key_id foreign_key_id) returns schema_id immutable return schema_id((foreign_key_id).schema_name);
create cast (foreign_key_id as schema_id) with function foreign_key_id_to_schema_id(foreign_key_id) as assignment;

create function foreign_table_id_to_schema_id(foreign_table_id foreign_table_id) returns schema_id immutable return schema_id((foreign_table_id).schema_name);
create cast (foreign_table_id as schema_id) with function foreign_table_id_to_schema_id(foreign_table_id) as assignment;

create function function_id_to_schema_id(function_id function_id) returns schema_id immutable return schema_id((function_id).schema_name);
create cast (function_id as schema_id) with function function_id_to_schema_id(function_id) as assignment;

create function procedure_id_to_schema_id(procedure_id procedure_id) returns schema_id immutable return schema_id((procedure_id).schema_name);
create cast (procedure_id as schema_id) with function procedure_id_to_schema_id(procedure_id) as assignment;

create function operator_id_to_schema_id(operator_id operator_id) returns schema_id immutable return schema_id((operator_id).schema_name);
create cast (operator_id as schema_id) with function operator_id_to_schema_id(operator_id) as assignment;

create function policy_id_to_schema_id(policy_id policy_id) returns schema_id immutable return schema_id((policy_id).schema_name);
create cast (policy_id as schema_id) with function policy_id_to_schema_id(policy_id) as assignment;

create function relation_id_to_schema_id(relation_id relation_id) returns schema_id immutable return schema_id((relation_id).schema_name);
create cast (relation_id as schema_id) with function relation_id_to_schema_id(relation_id) as assignment;

create function row_id_to_schema_id(row_id row_id) returns schema_id immutable return schema_id((row_id).schema_name);
create cast (row_id as schema_id) with function row_id_to_schema_id(row_id) as assignment;

create function sequence_id_to_schema_id(sequence_id sequence_id) returns schema_id immutable return schema_id((sequence_id).schema_name);
create cast (sequence_id as schema_id) with function sequence_id_to_schema_id(sequence_id) as assignment;

create function table_id_to_schema_id(table_id table_id) returns schema_id immutable return schema_id((table_id).schema_name);
create cast (table_id as schema_id) with function table_id_to_schema_id(table_id) as assignment;

create function table_privilege_id_to_schema_id(table_privilege_id table_privilege_id) returns schema_id immutable return schema_id((table_privilege_id).schema_name);
create cast (table_privilege_id as schema_id) with function table_privilege_id_to_schema_id(table_privilege_id) as assignment;

create function trigger_id_to_schema_id(trigger_id trigger_id) returns schema_id immutable return schema_id((trigger_id).schema_name);
create cast (trigger_id as schema_id) with function trigger_id_to_schema_id(trigger_id) as assignment;

create function type_id_to_schema_id(type_id type_id) returns schema_id immutable return schema_id((type_id).schema_name);
create cast (type_id as schema_id) with function type_id_to_schema_id(type_id) as assignment;

create function view_id_to_schema_id(view_id view_id) returns schema_id immutable return schema_id((view_id).schema_name);
create cast (view_id as schema_id) with function view_id_to_schema_id(view_id) as assignment;

create function column_id_to_relation_id(column_id column_id) returns relation_id immutable return relation_id((column_id).schema_name, (column_id).relation_name);
create cast (column_id as relation_id) with function column_id_to_relation_id(column_id) as assignment;

create function constraint_id_to_relation_id(constraint_id constraint_id) returns relation_id immutable return relation_id((constraint_id).schema_name, (constraint_id).relation_name);
create cast (constraint_id as relation_id) with function constraint_id_to_relation_id(constraint_id) as assignment;

create function field_id_to_relation_id(field_id field_id) returns relation_id immutable return relation_id((field_id).schema_name, (field_id).relation_name);
create cast (field_id as relation_id) with function field_id_to_relation_id(field_id) as assignment;

create function foreign_key_id_to_relation_id(foreign_key_id foreign_key_id) returns relation_id immutable return relation_id((foreign_key_id).schema_name, (foreign_key_id).relation_name);
create cast (foreign_key_id as relation_id) with function foreign_key_id_to_relation_id(foreign_key_id) as assignment;

create function policy_id_to_relation_id(policy_id policy_id) returns relation_id immutable return relation_id((policy_id).schema_name, (policy_id).relation_name);
create cast (policy_id as relation_id) with function policy_id_to_relation_id(policy_id) as assignment;

create function row_id_to_relation_id(row_id row_id) returns relation_id immutable return relation_id((row_id).schema_name, (row_id).relation_name);
create cast (row_id as relation_id) with function row_id_to_relation_id(row_id) as assignment;

create function table_privilege_id_to_relation_id(table_privilege_id table_privilege_id) returns relation_id immutable return relation_id((table_privilege_id).schema_name, (table_privilege_id).relation_name);
create cast (table_privilege_id as relation_id) with function table_privilege_id_to_relation_id(table_privilege_id) as assignment;

create function trigger_id_to_relation_id(trigger_id trigger_id) returns relation_id immutable return relation_id((trigger_id).schema_name, (trigger_id).relation_name);
create cast (trigger_id as relation_id) with function trigger_id_to_relation_id(trigger_id) as assignment;

create function field_id_to_column_id(field_id field_id) returns column_id immutable return column_id((field_id).schema_name, (field_id).relation_name, (field_id).column_name);
create cast (field_id as column_id) with function field_id_to_column_id(field_id) as assignment;

perform omni_types.sum_type('object_id', variadic array_agg(t.oid::regtype))
from pg_type t
    inner join pg_namespace ns on ns.oid = t.typnamespace and ns.nspname = current_schema
where t.typname like '%_id';

    /*
This code is partially based on meta from the Aquameta project (https://github.com/aquameta/meta),
licensed under the terms of BSD-2 Clause License. It started off as an integration but became a
significant re-working.
*/

-- for some reason, generator isn't generating these.  they're commented out.
create function field_id_to_row_id(field_id field_id) returns row_id as $_$ select row_id((field_id).schema_name, (field_id).relation_name, (field_id).pk_column_names, (field_id).pk_values) $_$ immutable language sql;
create cast (field_id as row_id) with function field_id_to_row_id(field_id) as assignment;


------------------------------------------------------------------------
-- helpers
-- non-standard constructors that do sensible things
------------------------------------------------------------------------

-- field_id constructor taking a row_id
create function field_id(row_id row_id, column_name name) returns field_id as $$
    select field_id(row_id.schema_name, row_id.relation_name, row_id.pk_column_names, row_id.pk_values, column_name);
$$ language sql;


-- field_id constructor non-array pk
create function field_id( schema_name name, relation_name name, pk_column_name name, pk_value name, column_name name) returns field_id as $$
    select field_id(schema_name, relation_name, array[pk_column_name], array[pk_value], column_name);
$$ language sql;


-- single key row_id constructor
create function row_id(schema_name text, relation_name name, pk_column_name name, pk_value name) returns row_id as $_$ select row_id(schema_name, relation_name, array[pk_column_name], array[pk_value]) $_$ immutable language sql;


------------------------------------------------------------------------
-- pk_stmt()
-- helper function for iterating primary key arrays and generating a stmt fragment.
------------------------------------------------------------------------

-- template is rendered by format(), using positional argument notation.
--     1: pk_column_names[i]
--     2: pk_values[i]
--     3: i
--
-- select _pk_stmt (
--     array['id','other_id'],
--     array[public.uuid_generate_v4()::text,public.uuid_generate_v4()::text],
--     '(row_id).pk_values[%3$s] = x.%1$I'
-- );
--                               _pk_stmt                               
-- ---------------------------------------------------------------------
--  (row_id).pk_values[1] = x.id and (row_id).pk_values[2] = x.other_id

create or replace function _pk_stmt(pk_column_names text[], pk_values text[], template text, delimeter text default ' and ') returns text as $$
    declare
        pk_comparisons text[];
        column_name text;
        i integer;
    begin
        i := 1;
        foreach column_name in array pk_column_names loop
            pk_comparisons[i] := format(template, pk_column_names[i], pk_values[i], i);
            i := i + 1;
        end loop;
        return array_to_string(pk_comparisons, delimeter);
    end
$$ language plpgsql;

/*
select _pk_stmt (
    row_id(
        'public',
        'foo',
        array['id','other_id'],
        array[public.uuid_generate_v4()::text,public.uuid_generate_v4()::text]
    ),
    '(row_id).pk_values[%3$s] = x.%1$I',
    ' OR '
);
                              _pk_stmt                              
--------------------------------------------------------------------
 (row_id).pk_values[1] = x.id OR (row_id).pk_values[2] = x.other_id
*/

create function _pk_stmt(row_id row_id, template text, delimeter text default ' and ') returns text as $$
    select _pk_stmt((row_id).pk_column_names, (row_id).pk_values, template, delimeter);
$$ language sql;


create or replace function field_id_literal_value(field_id field_id) returns text as $$
declare
    literal_value text;
    stmt text;
    pk_stmt text;
begin
    pk_stmt := _pk_stmt (
        (field_id).pk_column_names,
        (field_id).pk_values,
        '%1$I = %2$L'
    );

    stmt := format('select %I::text from %I.%I where %s',
        (field_id).column_name,
        (field_id).schema_name,
        (field_id).relation_name,
        pk_stmt
    );
    -- raise notice 'stmt: %', stmt;

    execute stmt into literal_value;

    return literal_value;
-- TODO: is this correct?  this fires when the table doesn't exist etc.
exception when others then
    raise warning 'field_id_literal_value exception on %: %', field_id, SQLERRM;
    return null;
end
$$ language plpgsql stable;



create or replace function row_exists(in row_id row_id, out answer boolean) as $$
    declare
        stmt text;
        pk_comparisons text[];
        pk_comparison_stmt text;
        column_name text;
        i integer;
    begin
        -- generate the pk comparisons line
        i := 1;
        foreach column_name in array row_id.pk_column_names loop
            pk_comparisons[i] := quote_ident((row_id).pk_column_names[i]) || '::text = ' || quote_literal((row_id).pk_values[i]);
            i := i + 1;
        end loop;
        pk_comparison_stmt := array_to_string(pk_comparisons, ' and ');


        stmt := format (
            -- 'select (count(*) = 1) from %I.%I where %I::text = %L',
            'select (count(*) = 1) from %I.%I where %s',
                (row_id).schema_name,
                (row_id).relation_name,
                pk_comparison_stmt
            );

        raise debug '%', stmt;
        execute stmt into answer;

    exception
        when undefined_table then
            answer := false;
    end;
$$ language plpgsql;

    /*
This code is partially based on meta from the Aquameta project (https://github.com/aquameta/meta),
licensed under the terms of BSD-2 Clause License. It started off as an integration but became a
significant re-working.
*/

-- returns an array of types, given an identity_args string (as provided by pg_get_function_identity_arguments())
create or replace function _get_function_type_sig_array(p pg_proc) returns text[]
    language sql as
$$
select
    coalesce(array_agg(format_type(typ, null) order by ordinality), '{}')
from
    (select
         typ,
         ordinality
     from
         unnest(p.proargtypes) with ordinality as t(typ, ordinality)) as arg(typ, ordinality)
$$ set search_path = '', stable;

create function resolved_type_name(t pg_type) returns name
stable
return case
           when t.typelem != 0 and t.typlen = -1
               then (select
                         typname || '[]'
                     from
                         pg_type bt
                     where
                         bt.oid = t.typelem)
           else t.typname
    end;

create function _pg_get_expr(pg_node_tree, oid) returns text
 set search_path = ''
stable
 return pg_get_expr($1, $2);

create function _pg_get_expr(pg_node_tree, oid, boolean) returns text
 set search_path = ''
stable
 return pg_get_expr($1, $2, $3);


/******************************************************************************
 * siuda
 *****************************************************************************/
create type siuda as enum ('select', 'insert', 'update', 'delete', 'all');

create function siuda(c char) returns siuda as $$
begin
    case c
        when 'r' then
            return 'select'::siuda;
        when 'a' then
            return 'insert'::siuda;
        when 'w' then
            return 'update'::siuda;
        when 'd' then
            return 'delete'::siuda;
        when '*' then
            return 'all'::siuda;
    end case;
end;
$$ immutable language plpgsql;


create cast (char as siuda)
with function siuda(char)
as assignment;


/******************************************************************************
 * schema
 *****************************************************************************/
create view schema as
    select schema_id(schema_name) id, schema_name::text as name
    from information_schema.schemata
--    where schema_name not in ('pg_catalog', 'information_schema')
;


/******************************************************************************
 * type
 *****************************************************************************/

create or replace view type as
    select
        type_id(n.nspname, resolved_type_name(t)) as id,
        n.nspname::text                           as schema_name,
        case
            when t.typelem != 0 and t.typlen = -1
                then (select typname || '[]' from pg_type bt where bt.oid = t.typelem)
            else t.typname
            end                                   as name
    from
        pg_catalog.pg_type                t
        left join pg_catalog.pg_namespace n on n.oid = t.typnamespace;


create or replace view type_basic as
    select
        type_id(n.nspname, resolved_type_name(t)) as id
    from
        pg_catalog.pg_type                t
        left join pg_catalog.pg_namespace n on n.oid = t.typnamespace
    where
        typtype = 'b';

create or replace view type_array as
    select
        type_id(n.nspname, resolved_type_name(t))   as id,
        type_id(n1.nspname, resolved_type_name(t1)) as base_type_id
    from
        pg_catalog.pg_type      t
        inner join pg_type      t1 on t1.oid = t.typelem
        inner join pg_namespace n on n.oid = t.typnamespace
        inner join pg_namespace n1 on n1.oid = t1.typnamespace
    where
        t.typelem != 0;

create or replace view type_composite as
    select
        type_id(n.nspname, resolved_type_name(t)) as id
    from
        pg_catalog.pg_type                t
        left join pg_catalog.pg_namespace n on n.oid = t.typnamespace
    where
        typtype = 'c';

create or replace view type_composite_attribute as
    select
        type_id(n.nspname, resolved_type_name(t)) as id,
        a.attname                                 as name
    from
        pg_catalog.pg_type                t
        left join pg_catalog.pg_namespace n on n.oid = t.typnamespace
        join      pg_attribute            a on a.attrelid = t.typrelid
    where
        t.typtype = 'c' and
        a.attnum > 0 and
        not a.attisdropped;

create or replace view type_composite_attribute_position as
    select
        type_id(n.nspname, resolved_type_name(t)) as id,
        a.attname                                 as name,
        a.attnum                                  as position
    from
        pg_catalog.pg_type                t
        left join pg_catalog.pg_namespace n on n.oid = t.typnamespace
        join      pg_attribute            a on a.attrelid = t.typrelid
    where
        t.typtype = 'c' and
        a.attnum > 0 and
        not a.attisdropped;

create or replace view type_composite_attribute_collation as
    select
        type_id(n.nspname, resolved_type_name(t)) as id,
        a.attname                                 as name,
        co.collname                               as collation
    from
        pg_catalog.pg_type                t
        left join pg_catalog.pg_namespace n on n.oid = t.typnamespace
        join      pg_attribute            a on a.attrelid = t.typrelid
        join      pg_type                 ct on ct.oid = a.atttypid
        left join pg_collation            co on co.oid = a.attcollation
    where
        t.typtype = 'c' and
        a.attnum > 0 and
        not a.attisdropped and
        a.attcollation <> ct.typcollation;

create or replace view type_domain as
    select
        type_id(n.nspname, resolved_type_name(t))   as id,
        type_id(n1.nspname, resolved_type_name(t1)) as base_type_id
    from
        pg_catalog.pg_type                 t
        left join  pg_catalog.pg_namespace n on n.oid = t.typnamespace
        inner join pg_catalog.pg_type      t1 on t1.oid = t.typbasetype
        left join  pg_catalog.pg_namespace n1 on n1.oid = t1.typnamespace
    where
        t.typtype = 'd';

create or replace view type_enum as
    select
        type_id(n.nspname, resolved_type_name(t)) as id
    from
        pg_catalog.pg_type                t
        left join pg_catalog.pg_namespace n on n.oid = t.typnamespace
    where
        typtype = 'e';

create or replace view type_enum_label as
    select
        type_id(n.nspname, resolved_type_name(t)) as id,
        enumlabel                                 as label,
        enumsortorder                             as sortorder
    from
        pg_catalog.pg_type                 t
        left join  pg_catalog.pg_namespace n on n.oid = t.typnamespace
        inner join pg_enum                 e on e.enumtypid = t.oid
    where
        typtype = 'e';

create or replace view type_pseudo as
    select
        type_id(n.nspname, resolved_type_name(t)) as id
    from
        pg_catalog.pg_type                t
        left join pg_catalog.pg_namespace n on n.oid = t.typnamespace
    where
        typtype = 'p';

create or replace view type_range as
    select
        type_id(n.nspname, resolved_type_name(t)) as id
    from
        pg_catalog.pg_type                t
        left join pg_catalog.pg_namespace n on n.oid = t.typnamespace
    where
        typtype = 'r';

create or replace view type_multirange as
    select
        type_id(n.nspname, resolved_type_name(t)) as id
    from
        pg_catalog.pg_type                t
        left join pg_catalog.pg_namespace n on n.oid = t.typnamespace
    where
        typtype = 'm';



/******************************************************************************
 * cast
 *****************************************************************************/
create view "cast" as
    select
        cast_id(ts.typname, pg_catalog.format_type(castsource, null), tt.typname,
                pg_catalog.format_type(casttarget, null)) as id,
        pg_catalog.format_type(castsource, null)          as "from",
        pg_catalog.format_type(casttarget, null)          as "to"
    from
        pg_catalog.pg_cast                c
        left join pg_catalog.pg_type      ts
                  on c.castsource = ts.oid
        left join pg_catalog.pg_namespace ns
                  on ns.oid = ts.typnamespace
        left join pg_catalog.pg_type      tt
                  on c.casttarget = tt.oid;

create view cast_binary_coercible as
    select
        cast_id(ts.typname, pg_catalog.format_type(castsource, null), tt.typname,
                pg_catalog.format_type(casttarget, null)) as id
    from
        pg_catalog.pg_cast                c
        inner join pg_catalog.pg_type      ts
                  on c.castsource = ts.oid
        inner join pg_catalog.pg_namespace ns
                  on ns.oid = ts.typnamespace
        inner join pg_catalog.pg_type      tt
                  on c.casttarget = tt.oid
    where
        castfunc = 0;

create view cast_function as
    select
        cast_id(ts.typname, pg_catalog.format_type(castsource, null), tt.typname,
                pg_catalog.format_type(casttarget, null)) as id,
        function_id(
            pns.nspname,
            p.proname,
            _get_function_type_sig_array(p)
        )
    from
        pg_catalog.pg_cast                 c
        inner join pg_catalog.pg_proc      p
                   on c.castfunc = p.oid
        inner join pg_catalog.pg_namespace pns
                   on pns.oid = p.pronamespace
        inner join  pg_catalog.pg_type      ts
                   on c.castsource = ts.oid
        inner join  pg_catalog.pg_namespace ns
                   on ns.oid = ts.typnamespace
        inner join  pg_catalog.pg_type      tt
                   on c.casttarget = tt.oid
where castfunc != 0;

create view cast_implicit_in_assignment as
    select
        cast_id(ts.typname, pg_catalog.format_type(castsource, null), tt.typname,
                pg_catalog.format_type(casttarget, null)) as id
    from
        pg_catalog.pg_cast                 c
        inner join  pg_catalog.pg_type      ts
                    on c.castsource = ts.oid
        inner join  pg_catalog.pg_namespace ns
                    on ns.oid = ts.typnamespace
        inner join  pg_catalog.pg_type      tt
                    on c.casttarget = tt.oid
where castcontext = 'a';

create view cast_explicit as
    select
        cast_id(ts.typname, pg_catalog.format_type(castsource, null), tt.typname,
                pg_catalog.format_type(casttarget, null)) as id
    from
        pg_catalog.pg_cast                 c
        inner join  pg_catalog.pg_type      ts
                    on c.castsource = ts.oid
        inner join  pg_catalog.pg_namespace ns
                    on ns.oid = ts.typnamespace
        inner join  pg_catalog.pg_type      tt
                    on c.casttarget = tt.oid
    where castcontext = 'e';

create view cast_implicit as
    select
        cast_id(ts.typname, pg_catalog.format_type(castsource, null), tt.typname,
                pg_catalog.format_type(casttarget, null)) as id
    from
        pg_catalog.pg_cast                 c
        inner join  pg_catalog.pg_type      ts
                    on c.castsource = ts.oid
        inner join  pg_catalog.pg_namespace ns
                    on ns.oid = ts.typnamespace
        inner join  pg_catalog.pg_type      tt
                    on c.casttarget = tt.oid
    where castcontext = 'i';

/******************************************************************************
 * operator
 *****************************************************************************/
create view "operator" as
SELECT operator_id(n.nspname, o.oprname, trns.nspname, tr.typname, trns.nspname, tr.typname) as id,
    n.nspname::text as schema_name,
    o.oprname::text as name,
    CASE WHEN o.oprkind='l' THEN NULL ELSE pg_catalog.format_type(o.oprleft, NULL) END AS "Left arg type",
    CASE WHEN o.oprkind='r' THEN NULL ELSE pg_catalog.format_type(o.oprright, NULL) END AS "Right arg type",
    pg_catalog.format_type(o.oprresult, NULL) AS "Result type",
    coalesce(pg_catalog.obj_description(o.oid, 'pg_operator'),
        pg_catalog.obj_description(o.oprcode, 'pg_proc')) AS "Description"
FROM pg_catalog.pg_operator o
    LEFT JOIN pg_catalog.pg_namespace n ON n.oid = o.oprnamespace
    JOIN pg_catalog.pg_type tl ON o.oprleft = tl.oid
    JOIN pg_catalog.pg_namespace tlns on tl.typnamespace = tlns.oid
    JOIN pg_catalog.pg_type tr ON o.oprleft = tr.oid
    JOIN pg_catalog.pg_namespace trns on tr.typnamespace = trns.oid
WHERE n.nspname <> 'pg_catalog'
    AND n.nspname <> 'information_schema'
--    AND pg_catalog.pg_operator_is_visible(o.oid)
-- ORDER BY 1, 2, 3, 4;
;


/******************************************************************************
 * sequence
 *****************************************************************************/
create view "sequence" as
    select sequence_id(ns.nspname, c.relname) as id,
           schema_id(ns.nspname) as schema_id,
           ns.nspname as schema_name,
           c.relname::text as name
    from pg_sequence s
inner join pg_class c on c.oid = s.seqrelid
inner join pg_namespace ns on ns.oid = c.relnamespace;

create view "sequence_start_value" as
    select sequence_id(ns.nspname, c.relname) as id,
           schema_id(ns.nspname) as schema_id,
           s.seqstart as start_value
    from pg_sequence s
         inner join pg_class c on c.oid = s.seqrelid
         inner join pg_namespace ns on ns.oid = c.relnamespace;

create view "sequence_minimum_value" as
    select sequence_id(ns.nspname, c.relname) as id,
           schema_id(ns.nspname) as schema_id,
           s.seqmin as minimum_value
    from pg_sequence s
         inner join pg_class c on c.oid = s.seqrelid
         inner join pg_namespace ns on ns.oid = c.relnamespace;

create view "sequence_maximum_value" as
    select sequence_id(ns.nspname, c.relname) as id,
           schema_id(ns.nspname) as schema_id,
           s.seqmax as maximum_value
    from pg_sequence s
         inner join pg_class c on c.oid = s.seqrelid
         inner join pg_namespace ns on ns.oid = c.relnamespace;

create view "sequence_increment" as
    select sequence_id(ns.nspname, c.relname) as id,
           schema_id(ns.nspname) as schema_id,
           s.seqincrement as increment
    from pg_sequence s
         inner join pg_class c on c.oid = s.seqrelid
         inner join pg_namespace ns on ns.oid = c.relnamespace;

create view "sequence_cache" as
    select sequence_id(ns.nspname, c.relname) as id,
           schema_id(ns.nspname) as schema_id,
           s.seqcache as cache
    from pg_sequence s
         inner join pg_class c on c.oid = s.seqrelid
         inner join pg_namespace ns on ns.oid = c.relnamespace;

create view "sequence_cycle" as
    select sequence_id(ns.nspname, c.relname) as id,
           schema_id(ns.nspname) as schema_id
    from pg_sequence s
         inner join pg_class c on c.oid = s.seqrelid
         inner join pg_namespace ns on ns.oid = c.relnamespace
    where s.seqcycle;

create view "sequence_type" as
    select sequence_id(ns.nspname, c.relname) as id,
           schema_id(ns.nspname) as schema_id,
           type_id(tns.nspname, resolved_type_name(t))
    from pg_sequence s
         inner join pg_class c on c.oid = s.seqrelid
         inner join pg_namespace ns on ns.oid = c.relnamespace
         inner join pg_type t on t.oid = s.seqtypid
         inner join pg_namespace tns on tns.oid = t.typnamespace;

create view "sequence_table" as
    select
        sequence_id(ns.nspname, c.relname) as id,
        schema_id(ns.nspname)              as schema_id,
        table_id(rns.nspname, r.relname)
    from
        pg_sequence             s
        inner join pg_class     c on c.oid = s.seqrelid
        inner join pg_namespace ns on ns.oid = c.relnamespace
        left join  pg_depend    d
                   on d.objid = c.oid
                       and d.classid = 'pg_class'::regclass
                       and d.deptype in ('a', 'i')
        left join pg_class r on r.oid = d.refobjid
        left join pg_namespace rns on rns.oid = r.relnamespace
    where
        r is distinct from null and
        c.relkind = 'S';

/******************************************************************************
 * table
 *****************************************************************************/
create view "table" as
    select
        relation_id(ns.nspname, c.relname) as id,
        schema_id(ns.nspname)              as schema_id,
        ns.nspname                         as schema_name,
        c.relname                          as name
    from
        pg_class                c
        inner join pg_namespace ns on ns.oid = c.relnamespace
    where
        c.relkind = 'r' or
        c.relkind = 'p';

create view table_partitioned as
    select
        relation_id(ns.nspname, c.relname) as id
    from
        pg_class                c
        inner join pg_namespace ns on ns.oid = c.relnamespace
    where
        c.relkind = 'p';

create view table_permanent as
    select
        relation_id(ns.nspname, c.relname) as id
    from
        pg_class                c
        inner join pg_namespace ns on ns.oid = c.relnamespace
    where
        c.relkind = 'r' and
        c.relpersistence = 'p';

create view table_temporary as
    select
        relation_id(ns.nspname, c.relname) as id
    from
        pg_class                c
        inner join pg_namespace ns on ns.oid = c.relnamespace
    where
        c.relkind = 'r' and
        c.relpersistence = 't';

create view table_rowsecurity as
    select relation_id(schemaname, tablename) as id
    from pg_catalog.pg_tables
where pg_tables.rowsecurity;

create view table_forcerowsecurity as
    select relation_id(schemaname, tablename) as id
    from pg_catalog.pg_tables
inner join pg_catalog.pg_namespace ns on pg_tables.schemaname = ns.nspname
inner join pg_catalog.pg_class c on pg_tables.tablename = c.relname and ns.oid = c.relnamespace
where relforcerowsecurity;

/******************************************************************************
 * view
 *****************************************************************************/
create view "view" as
    select relation_id(table_schema, table_name) as id,
           schema_id(table_schema) as schema_id,
           table_schema::text as schema_name,
           table_name::text as name
    from information_schema.views v;

create view view_definition as
    select relation_id(table_schema, table_name) as id,
           view_definition::text as query
    from information_schema.views v;

/******************************************************************************
 * relation_column
 *****************************************************************************/
create view relation_column as
    select column_id(c.table_schema, c.table_name, c.column_name) as id,
           relation_id(c.table_schema, c.table_name) as relation_id,
           c.column_name::text as name
    from information_schema.columns c;

create view relation_column_position as
select column_id(c.table_schema, c.table_name, c.column_name) as id,
       c.ordinal_position::integer                            as position
from information_schema.columns c;

create view relation_column_default as
    select
        column_id(ns.nspname, c.relname, a.attname)                                                     as id,
        cast(case when a.attgenerated = '' then _pg_get_expr(ad.adbin, ad.adrelid) end as text) as "default"
    from
        pg_attribute a inner join pg_attrdef ad on attrelid = adrelid and attnum = adnum
        inner join pg_class                                                               c on a.attrelid = c.oid
            and c.relkind = any ('{r,v,m,f,p}')
        inner join pg_namespace                                                           ns on ns.oid = c.relnamespace
    where
        a.attnum > 0;

create view relation_column_identity as
    select
        column_id(ns.nspname, c.relname, a.attname)                                                     as id,
        a.attidentity = 'a' as always
    from
        pg_attribute a
        inner join pg_class                                                               c on a.attrelid = c.oid
            and c.relkind = any ('{r,v,m,f,p}')
        inner join pg_namespace                                                           ns on ns.oid = c.relnamespace
    where
        a.attnum > 0 and a.attidentity in ('a','d');

create view relation_column_generated as
    select
        column_id(ns.nspname, c.relname, a.attname)                                                     as id,
        _pg_get_expr(ad.adbin, ad.adrelid)::text as generated
    from
        pg_attribute a inner join pg_attrdef ad on attrelid = adrelid and attnum = adnum
        inner join pg_class                                                               c on a.attrelid = c.oid
            and c.relkind = any ('{r,v,m,f,p}')
        inner join pg_namespace                                                           ns on ns.oid = c.relnamespace
    where
        a.attnum > 0 and a.attgenerated in ('s');


create view relation_column_type as
    select
        column_id(ns.nspname, c.relname, col.attname)                                                      as id,
        type_id(nt.nspname, resolved_type_name(t)) as "type_id"
    from
        (pg_attribute a left join pg_attrdef ad on attrelid = adrelid and attnum = adnum) col
        inner join pg_class                                                               c on col.attrelid = c.oid
            and c.relkind = any('{r,v,m,f,p}')
        inner join pg_namespace                                                           ns on ns.oid = c.relnamespace
        join       (pg_type t join pg_namespace nt on (t.typnamespace = nt.oid)) on col.atttypid = t.oid
    where col.attnum > 0;

create view relation_column_nullable as
    select column_id(c.table_schema, c.table_name, c.column_name) as id
    from information_schema.columns c
        where c.is_nullable = 'YES';

/******************************************************************************
 * relation
 *****************************************************************************/
create view relation as
    select relation_id(ns.nspname, c.relname) as id,
           schema_id(ns.nspname) as schema_id,
           ns.nspname::text as schema_name,
           c.relname::text as name
    from pg_class c
    inner join pg_namespace ns on ns.oid = c.relnamespace
    where reltype != 0;

create view relation_primary_key as
    select relation_id(ns.nspname, c.relname) as id,
           column_id(ns.nspname, c.relname, a.attname),
           position
    from pg_class c
    inner join pg_namespace ns on ns.oid = c.relnamespace
    inner join pg_constraint ct on ct.conrelid = c.oid and ct.contype = 'p'
    join lateral (select * from unnest(ct.conkey) with ordinality as t(key, position)) as k on true
    inner join pg_attribute a on a.attrelid = c.oid and a.attnum = key
    where reltype != 0;

create view relation_foreign_key as
    select relation_id(ns.nspname, c.relname) as id,
           column_id(ns.nspname, c.relname, a.attname),
           position
    from pg_class c
    inner join pg_namespace ns on ns.oid = c.relnamespace
    inner join pg_constraint ct on ct.conrelid = c.oid and ct.contype = 'f'
    join lateral (select * from unnest(ct.conkey) with ordinality as t(key, position)) as k on true
    inner join pg_attribute a on a.attrelid = c.oid and a.attnum = key
    where reltype != 0;

create view relation_foreign_key_constraint as
    select relation_id(ns.nspname, c.relname) as id,
           constraint_id(ns.nspname, c.relname, ct.conname)
    from pg_class c
    inner join pg_namespace ns on ns.oid = c.relnamespace
    inner join pg_constraint ct on ct.conrelid = c.oid and ct.contype = 'f'
    join lateral (select * from unnest(ct.conkey) with ordinality as t(key, position)) as k on true
    inner join pg_attribute a on a.attrelid = c.oid and a.attnum = key
    where reltype != 0;

/******************************************************************************
 * function
 *****************************************************************************/

create or replace function _pg_get_function_sqlbody(p pg_proc) returns text language plpgsql stable
as $$
begin
    if (current_setting('server_version_num')::int/10000) = 13 then
        return p.prosrc;
    else
        return coalesce(pg_catalog.pg_get_function_sqlbody(p.oid), p.prosrc);
    end if;
end;
$$;

create or replace function _pg_get_function_result(p oid) returns text language sql stable
as $$ select pg_get_function_result(p); $$
    -- Force qualified naming
    set search_path = '';


create or replace view "callable" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace)
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id, -- function_id
        schema_id(schema_name) as schema_id,
        schema_name,
        name,
        type_sig
    from
        orig;


create or replace view "callable_function" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.prokind = 'f')
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_argument_name" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig,
                argname, ordinality
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
                inner join lateral (select * from unnest(proargnames) with ordinality as a(argname, ordinality)) as t(argname, ordinality) on true
           )
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id,
        argname as argument_name,
        ordinality as argument_position
    from
        orig where argname != '';

create or replace view "callable_argument_type" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig,
                argtype, ordinality
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
                inner join lateral (select * from unnest(coalesce(proallargtypes, proargtypes)) with ordinality as a(argtype, ordinality)) as t(argtype, ordinality) on true
        )
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id,
        type_id(tns.nspname, resolved_type_name(t)) as type_id,
        ordinality as argument_position
    from orig
         inner join pg_type t on t.oid = argtype
         inner join pg_namespace tns on tns.oid = t.typnamespace;

create or replace view "callable_argument_mode" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig,
                argmode, ordinality
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
                inner join lateral (select * from unnest(coalesce(proargmodes, array_fill('i'::"char", array[coalesce(cardinality(proallargtypes), pronargs)]))) with ordinality as a(argmode, ordinality)) as t(argmode, ordinality) on true
        )
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id,
        ordinality as argument_position,
    case when argmode = 'i' then 'in'
         when argmode = 'o' then 'out'
         when argmode = 'b' then 'inout'
         when argmode = 'v' then 'variadic'
         when argmode = 't' then 'table'
    end as argument_mode
    from orig;


create or replace view "callable_argument_default" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig,
                argdefault, ordinality
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
                inner join lateral (select pg_get_function_arg_default(p.oid, ordinality), ordinality from generate_series(1, p.pronargs) ordinality) as t(argdefault, ordinality) on true
        )
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id,
        argdefault as argument_default,
        ordinality as argument_position
    from orig
    where argdefault is not null;

create or replace view "callable_aggregate" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.prokind = 'a')
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_window" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.prokind = 'w')
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_procedure" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.prokind = 'p')
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_immutable" as
with
    orig as (
        select
            n.nspname                                                               as "schema_name",
            p.proname                                                               as "name",
            _get_function_type_sig_array(p) as type_sig
        from
            pg_proc                            p
            inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
        where
            p.provolatile = 'i')
select
    function_id(
            schema_name,
            name,
            type_sig
    )                      as id
from
    orig;

create or replace view "callable_stable" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.provolatile = 's')
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_volatile" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.provolatile = 'v')
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_parallel_safe" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.proparallel = 's')
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_parallel_restricted" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.proparallel = 'r')
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_parallel_unsafe" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.proparallel = 'u')
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_security_definer" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.prosecdef)
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_security_invoker" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                not p.prosecdef)
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id
    from
        orig;

create or replace view "callable_language" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig,
                p.prolang
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                not p.prosecdef)
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id,
        l.lanname as language
    from orig
         inner join pg_language l on l.oid = prolang;

create or replace view "callable_acl" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig,
                p.proacl
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                not p.prosecdef)
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id,
        aclitem as acl
    from orig
      inner join lateral (select * from unnest(proacl)) acl(aclitem) on true;

create or replace view "callable_owner" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig,
                p.proowner
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                not p.prosecdef)
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id,
        pg_get_userbyid(proowner) as owner
    from orig;

create or replace view "callable_body" as
    with
        orig as (
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig,
                _pg_get_function_sqlbody(p) as body
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                not p.prosecdef)
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                      as id,
        body
    from orig;

create or replace view "callable_return_type" as
    with
        orig as (
            -- slightly modified version of query output by \df+
            select
                n.nspname                                                               as "schema_name",
                p.proname                                                               as "name",
                _get_function_type_sig_array(p) as type_sig,
                _pg_get_function_result(p.oid)                                          as "return_type",
                p.prorettype                                                            as "return_type_id",
                p.proretset as "returns_set"
            from
                pg_proc                            p
                inner join pg_catalog.pg_namespace n on n.oid = p.pronamespace
            where
                p.prokind = 'f')
    select
        function_id(
                schema_name,
                name,
                type_sig
        )                                         as id,
        returns_set as setof,
        type_id(ns.nspname, resolved_type_name(t)) as return_type_id
    from
        orig
        inner join pg_type      t on t.oid = return_type_id
        inner join pg_namespace ns on ns.oid = t.typnamespace;

/******************************************************************************
 * trigger
 *****************************************************************************/
create view trigger as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id,
           t.id as relation_id,
           t_pgn.nspname::text as schema_name,
           pgc.relname::text as relation_name,
           pg_trigger.tgname::text as name

    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace

    inner join "schema" t_s
            on t_s.name = t_pgn.nspname

    inner join "table" t
            on t.schema_id = t_s.id and
               t.name = pgc.relname

    where
           not tgisinternal;

create view trigger_callable as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id,
           function_id(
               f_pgn.nspname,
               pgp.proname,
               _get_function_type_sig_array(pgp)
           ) as callable_id

    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace

    inner join "schema" t_s
            on t_s.name = t_pgn.nspname

    inner join "table" t
            on t.schema_id = t_s.id and
               t.name = pgc.relname

    inner join pg_proc pgp
            on pgp.oid = tgfoid

    inner join pg_namespace f_pgn
            on f_pgn.oid = pgp.pronamespace
    where
           not tgisinternal;

create view trigger_when as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id,
           case when (tgtype >> 1 & 1)::bool then 'before'
                when (tgtype >> 6 & 1)::bool then 'instead'
                else 'after'
           end as "when"

    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    where 
           not tgisinternal;

create view trigger_insert as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id
    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    where
           not tgisinternal
           and (tgtype >> 2 & 1)::bool;

create view trigger_delete as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id
    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    where
           not tgisinternal
           and (tgtype >> 3 & 1)::bool;

create view trigger_update as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id
    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    where

           not tgisinternal
           and (tgtype >> 4 & 1)::bool;

create view trigger_truncate as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id
    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    where
           not tgisinternal
           and (tgtype >> 5 & 1)::bool;

create view trigger_level as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id,
    case when (tgtype & 1)::bool then 'row'
         else 'statement'
    end as level

    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    where
           not tgisinternal;

create view trigger_enabled as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id
    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    where
           not tgisinternal
           and tgenabled <> 'D';

create view trigger_deferrable as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id
    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    where
           not tgisinternal
           and tgdeferrable;

create view trigger_initially_deferred as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id
    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    where
           not tgisinternal
           and tginitdeferred;

create view trigger_arguments as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id,
           tgnargs as number_of_arguments,
           (select array_agg(convert_from(decode(unnest, 'hex'), 'utf8')) from unnest(regexp_split_to_array(encode(tgargs, 'hex'), '00')) where unnest <> '') as arguments
    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    where
           not tgisinternal;

create view trigger_columns as
    select trigger_id(t_pgn.nspname, pgc.relname, pg_trigger.tgname) as id,
           column_id(t_pgn.nspname, pgc.relname, pga.attname) as column_id
    from pg_trigger

    inner join pg_class pgc
            on pgc.oid = tgrelid

    inner join pg_namespace t_pgn
            on t_pgn.oid = pgc.relnamespace
    cross join lateral unnest(tgattr) attribute
    join pg_attribute pga
            on pga.attrelid = tgrelid
               and pga.attnum = attribute
    where
           not tgisinternal;

/******************************************************************************
 * role
 *****************************************************************************/
create view role as
   select role_id(pgr.rolname) as id,
          pgr.rolname::text  as name
   from pg_roles pgr
    union all
   select role_id('0'::oid::regrole::text) as id,
    'PUBLIC' as name;

create view role_superuser as
   select role_id(pgr.rolname) as id
   from pg_roles pgr
   where pgr.rolsuper;

create view role_inherit as
   select role_id(pgr.rolname) as id
   from pg_roles pgr
   where pgr.rolinherit;

create view role_create_role as
   select role_id(pgr.rolname) as id
   from pg_roles pgr
   where pgr.rolcreaterole;

create view role_create_db as
   select role_id(pgr.rolname) as id
   from pg_roles pgr
   where pgr.rolcreatedb;

create view role_can_login as
   select role_id(pgr.rolname) as id
   from pg_roles pgr
   where pgr.rolcanlogin;

create view role_replication as
   select role_id(pgr.rolname) as id
   from pg_roles pgr
   where pgr.rolreplication;

create view role_connection_limit as
   select role_id(pgr.rolname) as id,
          pgr.rolconnlimit   as connection_limit
   from pg_roles pgr;

create view role_valid_until as
   select role_id(pgr.rolname) as id,
          pgr.rolvaliduntil   as valid_until
   from pg_roles pgr where pgr.rolvaliduntil is not null;

create view role_setting as
    select role_setting_id(pgr.rolname, pgd.datname, split_part(setting, '=', 1)) as id,
           role_id(pgr.rolname) as role_id,
           split_part(setting, '=', 1) as setting_name,
           split_part(setting, '=', 2) as setting_value
    from pg_db_role_setting pgrs
        join pg_roles pgr on pgr.oid = pgrs.setrole
        left join pg_database pgd on pgd.oid = pgrs.setdatabase,
        unnest(pgrs.setconfig) as setting;

create view role_member as
select
    role_id(r.rolname::text) as id,
    role_id(r2.rolname::text) as member_role_id
from pg_auth_members m
    join pg_roles r on r.oid = m.roleid
    join pg_roles r2 on r2.oid = m.member;


/******************************************************************************
 * policy
 *****************************************************************************/

create view policy as
select policy_id(n.nspname, c.relname, p.polname) as id,
    p.polname::text as name,
    relation_id(n.nspname, c.relname) as relation_id,
    c.relname::text as relation_name,
    n.nspname::text as schema_name,
    p.polcmd::char::siuda as command,
    _pg_get_expr(p.polqual, p.polrelid, True) as using,
    _pg_get_expr(p.polwithcheck, p.polrelid, True) as check
from pg_policy p
    join pg_class c on c.oid = p.polrelid
    join pg_namespace n on n.oid = c.relnamespace;


/******************************************************************************
 * policy_role
 *****************************************************************************/
create view policy_role as
select
--    policy_id((relation_id).schema_name, (relation_id).name, policy_name)::text || '<-->' || role_id::text as id,
    policy_id((relation_id).schema_name, (relation_id).name, policy_name) as id,
    policy_name::text,
    relation_id,
    (relation_id).name as relation_name,
    (relation_id).schema_name as schema_name,
    role_id,
    (role_id).name as role_name
from (
    select
        p.polname as policy_name,
        relation_id(n.nspname, c.relname) as relation_id,
        unnest(p.polroles::regrole[]::text[]::role_id[]) as role_id
    from pg_policy p
        join pg_class c on c.oid = p.polrelid
        join pg_namespace n on n.oid = c.relnamespace
) a;


/******************************************************************************
 * constraint_relation_non_null*
 *****************************************************************************/

create view constraint_relation_non_null_column as
    select
        constraint_id(n.nspname, r.relname, a.attname || '_not_null') as id,
        column_id(n.nspname, r.relname, a.attname)
    from
        pg_namespace n,
        pg_class     r,
        pg_attribute a
    where
        n.oid = r.relnamespace and
        r.oid = a.attrelid and
        a.attnum > 0 and
        not a.attisdropped and
        a.attnotnull and
        r.relkind in ('r', 'p') and
        pg_has_role(r.relowner, 'USAGE');

create view constraint_relation_non_null as
    select
        constraint_id(n.nspname, r.relname, a.attname || '_not_null') as id,
        relation_id(n.nspname, r.relname),
        a.attname || '_not_null'                                      as name
    from
        pg_namespace n,
        pg_class     r,
        pg_attribute a
    where
        n.oid = r.relnamespace and
        r.oid = a.attrelid and
        a.attnum > 0 and
        not a.attisdropped and
        a.attnotnull and
        r.relkind in ('r', 'p') and
        pg_has_role(r.relowner, 'USAGE');


/******************************************************************************
 * constraint_relation_check*
 *****************************************************************************/
create view constraint_relation_check as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id,
        relation_id(cns.nspname, cc.relname),
        c.conname                                        as name

    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
    where
        c.contype = 'c'
    union all
    select *
    from
        constraint_relation_non_null;

create view constraint_relation_check_expr as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id,
        _pg_get_expr(c.conbin, coalesce(cc.oid, 0))       as expr

    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
    where
        c.contype = 'c'
    union all
    select
        id,
        (column_id).name || ' IS NOT NULL' as expr
    from
        constraint_relation_non_null_column cruc;

/******************************************************************************
 * constraint_relation_unique*
 *****************************************************************************/
create view constraint_relation_unique as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id,
        relation_id(cns.nspname, cc.relname),
        c.conname                                        as name

    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
    where
        c.contype = 'u';

create view constraint_relation_unique_column as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id,
        column_id(ns.nspname, cc.relname, a.attname)
    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
        inner join pg_attribute a on a.attrelid = cc.oid and a.attnum = any(c.conkey)
    where
        c.contype = 'u';

/******************************************************************************
 * constraint_relation_foreign_key*
 *****************************************************************************/
create view constraint_relation_foreign_key as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id,
        relation_id(cns.nspname, cc.relname),
        c.conname                                        as name

    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
    where
        c.contype = 'f';

create view constraint_relation_foreign_key_update as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id,
        (case when c.confupdtype = 'a' then 'no action'
         when c.confupdtype = 'r' then 'restrict'
         when c.confupdtype = 'c' then 'cascade'
         when c.confupdtype = 'n' then 'set null'
         when c.confupdtype = 'd' then 'set default'
         else null end) as update_action
    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
    where
        c.contype = 'f';

create view constraint_relation_foreign_key_delete as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id,
        (case when c.confdeltype = 'a' then 'no action'
         when c.confdeltype = 'r' then 'restrict'
         when c.confdeltype = 'c' then 'cascade'
         when c.confdeltype = 'n' then 'set null'
         when c.confdeltype = 'd' then 'set default'
         else null end) as delete_action
    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
    where
        c.contype = 'f';

create view constraint_relation_foreign_key_match_full as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id
    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
    where
        c.contype = 'f' and c.confmatchtype = 'f';

create view constraint_relation_foreign_key_match_partial as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id
    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
    where
        c.contype = 'f' and c.confmatchtype = 'p';

create view constraint_relation_foreign_key_match_simple as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id
    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
    where
        c.contype = 'f' and c.confmatchtype = 's';

create view constraint_relation_foreign_key_references as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id,
        relation_id(rns.nspname, rc.relname)
    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
        inner join pg_class rc on rc.oid = c.confrelid
        inner join pg_namespace rns on rns.oid = rc.relnamespace
    where
        c.contype = 'f';

create view constraint_relation_foreign_key_references_column as
    select
        constraint_id(ns.nspname, cc.relname, c.conname) as id,
        column_id(rns.nspname, rc.relname, a.attname),
        position
    from
        pg_constraint           c
        inner join pg_namespace ns on ns.oid = c.connamespace
        inner join pg_class     cc on cc.oid = c.conrelid
        inner join pg_namespace cns on cns.oid = cc.relnamespace
        inner join pg_class rc on rc.oid = c.confrelid
        inner join pg_namespace rns on rns.oid = rc.relnamespace
        join lateral (select * from unnest(c.confkey) with ordinality as t(key, position)) as k on true
        inner join pg_attribute a on a.attrelid = rc.oid and a.attnum = key
    where
        c.contype = 'f';
/******************************************************************************
 * extension
 *****************************************************************************/
create view extension as
    select extension_id(ext.extname) as id,
           schema_id(pgn.nspname) as schema_id,
           pgn.nspname::text as schema_name,
           ext.extname::text as name,
           ext.extversion as version

    from pg_catalog.pg_extension ext
    inner join pg_catalog.pg_namespace pgn
            on pgn.oid = ext.extnamespace;


/******************************************************************************
 * foreign_data_wrapper
 *****************************************************************************/
create view foreign_data_wrapper as
    select id,
           name::text,
--            handler_id,
--            validator_id,
           opt as options

    from (
        select foreign_data_wrapper_id(fdwname) as id,
               fdwname as name,
--                h_f.id as handler_id,
--                v_f.id as validator_id,
               jsonb_build_object(variadic string_to_array(unnest(coalesce(fdwoptions, array['']::text[])), '=')) as opt

        from pg_catalog.pg_foreign_data_wrapper

        left join pg_proc p_h
               on p_h.oid = fdwhandler

        left join pg_namespace h_n
               on h_n.oid = p_h.pronamespace

--         left join function h_f
--                on h_f.schema_name = h_n.nspname and
--                   h_f.name = p_h.proname

        left join pg_proc p_v
               on p_v.oid = fdwvalidator

        left join pg_namespace v_n
               on v_n.oid = p_v.pronamespace

--         left join "function" v_f
--                on v_f.schema_name = v_n.nspname and
--                   v_f.name = p_v.proname
    ) q

    group by id,
             opt,
             name;
--              handler_id,
--              validator_id;


/******************************************************************************
 * foreign_server
 *****************************************************************************/
create view foreign_server as
    select id,
           foreign_data_wrapper_id,
           name::text,
           "type",
           version,
           opt as options

    from (
        select foreign_server_id(srvname) as id,
               foreign_data_wrapper_id(fdwname) as foreign_data_wrapper_id,
               srvname as name,
               srvtype as "type",
               srvversion as version,
               jsonb_build_object(variadic string_to_array(unnest(coalesce(srvoptions, array['']::text[])), '=')) as opt

        from pg_catalog.pg_foreign_server fs
        inner join pg_catalog.pg_foreign_data_wrapper fdw
                on fdw.oid = fs.srvfdw
    ) q

    group by id,
             opt,
             foreign_data_wrapper_id,
             name,
             "type",
             version;



/******************************************************************************
 * foreign_table
 *****************************************************************************/
create view foreign_table as
    select id,
           foreign_server_id,
           schema_id,
           schema_name::text,
           name::text,
           opt as options

    from (
        select relation_id(pgn.nspname, pgc.relname) as id,
               schema_id(pgn.nspname) as schema_id,
               foreign_server_id(pfs.srvname) as foreign_server_id,
               pgn.nspname as schema_name,
               pgc.relname as name,
               jsonb_build_object(variadic string_to_array(unnest(coalesce(ftoptions, array['']::text[])), '=')) as opt

        from pg_catalog.pg_foreign_table pft
        inner join pg_catalog.pg_class pgc
                on pgc.oid = pft.ftrelid
        inner join pg_catalog.pg_namespace pgn
                on pgn.oid = pgc.relnamespace
        inner join pg_catalog.pg_foreign_server pfs
                on pfs.oid = pft.ftserver
    ) q

    group by id,
             opt,
             schema_id,
             foreign_server_id,
             schema_name,
             name;


/******************************************************************************
 * foreign_column
 *****************************************************************************/

create view foreign_column as
    select column_id(c.table_schema, c.table_name, c.column_name) as id,
           relation_id(c.table_schema, c.table_name) as foreign_table_id,
           c.table_schema::text as schema_name,
           c.table_name::text as foreign_table_name,
           c.column_name::text as name,
           quote_ident(c.udt_schema) || '.' || quote_ident(c.udt_name) as "type",
           (c.is_nullable = 'YES') as nullable

    from pg_catalog.pg_foreign_table pft
    inner join pg_catalog.pg_class pgc
            on pgc.oid = pft.ftrelid
    inner join pg_catalog.pg_namespace pgn
            on pgn.oid = pgc.relnamespace
    inner join information_schema.columns c
            on c.table_schema = pgn.nspname and
               c.table_name = pgc.relname;

/******************************************************************************
 * index
 *****************************************************************************/
create view "index" as
select index_id(ns.nspname, c.relname) as id,
       ns.nspname as schema_name,
       c.relname as name
from pg_index i
         inner join pg_class c on c.oid = i.indexrelid
         inner join pg_class cr on cr.oid = i.indrelid and cr.relkind != 't'
         inner join pg_namespace ns on ns.oid = c.relnamespace;

create view "index_relation" as
select index_id(ns.nspname, c.relname) as id,
       relation_id(rns.nspname, rc.relname)
from pg_index i
         inner join pg_class c on c.oid = i.indexrelid
         inner join pg_class cr on cr.oid = i.indrelid and cr.relkind != 't'
         inner join pg_namespace ns on ns.oid = c.relnamespace
         inner join pg_class rc on rc.oid = i.indrelid
         inner join pg_namespace rns on rns.oid = rc.relnamespace;

create view "index_unique" as
select index_id(ns.nspname, c.relname) as id
from pg_index i
         inner join pg_class c on c.oid = i.indexrelid
         inner join pg_class cr on cr.oid = i.indrelid and cr.relkind != 't'
         inner join pg_namespace ns on ns.oid = c.relnamespace
where i.indisunique;

if current_setting('server_version_num')::int / 10000 > 14 then
create view "index_unique_null_values_distinct" as
select index_id(ns.nspname, c.relname) as id
from pg_index i
         inner join pg_class c on c.oid = i.indexrelid
         inner join pg_class cr on cr.oid = i.indrelid and cr.relkind != 't'
         inner join pg_namespace ns on ns.oid = c.relnamespace
where i.indisunique
  and not i.indnullsnotdistinct;
else
create view "index_unique_null_values_distinct" as
    select index_id(ns.nspname, c.relname) as id
    from pg_index i
         inner join pg_class c on c.oid = i.indexrelid
         inner join pg_class cr on cr.oid = i.indrelid and cr.relkind != 't'
         inner join pg_namespace ns on ns.oid = c.relnamespace;
end if;

create view "index_primary_key" as
select index_id(ns.nspname, c.relname) as id
from pg_index i
         inner join pg_class c on c.oid = i.indexrelid
         inner join pg_class cr on cr.oid = i.indrelid and cr.relkind != 't'
         inner join pg_namespace ns on ns.oid = c.relnamespace
where i.indisprimary;

create view "index_unique_immediate" as
select index_id(ns.nspname, c.relname) as id
from pg_index i
         inner join pg_class c on c.oid = i.indexrelid
         inner join pg_class cr on cr.oid = i.indrelid and cr.relkind != 't'
         inner join pg_namespace ns on ns.oid = c.relnamespace
where i.indisunique
  and i.indimmediate;

create view "index_replica_identity" as
    select index_id(ns.nspname, c.relname) as id
    from pg_index i
         inner join pg_class c on c.oid = i.indexrelid
         inner join pg_class cr on cr.oid = i.indrelid and cr.relkind != 't'
         inner join pg_namespace ns on ns.oid = c.relnamespace
    where i.indisreplident;

create view "index_attribute" as
    select index_id(ns.nspname, c.relname) as id,
           attribute, position
    from pg_index i
         inner join pg_class c on c.oid = i.indexrelid
         inner join pg_class cr on cr.oid = i.indrelid and cr.relkind != 't'
         inner join pg_namespace ns on ns.oid = c.relnamespace
    inner join lateral (select position, pg_get_indexdef(indexrelid::oid, position, true) as attribute from generate_series(1,indnatts) position) t
    on true
where indnatts > 0;

create view "index_partial" as
    select index_id(ns.nspname, c.relname) as id,
           _pg_get_expr(indpred, indrelid) as condition
    from pg_index i
         inner join pg_class c on c.oid = i.indexrelid
         inner join pg_class cr on cr.oid = i.indrelid and cr.relkind != 't'
         inner join pg_namespace ns on ns.oid = c.relnamespace
    where i.indpred is not null;

--- language

create view language as
    select language_id(lanname) as id
from pg_language;

create view language_trusted as
    select language_id(lanname) as id
from pg_language where lanpltrusted;

create view language_internal as
    select language_id(lanname) as id
from pg_language where not lanispl;

create view language_handler as
    select language_id(lanname) as id,
           function_id(ns.nspname, p.proname, _get_function_type_sig_array(p)) as handler_id
from pg_language l
     inner join pg_proc p on p.oid = l.lanplcallfoid
     inner join pg_namespace ns on ns.oid = p.pronamespace
where lanispl;

create view language_inline_handler as
    select language_id(lanname) as id,
           function_id(ns.nspname, p.proname, _get_function_type_sig_array(p)) as handler_id
from pg_language l
     inner join pg_proc p on p.oid = l.laninline
     inner join pg_namespace ns on ns.oid = p.pronamespace
where lanispl and laninline != 0;

create view language_validator as
    select language_id(lanname) as id,
           function_id(ns.nspname, p.proname, _get_function_type_sig_array(p)) as handler_id
from pg_language l
     inner join pg_proc p on p.oid = l.laninline
     inner join pg_namespace ns on ns.oid = p.pronamespace
where lanispl and lanvalidator != 0;

create view language_acl as
    select language_id(lanname) as id,
           aclitem as acl
from pg_language
  inner join lateral (select * from unnest(lanacl)) acl(aclitem) on true;

-- TODO: language owner

--- descriptions/comments

create view comment as
    -- relation
    select
        relation_id(ns.nspname, c.relname)::object_id as id,
        description                                   as comment
    from
        pg_description          d
        inner join pg_class     c on c.oid = d.objoid and d.classoid = 'pg_class'::regclass
        inner join pg_namespace ns on ns.oid = c.relnamespace
        where d.objsubid = 0
    union all
    -- callable
    select
        function_id(ns.nspname, p.proname, _get_function_type_sig_array(p))::object_id as id,
        description                                   as comment
    from
        pg_description          d
        inner join pg_proc     p on p.oid = d.objoid and d.classoid = 'pg_proc'::regclass
        inner join pg_namespace ns on ns.oid = p.pronamespace
    union all
    -- column
    select
        column_id(ns.nspname, c.relname, a.attname)::object_id as id,
        description                                   as comment
    from
        pg_description          d
        inner join pg_class     c on c.oid = d.objoid and d.classoid = 'pg_class'::regclass
        inner join pg_attribute a on a.attrelid = c.oid and a.attnum > 0
        inner join pg_namespace ns on ns.oid = c.relnamespace
        where d.objsubid != 0
    union all
    -- cast
    select
        cast_id(st_ns.nspname, st.typname, tt_ns.nspname, tt.typname)::object_id as id,
        description                                                              as comment
    from
        pg_description          d
        inner join pg_cast      c on c.oid = d.objoid and d.classoid = 'pg_cast'::regclass
        inner join pg_type      st on st.oid = c.castsource
        inner join pg_type      tt on tt.oid = c.casttarget
        inner join pg_namespace st_ns on st_ns.oid = st.typnamespace
        inner join pg_namespace tt_ns on tt_ns.oid = tt.typnamespace;

--- dependencies

create view obj_object_id as
    select
        'pg_class'::regclass::oid                     as classid,
        c.oid                                         as objid,
        0::oid                                        as objsubid,
        relation_id(ns.nspname, c.relname)::object_id as object_id
    from
        pg_class                c
        inner join pg_namespace ns on ns.oid = c.relnamespace
    union all
    select
        'pg_extension'::regclass::oid      as classid,
        e.oid                              as objid,
        0::oid                             as objsubid,
        extension_id(e.extname)::object_id as object_id
    from
        pg_extension            e
        inner join pg_namespace ns on ns.oid = e.extnamespace
    union all
    select

        'pg_type'::regclass::oid                              as classid,
        t.oid                                                 as objid,
        0::oid                                                as objsubid,
        type_id(ns.nspname, resolved_type_name(t))::object_id as object_id
    from
        pg_type                 t
        inner join pg_namespace ns on ns.oid = t.typnamespace
    union all
    select
        'pg_namespace'::regclass::oid    as classid,
        ns.oid                           as objid,
        0::oid                           as objsubid,
        schema_id(ns.nspname)::object_id as object_id
    from
        pg_namespace ns
    union all
    select
        'pg_proc'::regclass::oid                                                       as classid,
        p.oid                                                                          as objid,
        0::oid                                                                         as objsubid,
        function_id(ns.nspname, p.proname, _get_function_type_sig_array(p))::object_id as object_id
    from
        pg_proc                 p
        inner join pg_namespace ns on ns.oid = p.pronamespace
    union all
    select
        'pg_language'::regclass::oid    as classid,
        l.oid                           as objid,
        0::oid                          as objsubid,
        language_id(lanname)::object_id as object_id
    from
        pg_language l;
--- TODO: handle the rest of the cases (^^)

create view dependency as
    with
        pre as materialized (
            -- relation
            select
                relation_id(ns.nspname, c.relname)::object_id as id,
                d                                             as dependency
            from
                pg_depend               d
                inner join pg_class     c
                           on c.oid = d.objid and d.classid = 'pg_class'::regclass and c.relkind != 't'
                inner join pg_namespace ns on ns.oid = c.relnamespace
            where
                d.objsubid = 0
            union all
            -- callable
            select
                function_id(ns.nspname, p.proname, _get_function_type_sig_array(p))::object_id as id,
                d                                                                              as dependency
            from
                pg_depend               d
                inner join pg_proc      p on p.oid = d.objid and d.classid = 'pg_proc'::regclass
                inner join pg_namespace ns on ns.oid = p.pronamespace
            where
                d.objsubid = 0
            union all
            -- column
            select
                column_id(ns.nspname, c.relname, a.attname)::object_id as id,
                d                                                      as dependency
            from
                pg_depend               d
                inner join pg_class     c on c.oid = d.objid and d.classid = 'pg_class'::regclass
                inner join pg_attribute a on a.attrelid = c.oid and a.attnum > 0
                inner join pg_namespace ns on ns.oid = c.relnamespace
            where
                d.objsubid != 0
            union all
            --- columns also depend on what tables depend on
            select
                column_id(ns.nspname, c.relname, a.attname)::object_id as id,
                (('pg_attribute'::regclass, c.oid, a.attnum, d.refclassid, d.refobjid, d.refobjsubid,
                  d.deptype)::pg_depend)                               as dependency
            from
                pg_attribute            a
                inner join pg_class     c on c.oid = a.attrelid and c.reltype != 0
                inner join pg_namespace ns on ns.oid = c.relnamespace
                inner join pg_depend    d on d.objid = c.oid and d.classid = 'pg_class'::regclass
            where
                a.attnum > 0
            union all
            -- cast
            select
                cast_id(st_ns.nspname, st.typname, tt_ns.nspname, tt.typname)::object_id as id,
                d                                                                        as dependency
            from
                pg_depend               d
                inner join pg_cast      c on c.oid = d.objid and d.classid = 'pg_cast'::regclass
                inner join pg_type      st on st.oid = c.castsource
                inner join pg_type      tt on tt.oid = c.casttarget
                inner join pg_namespace st_ns on st_ns.oid = st.typnamespace
                inner join pg_namespace tt_ns on tt_ns.oid = tt.typnamespace
            where
                d.objsubid = 0
            -- type
            union all
            select
                type_id(ns.nspname, resolved_type_name(t))::object_id as id,
                d                                                     as dependency
            from
                pg_depend               d
                inner join pg_type      t on t.oid = d.objid and d.classid = 'pg_type'::regclass
                inner join pg_namespace ns on ns.oid = t.typnamespace
            ---- not all types were properly connected before Postgres 17 (arrays)
            union all
            select
                type_id(ns.nspname, resolved_type_name(t))::object_id as id,
                d                                                     as dependency
            from
                pg_type                 t
                inner join pg_namespace ns on ns.oid = t.typnamespace
                inner join pg_depend    d on d.objid = t.typelem and d.classid = 'pg_type'::regclass
            where
                (current_setting('server_version_num')::int / 10000) < 17 and
                t.typelem != 0
            ---- not all types were properly connected before Postgres 17 (relations)
            union all
            select
                type_id(ns.nspname, resolved_type_name(t))::object_id as id,
                d                                                     as dependency
            from
                pg_type                 t
                inner join pg_namespace ns on ns.oid = t.typnamespace
                inner join pg_depend    d on d.objid = t.typrelid and d.classid = 'pg_class'::regclass
            where
                (current_setting('server_version_num')::int / 10000) < 17 and
                t.typrelid != 0
            ---- not all types were properly connected before Postgres 17 (relation arrays)
            union all
            select
                type_id(ns.nspname, resolved_type_name(t))::object_id as id,
                d                                                     as dependency
            from
                pg_type                 t
                inner join pg_type      tr on tr.oid = t.typelem and tr.typrelid != 0
                inner join pg_namespace ns on ns.oid = t.typnamespace
                inner join pg_depend    d on d.objid = tr.typrelid and d.classid = 'pg_class'::regclass
            where
                (current_setting('server_version_num')::int / 10000) < 17 and
                t.typelem != 0
            -- TODO: add support for multirange types
            -- operator
            union all
            select
                operator_id(ns.nspname, o.oprname, lns.nspname, resolved_type_name(lt), rns.nspname,
                            resolved_type_name(rt))::object_id as id,
                d                                              as dependency
            from
                pg_depend               d
                inner join pg_operator  o on o.oid = d.objid and d.classid = 'pg_operator'::regclass
                inner join pg_namespace ns on ns.oid = o.oprnamespace
                left join  pg_type      lt on lt.oid = o.oprleft
                left join  pg_type      rt on rt.oid = o.oprright
                left join  pg_namespace lns on ns.oid = lt.typnamespace
                left join  pg_namespace rns on ns.oid = rt.typnamespace
            -- sequence
            union all
            select
                sequence_id(ns.nspname, r.relname)::object_id as id,
                d                                             as dependency
            from
                pg_depend               d
                inner join pg_sequence  s on s.seqrelid = d.objid and d.classid = 'pg_class'::regclass
                inner join pg_class     r on r.oid = s.seqrelid
                inner join pg_namespace ns on ns.oid = r.relnamespace
            -- index
            union all
            select
                index_id(ns.nspname, r.relname)::object_id as id,
                d                                          as dependency
            from
                pg_depend               d
                inner join pg_index     i on i.indrelid = d.objid and d.classid = 'pg_class'::regclass
                inner join pg_class     r on r.oid = i.indrelid
                inner join pg_namespace ns on ns.oid = r.relnamespace
            -- language
            union all
            select
                language_id(l.lanname)::object_id as id,
                d                                 as dependency
            from
                pg_depend              d
                inner join pg_language l on l.oid = d.objid and d.classid = 'pg_language'::regclass)

    select
        pre.id,
        oo.object_id as dependent_on
    from
        pre
        inner join obj_object_id oo
                   on oo.classid = (pre.dependency).refclassid and oo.objid = (pre.dependency).refobjid and
                      oo.objsubid = (pre.dependency).refobjsubid
    where
        (pre.dependency).deptype != 'i';

--- ACL

create view acl as
    -- callable
    select
        function_id(ns.nspname, p.proname, _get_function_type_sig_array(p))::object_id as id,
        acl.*
    from
        pg_proc                                                                                   p
        inner join pg_namespace                                                                   ns on ns.oid = p.pronamespace
        join       lateral ( select
                                 role_id(grantor::regrole::name) as grantor,
                                 role_id(grantee::regrole::name) as grantee,
                                 privilege_type,
                                 is_grantable,
                                 p.proacl is null                as "default"
                             from
                                 aclexplode(coalesce(p.proacl, acldefault('f', p.proowner))) ) as acl on true
-- type
    union all
    select
        type_id(ns.nspname, resolved_type_name(t))::object_id as id,
        acl.*
    from
        pg_type                                                                                   t
        inner join pg_namespace                                                                   ns on ns.oid = t.typnamespace
        join       lateral ( select
                                 role_id(grantor::regrole::name) as grantor,
                                 role_id(grantee::regrole::name) as grantee,
                                 privilege_type,
                                 is_grantable,
                                 t.typacl is null                as "default"
                             from
                                 aclexplode(coalesce(t.typacl, acldefault('T', t.typowner))) ) as acl on true
-- column
    union all
    select
        column_id(ns.nspname, c.relname, a.attname)::object_id as id,
        acl.*
    from
        pg_attribute                                                                              a
        inner join pg_class                                                                       c on c.oid = a.attrelid and c.reltype != 0
        inner join pg_namespace                                                                   ns on ns.oid = c.relnamespace
        join       lateral ( select
                                 role_id(grantor::regrole::name) as grantor,
                                 role_id(grantee::regrole::name) as grantee,
                                 privilege_type,
                                 is_grantable,
                                 a.attacl is null                as "default"
                             from
                                 aclexplode(coalesce(a.attacl, acldefault('c', c.relowner))) ) as acl on true
-- relation
    union all
    select
        relation_id(ns.nspname, c.relname)::object_id as id,
        acl.*
    from
        pg_class                                                                                  c
        inner join pg_namespace                                                                   ns on ns.oid = c.relnamespace
        join       lateral ( select
                                 role_id(grantor::regrole::name) as grantor,
                                 role_id(grantee::regrole::name) as grantee,
                                 privilege_type,
                                 is_grantable,
                                 c.relacl is null                as "default"
                             from
                                 aclexplode(coalesce(c.relacl, acldefault('r', c.relowner))) ) as acl on true
    where reltype != 0
-- schema
    union all
    select
        schema_id(ns.nspname)::object_id as id,
        acl.*
    from
        pg_namespace                                                                          ns
        join lateral ( select
                           role_id(grantor::regrole::name) as grantor,
                           role_id(grantee::regrole::name) as grantee,
                           privilege_type,
                           is_grantable,
                           ns.nspacl is null               as "default"
                       from
                           aclexplode(coalesce(ns.nspacl, acldefault('n', ns.nspowner))) ) as acl on true
-- language
    union all
    select
        language_id(l.lanname)::object_id as id,
        acl.*
    from
        pg_language                                                                         l
        join lateral ( select
                           role_id(grantor::regrole::name) as grantor,
                           role_id(grantee::regrole::name) as grantee,
                           privilege_type,
                           is_grantable,
                           l.lanacl is null                as "default"
                       from
                           aclexplode(coalesce(l.lanacl, acldefault('l', l.lanowner))) ) as acl on true
-- fdw
    union all
    select
        foreign_data_wrapper_id(fdw.fdwname)::object_id as id,
        acl.*
    from
        pg_foreign_data_wrapper                                                                 fdw
        join lateral ( select
                           role_id(grantor::regrole::name) as grantor,
                           role_id(grantee::regrole::name) as grantee,
                           privilege_type,
                           is_grantable,
                           fdw.fdwacl is null              as "default"
                       from
                           aclexplode(coalesce(fdw.fdwacl, acldefault('F', fdw.fdwowner))) ) as acl on true
-- fdw server
    union all
    select
        foreign_server_id(f.srvname)::object_id as id,
        acl.*
    from
        pg_catalog.pg_foreign_server                                                        f
        join lateral ( select
                           role_id(grantor::regrole::name) as grantor,
                           role_id(grantee::regrole::name) as grantee,
                           privilege_type,
                           is_grantable,
                           f.srvacl is null                as "default"
                       from
                           aclexplode(coalesce(f.srvacl, acldefault('S', f.srvowner))) ) as acl on true
-- TODO: database, tablespace, largeobject (pending corresponding types)
;


    -- This is not perfect because of the potential pre-existing functions,
    -- but there's just so many functions there in `meta`
    for rec in select callable.*, (callable_function) is distinct from null as function
               from
                   callable
                   natural left join callable_function
                   natural left join callable_procedure
               where
                   schema_name = schema
        loop
            execute format('alter %4$s %2$I.%1$I(%3$s) set search_path to public, %2$I', rec.name, schema,
                           (select string_agg(p, ',') from unnest(rec.type_sig) t(p)),
                           case
                               when rec.function then 'function'
                               else 'procedure'
                               end
                    );
        end loop;

    create function create_remote_meta(
    schema regnamespace,
    remote_schema name,
    server name,
    materialize boolean default false,
    dblink_preferred boolean default false) returns void
    language plpgsql
as
$create_remote_meta$
declare
    import_schema   regnamespace := schema;
    old_search_path text         := current_setting('search_path');
begin
    if materialize then
        declare
            import_schema_name text := '_' || (select nspname from pg_namespace where oid = schema) || '_foreign';
        begin
            execute format('create schema %I', import_schema_name);
            import_schema := import_schema_name::regnamespace;
        end;
    end if;

    perform from pg_foreign_server where srvname = server and not dblink_preferred;
    if not found then
        -- Attempt to see if this is a dblink connection
        if coalesce(server = any (dblink_get_connections()), false) then
            if not materialize then
                raise exception 'dblink remote meta can only be materialized' using detail = 'This is because dblink is query execution oriented';
            end if;
            perform set_config('search_path', import_schema || ',public', true);

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'schema');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'schema',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'cast');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'cast',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'cast_binary_coercible');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'cast_binary_coercible',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'cast_function');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'cast_function',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'cast_implicit');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'cast_implicit',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'cast_implicit_in_assignment');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'cast_implicit_in_assignment',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'cast_explicit');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'cast_explicit',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'operator');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'operator',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'sequence');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'sequence',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'sequence_minimum_value');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'sequence_minimum_value',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'sequence_start_value');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'sequence_start_value',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'sequence_maximum_value');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'sequence_maximum_value',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'sequence_cycle');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'sequence_cycle',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'sequence_increment');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'sequence_increment',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'sequence_type');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'sequence_type',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'sequence_cache');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'sequence_cache',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'sequence_table');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'sequence_table',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'relation_column');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'relation_column',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'relation_column_type');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'relation_column_type',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'relation_column_position');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'relation_column_position',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'relation_column_default');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'relation_column_default',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'relation_column_identity');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'relation_column_identity',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'relation_column_generated');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'relation_column_generated',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'relation_column_nullable');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'relation_column_nullable',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'relation');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'relation',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'table');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'table',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'table_permanent');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'table_permanent',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'table_temporary');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'table_temporary',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'table_partitioned');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'table_partitioned',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'table_rowsecurity');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'table_rowsecurity',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'table_forcerowsecurity');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'table_forcerowsecurity',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'view');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'view',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_function');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_function',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_procedure');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_procedure',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_return_type');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_return_type',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_owner');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_owner',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_argument_name');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_argument_name',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_argument_type');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_argument_type',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_argument_mode');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_argument_mode',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_argument_default');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_argument_default',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_immutable');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_immutable',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_volatile');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_volatile',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_stable');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_stable',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_aggregate');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_aggregate',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_window');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_window',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_parallel_safe');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_parallel_safe',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_parallel_unsafe');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_parallel_unsafe',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_parallel_restricted');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_parallel_restricted',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_security_definer');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_security_definer',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_security_invoker');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_security_invoker',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_language');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_language',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_acl');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_acl',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'callable_body');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'callable_body',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'trigger');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'trigger',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'role');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'role',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'role_superuser');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'role_superuser',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'role_inherit');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'role_inherit',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'role_create_role');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'role_create_role',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'role_create_db');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'role_create_db',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'role_can_login');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'role_can_login',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'role_replication');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'role_replication',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'role_connection_limit');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'role_connection_limit',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'role_valid_until');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'role_valid_until',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'role_setting');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'role_setting',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'policy');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'policy',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'policy_role');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'policy_role',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_non_null');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_non_null',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_non_null_column');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_non_null_column',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_check');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_check',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_check_expr');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_check_expr',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_unique');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_unique',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_unique_column');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_unique_column',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'extension');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'extension',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'foreign_data_wrapper');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'foreign_data_wrapper',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'foreign_server');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'foreign_server',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'foreign_table');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'foreign_table',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'foreign_column');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'foreign_column',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'relation_foreign_key');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'relation_foreign_key',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'relation_foreign_key_constraint');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'relation_foreign_key_constraint',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_foreign_key');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_foreign_key',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_foreign_key_update');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_foreign_key_update',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_foreign_key_delete');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_foreign_key_delete',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_foreign_key_match_full');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_foreign_key_match_full',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_foreign_key_match_partial');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_foreign_key_match_partial',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_foreign_key_match_simple');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_foreign_key_match_simple',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_foreign_key_references');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_foreign_key_references',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'constraint_relation_foreign_key_references_column');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'constraint_relation_foreign_key_references_column',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_basic');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_basic',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_array');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_array',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_composite');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_composite',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_composite_attribute');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_composite_attribute',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_composite_attribute_position');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_composite_attribute_position',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_composite_attribute_collation');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_composite_attribute_collation',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_domain');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_domain',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_enum');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_enum',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_enum_label');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_enum_label',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_pseudo');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_pseudo',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_range');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_range',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'type_multirange');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'type_multirange',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'index');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'index',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'index_relation');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'index_relation',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'index_unique');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'index_unique',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'index_unique_null_values_distinct');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'index_unique_null_values_distinct',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'index_primary_key');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'index_primary_key',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'index_unique_immediate');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'index_unique_immediate',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'index_replica_identity');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'index_replica_identity',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'index_attribute');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'index_attribute',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'index_partial');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'index_partial',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'language');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'language',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'language_internal');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'language_internal',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'language_trusted');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'language_trusted',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'language_handler');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'language_handler',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'language_inline_handler');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'language_inline_handler',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'language_validator');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'language_validator',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'language_acl');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'language_acl',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'comment');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'comment',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'dependency');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'dependency',  server, remote_schema, signature);
            end;

            declare
                signature text;
            begin
                select
                    string_agg(quote_ident(rc.name) || ' ' ||
                              -- FIXME: we are not quoting type name here (vv) because it may contain subscript for array types
                              (quote_ident(type.schema_name) || '.' || type.name), ', ' order by rcp.position)
                into signature
                from
                    omni_schema.relation_column                   rc
                    natural join omni_schema.relation_column_type rct
                    natural join omni_schema.relation_column_position rcp
                    inner join   omni_schema.type on type.id = type_id
                where
                    relation_id = omni_schema.relation_id('omni_schema', 'acl');

                execute format('create view %1$I as select * from public.dblink(%2$L, ''select * from %3$I.%1$I'') as t(%4$s)', 'acl',  server, remote_schema, signature);
            end;
            perform set_config('search_path', old_search_path, true);
        else
            raise exception 'no foreign server or dblink named % found', server;
        end if;
    else
        execute format(
                $import$
            import foreign schema %I limit to (
            "schema"
            ,
            "cast"
            ,
            "cast_binary_coercible"
            ,
            "cast_function"
            ,
            "cast_implicit"
            ,
            "cast_implicit_in_assignment"
            ,
            "cast_explicit"
            ,
            "operator"
            ,
            "sequence"
            ,
            "sequence_minimum_value"
            ,
            "sequence_start_value"
            ,
            "sequence_maximum_value"
            ,
            "sequence_cycle"
            ,
            "sequence_increment"
            ,
            "sequence_type"
            ,
            "sequence_cache"
            ,
            "sequence_table"
            ,
            "relation_column"
            ,
            "relation_column_type"
            ,
            "relation_column_position"
            ,
            "relation_column_default"
            ,
            "relation_column_identity"
            ,
            "relation_column_generated"
            ,
            "relation_column_nullable"
            ,
            "relation"
            ,
            "table"
            ,
            "table_permanent"
            ,
            "table_temporary"
            ,
            "table_partitioned"
            ,
            "table_rowsecurity"
            ,
            "table_forcerowsecurity"
            ,
            "view"
            ,
            "callable"
            ,
            "callable_function"
            ,
            "callable_procedure"
            ,
            "callable_return_type"
            ,
            "callable_owner"
            ,
            "callable_argument_name"
            ,
            "callable_argument_type"
            ,
            "callable_argument_mode"
            ,
            "callable_argument_default"
            ,
            "callable_immutable"
            ,
            "callable_volatile"
            ,
            "callable_stable"
            ,
            "callable_aggregate"
            ,
            "callable_window"
            ,
            "callable_parallel_safe"
            ,
            "callable_parallel_unsafe"
            ,
            "callable_parallel_restricted"
            ,
            "callable_security_definer"
            ,
            "callable_security_invoker"
            ,
            "callable_language"
            ,
            "callable_acl"
            ,
            "callable_body"
            ,
            "trigger"
            ,
            "role"
            ,
            "role_superuser"
            ,
            "role_inherit"
            ,
            "role_create_role"
            ,
            "role_create_db"
            ,
            "role_can_login"
            ,
            "role_replication"
            ,
            "role_connection_limit"
            ,
            "role_valid_until"
            ,
            "role_setting"
            ,
            "policy"
            ,
            "policy_role"
            ,
            "constraint_relation_non_null"
            ,
            "constraint_relation_non_null_column"
            ,
            "constraint_relation_check"
            ,
            "constraint_relation_check_expr"
            ,
            "constraint_relation_unique"
            ,
            "constraint_relation_unique_column"
            ,
            "extension"
            ,
            "foreign_data_wrapper"
            ,
            "foreign_server"
            ,
            "foreign_table"
            ,
            "foreign_column"
            ,
            "relation_foreign_key"
            ,
            "relation_foreign_key_constraint"
            ,
            "constraint_relation_foreign_key"
            ,
            "constraint_relation_foreign_key_update"
            ,
            "constraint_relation_foreign_key_delete"
            ,
            "constraint_relation_foreign_key_match_full"
            ,
            "constraint_relation_foreign_key_match_partial"
            ,
            "constraint_relation_foreign_key_match_simple"
            ,
            "constraint_relation_foreign_key_references"
            ,
            "constraint_relation_foreign_key_references_column"
            ,
            "type"
            ,
            "type_basic"
            ,
            "type_array"
            ,
            "type_composite"
            ,
            "type_composite_attribute"
            ,
            "type_composite_attribute_position"
            ,
            "type_composite_attribute_collation"
            ,
            "type_domain"
            ,
            "type_enum"
            ,
            "type_enum_label"
            ,
            "type_pseudo"
            ,
            "type_range"
            ,
            "type_multirange"
            ,
            "index"
            ,
            "index_relation"
            ,
            "index_unique"
            ,
            "index_unique_null_values_distinct"
            ,
            "index_primary_key"
            ,
            "index_unique_immediate"
            ,
            "index_replica_identity"
            ,
            "index_attribute"
            ,
            "index_partial"
            ,
            "language"
            ,
            "language_internal"
            ,
            "language_trusted"
            ,
            "language_handler"
            ,
            "language_inline_handler"
            ,
            "language_validator"
            ,
            "language_acl"
            ,
            "comment"
            ,
            "dependency"
            ,
            "acl"
           ) from server %I into %s
        $import$,
                remote_schema, server, import_schema);

    end if;
    if materialize then
        -- FIXME: hard-coded schema here
        perform omni_schema.materialize_meta(import_schema, schema);
    end if;
    return;
end;
$create_remote_meta$;

    create function materialize_meta(
    from_schema regnamespace,
    to_schema regnamespace) returns void
    language plpgsql
as
$create_remote_meta$
declare
    old_search_path text := current_setting('search_path');
begin
    perform set_config('search_path', to_schema || ',' || from_schema, true);

    create materialized view if not exists "schema" as
        select * from "schema" with no data;
    create materialized view if not exists "cast" as
        select * from "cast" with no data;
    create materialized view if not exists "cast_binary_coercible" as
        select * from "cast_binary_coercible" with no data;
    create materialized view if not exists "cast_function" as
        select * from "cast_function" with no data;
    create materialized view if not exists "cast_implicit" as
        select * from "cast_implicit" with no data;
    create materialized view if not exists "cast_implicit_in_assignment" as
        select * from "cast_implicit_in_assignment" with no data;
    create materialized view if not exists "cast_explicit" as
        select * from "cast_explicit" with no data;
    create materialized view if not exists "operator" as
        select * from "operator" with no data;
    create materialized view if not exists "sequence" as
        select * from "sequence" with no data;
    create materialized view if not exists "sequence_minimum_value" as
        select * from "sequence_minimum_value" with no data;
    create materialized view if not exists "sequence_start_value" as
        select * from "sequence_start_value" with no data;
    create materialized view if not exists "sequence_maximum_value" as
        select * from "sequence_maximum_value" with no data;
    create materialized view if not exists "sequence_cycle" as
        select * from "sequence_cycle" with no data;
    create materialized view if not exists "sequence_increment" as
        select * from "sequence_increment" with no data;
    create materialized view if not exists "sequence_type" as
        select * from "sequence_type" with no data;
    create materialized view if not exists "sequence_cache" as
        select * from "sequence_cache" with no data;
    create materialized view if not exists "sequence_table" as
        select * from "sequence_table" with no data;
    create materialized view if not exists "relation_column" as
        select * from "relation_column" with no data;
    create materialized view if not exists "relation_column_type" as
        select * from "relation_column_type" with no data;
    create materialized view if not exists "relation_column_position" as
        select * from "relation_column_position" with no data;
    create materialized view if not exists "relation_column_default" as
        select * from "relation_column_default" with no data;
    create materialized view if not exists "relation_column_identity" as
        select * from "relation_column_identity" with no data;
    create materialized view if not exists "relation_column_generated" as
        select * from "relation_column_generated" with no data;
    create materialized view if not exists "relation_column_nullable" as
        select * from "relation_column_nullable" with no data;
    create materialized view if not exists "relation" as
        select * from "relation" with no data;
    create materialized view if not exists "table" as
        select * from "table" with no data;
    create materialized view if not exists "table_permanent" as
        select * from "table_permanent" with no data;
    create materialized view if not exists "table_temporary" as
        select * from "table_temporary" with no data;
    create materialized view if not exists "table_partitioned" as
        select * from "table_partitioned" with no data;
    create materialized view if not exists "table_rowsecurity" as
        select * from "table_rowsecurity" with no data;
    create materialized view if not exists "table_forcerowsecurity" as
        select * from "table_forcerowsecurity" with no data;
    create materialized view if not exists "view" as
        select * from "view" with no data;
    create materialized view if not exists "callable" as
        select * from "callable" with no data;
    create materialized view if not exists "callable_function" as
        select * from "callable_function" with no data;
    create materialized view if not exists "callable_procedure" as
        select * from "callable_procedure" with no data;
    create materialized view if not exists "callable_return_type" as
        select * from "callable_return_type" with no data;
    create materialized view if not exists "callable_owner" as
        select * from "callable_owner" with no data;
    create materialized view if not exists "callable_argument_name" as
        select * from "callable_argument_name" with no data;
    create materialized view if not exists "callable_argument_type" as
        select * from "callable_argument_type" with no data;
    create materialized view if not exists "callable_argument_mode" as
        select * from "callable_argument_mode" with no data;
    create materialized view if not exists "callable_argument_default" as
        select * from "callable_argument_default" with no data;
    create materialized view if not exists "callable_immutable" as
        select * from "callable_immutable" with no data;
    create materialized view if not exists "callable_volatile" as
        select * from "callable_volatile" with no data;
    create materialized view if not exists "callable_stable" as
        select * from "callable_stable" with no data;
    create materialized view if not exists "callable_aggregate" as
        select * from "callable_aggregate" with no data;
    create materialized view if not exists "callable_window" as
        select * from "callable_window" with no data;
    create materialized view if not exists "callable_parallel_safe" as
        select * from "callable_parallel_safe" with no data;
    create materialized view if not exists "callable_parallel_unsafe" as
        select * from "callable_parallel_unsafe" with no data;
    create materialized view if not exists "callable_parallel_restricted" as
        select * from "callable_parallel_restricted" with no data;
    create materialized view if not exists "callable_security_definer" as
        select * from "callable_security_definer" with no data;
    create materialized view if not exists "callable_security_invoker" as
        select * from "callable_security_invoker" with no data;
    create materialized view if not exists "callable_language" as
        select * from "callable_language" with no data;
    create materialized view if not exists "callable_acl" as
        select * from "callable_acl" with no data;
    create materialized view if not exists "callable_body" as
        select * from "callable_body" with no data;
    create materialized view if not exists "trigger" as
        select * from "trigger" with no data;
    create materialized view if not exists "role" as
        select * from "role" with no data;
    create materialized view if not exists "role_superuser" as
        select * from "role_superuser" with no data;
    create materialized view if not exists "role_inherit" as
        select * from "role_inherit" with no data;
    create materialized view if not exists "role_create_role" as
        select * from "role_create_role" with no data;
    create materialized view if not exists "role_create_db" as
        select * from "role_create_db" with no data;
    create materialized view if not exists "role_can_login" as
        select * from "role_can_login" with no data;
    create materialized view if not exists "role_replication" as
        select * from "role_replication" with no data;
    create materialized view if not exists "role_connection_limit" as
        select * from "role_connection_limit" with no data;
    create materialized view if not exists "role_valid_until" as
        select * from "role_valid_until" with no data;
    create materialized view if not exists "role_setting" as
        select * from "role_setting" with no data;
    create materialized view if not exists "policy" as
        select * from "policy" with no data;
    create materialized view if not exists "policy_role" as
        select * from "policy_role" with no data;
    create materialized view if not exists "constraint_relation_non_null" as
        select * from "constraint_relation_non_null" with no data;
    create materialized view if not exists "constraint_relation_non_null_column" as
        select * from "constraint_relation_non_null_column" with no data;
    create materialized view if not exists "constraint_relation_check" as
        select * from "constraint_relation_check" with no data;
    create materialized view if not exists "constraint_relation_check_expr" as
        select * from "constraint_relation_check_expr" with no data;
    create materialized view if not exists "constraint_relation_unique" as
        select * from "constraint_relation_unique" with no data;
    create materialized view if not exists "constraint_relation_unique_column" as
        select * from "constraint_relation_unique_column" with no data;
    create materialized view if not exists "extension" as
        select * from "extension" with no data;
    create materialized view if not exists "foreign_data_wrapper" as
        select * from "foreign_data_wrapper" with no data;
    create materialized view if not exists "foreign_server" as
        select * from "foreign_server" with no data;
    create materialized view if not exists "foreign_table" as
        select * from "foreign_table" with no data;
    create materialized view if not exists "foreign_column" as
        select * from "foreign_column" with no data;
    create materialized view if not exists "relation_foreign_key" as
        select * from "relation_foreign_key" with no data;
    create materialized view if not exists "relation_foreign_key_constraint" as
        select * from "relation_foreign_key_constraint" with no data;
    create materialized view if not exists "constraint_relation_foreign_key" as
        select * from "constraint_relation_foreign_key" with no data;
    create materialized view if not exists "constraint_relation_foreign_key_update" as
        select * from "constraint_relation_foreign_key_update" with no data;
    create materialized view if not exists "constraint_relation_foreign_key_delete" as
        select * from "constraint_relation_foreign_key_delete" with no data;
    create materialized view if not exists "constraint_relation_foreign_key_match_full" as
        select * from "constraint_relation_foreign_key_match_full" with no data;
    create materialized view if not exists "constraint_relation_foreign_key_match_partial" as
        select * from "constraint_relation_foreign_key_match_partial" with no data;
    create materialized view if not exists "constraint_relation_foreign_key_match_simple" as
        select * from "constraint_relation_foreign_key_match_simple" with no data;
    create materialized view if not exists "constraint_relation_foreign_key_references" as
        select * from "constraint_relation_foreign_key_references" with no data;
    create materialized view if not exists "constraint_relation_foreign_key_references_column" as
        select * from "constraint_relation_foreign_key_references_column" with no data;
    create materialized view if not exists "type" as
        select * from "type" with no data;
    create materialized view if not exists "type_basic" as
        select * from "type_basic" with no data;
    create materialized view if not exists "type_array" as
        select * from "type_array" with no data;
    create materialized view if not exists "type_composite" as
        select * from "type_composite" with no data;
    create materialized view if not exists "type_composite_attribute" as
        select * from "type_composite_attribute" with no data;
    create materialized view if not exists "type_composite_attribute_position" as
        select * from "type_composite_attribute_position" with no data;
    create materialized view if not exists "type_composite_attribute_collation" as
        select * from "type_composite_attribute_collation" with no data;
    create materialized view if not exists "type_domain" as
        select * from "type_domain" with no data;
    create materialized view if not exists "type_enum" as
        select * from "type_enum" with no data;
    create materialized view if not exists "type_enum_label" as
        select * from "type_enum_label" with no data;
    create materialized view if not exists "type_pseudo" as
        select * from "type_pseudo" with no data;
    create materialized view if not exists "type_range" as
        select * from "type_range" with no data;
    create materialized view if not exists "type_multirange" as
        select * from "type_multirange" with no data;
    create materialized view if not exists "index" as
        select * from "index" with no data;
    create materialized view if not exists "index_relation" as
        select * from "index_relation" with no data;
    create materialized view if not exists "index_unique" as
        select * from "index_unique" with no data;
    create materialized view if not exists "index_unique_null_values_distinct" as
        select * from "index_unique_null_values_distinct" with no data;
    create materialized view if not exists "index_primary_key" as
        select * from "index_primary_key" with no data;
    create materialized view if not exists "index_unique_immediate" as
        select * from "index_unique_immediate" with no data;
    create materialized view if not exists "index_replica_identity" as
        select * from "index_replica_identity" with no data;
    create materialized view if not exists "index_attribute" as
        select * from "index_attribute" with no data;
    create materialized view if not exists "index_partial" as
        select * from "index_partial" with no data;
    create materialized view if not exists "language" as
        select * from "language" with no data;
    create materialized view if not exists "language_internal" as
        select * from "language_internal" with no data;
    create materialized view if not exists "language_trusted" as
        select * from "language_trusted" with no data;
    create materialized view if not exists "language_handler" as
        select * from "language_handler" with no data;
    create materialized view if not exists "language_inline_handler" as
        select * from "language_inline_handler" with no data;
    create materialized view if not exists "language_validator" as
        select * from "language_validator" with no data;
    create materialized view if not exists "language_acl" as
        select * from "language_acl" with no data;
    create materialized view if not exists "comment" as
        select * from "comment" with no data;
    create materialized view if not exists "dependency" as
        select * from "dependency" with no data;
    create materialized view if not exists "acl" as
        select * from "acl" with no data;

    create or replace function refresh_meta() returns void
        language plpgsql
    as
    $refresh_meta$
    begin
        perform progress_report('Materializing schema...');
        refresh materialized view "schema";
        perform progress_report('Materializing cast...');
        refresh materialized view "cast";
        perform progress_report('Materializing cast_binary_coercible...');
        refresh materialized view "cast_binary_coercible";
        perform progress_report('Materializing cast_function...');
        refresh materialized view "cast_function";
        perform progress_report('Materializing cast_implicit...');
        refresh materialized view "cast_implicit";
        perform progress_report('Materializing cast_implicit_in_assignment...');
        refresh materialized view "cast_implicit_in_assignment";
        perform progress_report('Materializing cast_explicit...');
        refresh materialized view "cast_explicit";
        perform progress_report('Materializing operator...');
        refresh materialized view "operator";
        perform progress_report('Materializing sequence...');
        refresh materialized view "sequence";
        perform progress_report('Materializing sequence_minimum_value...');
        refresh materialized view "sequence_minimum_value";
        perform progress_report('Materializing sequence_start_value...');
        refresh materialized view "sequence_start_value";
        perform progress_report('Materializing sequence_maximum_value...');
        refresh materialized view "sequence_maximum_value";
        perform progress_report('Materializing sequence_cycle...');
        refresh materialized view "sequence_cycle";
        perform progress_report('Materializing sequence_increment...');
        refresh materialized view "sequence_increment";
        perform progress_report('Materializing sequence_type...');
        refresh materialized view "sequence_type";
        perform progress_report('Materializing sequence_cache...');
        refresh materialized view "sequence_cache";
        perform progress_report('Materializing sequence_table...');
        refresh materialized view "sequence_table";
        perform progress_report('Materializing relation_column...');
        refresh materialized view "relation_column";
        perform progress_report('Materializing relation_column_type...');
        refresh materialized view "relation_column_type";
        perform progress_report('Materializing relation_column_position...');
        refresh materialized view "relation_column_position";
        perform progress_report('Materializing relation_column_default...');
        refresh materialized view "relation_column_default";
        perform progress_report('Materializing relation_column_identity...');
        refresh materialized view "relation_column_identity";
        perform progress_report('Materializing relation_column_generated...');
        refresh materialized view "relation_column_generated";
        perform progress_report('Materializing relation_column_nullable...');
        refresh materialized view "relation_column_nullable";
        perform progress_report('Materializing relation...');
        refresh materialized view "relation";
        perform progress_report('Materializing table...');
        refresh materialized view "table";
        perform progress_report('Materializing table_permanent...');
        refresh materialized view "table_permanent";
        perform progress_report('Materializing table_temporary...');
        refresh materialized view "table_temporary";
        perform progress_report('Materializing table_partitioned...');
        refresh materialized view "table_partitioned";
        perform progress_report('Materializing table_rowsecurity...');
        refresh materialized view "table_rowsecurity";
        perform progress_report('Materializing table_forcerowsecurity...');
        refresh materialized view "table_forcerowsecurity";
        perform progress_report('Materializing view...');
        refresh materialized view "view";
        perform progress_report('Materializing callable...');
        refresh materialized view "callable";
        perform progress_report('Materializing callable_function...');
        refresh materialized view "callable_function";
        perform progress_report('Materializing callable_procedure...');
        refresh materialized view "callable_procedure";
        perform progress_report('Materializing callable_return_type...');
        refresh materialized view "callable_return_type";
        perform progress_report('Materializing callable_owner...');
        refresh materialized view "callable_owner";
        perform progress_report('Materializing callable_argument_name...');
        refresh materialized view "callable_argument_name";
        perform progress_report('Materializing callable_argument_type...');
        refresh materialized view "callable_argument_type";
        perform progress_report('Materializing callable_argument_mode...');
        refresh materialized view "callable_argument_mode";
        perform progress_report('Materializing callable_argument_default...');
        refresh materialized view "callable_argument_default";
        perform progress_report('Materializing callable_immutable...');
        refresh materialized view "callable_immutable";
        perform progress_report('Materializing callable_volatile...');
        refresh materialized view "callable_volatile";
        perform progress_report('Materializing callable_stable...');
        refresh materialized view "callable_stable";
        perform progress_report('Materializing callable_aggregate...');
        refresh materialized view "callable_aggregate";
        perform progress_report('Materializing callable_window...');
        refresh materialized view "callable_window";
        perform progress_report('Materializing callable_parallel_safe...');
        refresh materialized view "callable_parallel_safe";
        perform progress_report('Materializing callable_parallel_unsafe...');
        refresh materialized view "callable_parallel_unsafe";
        perform progress_report('Materializing callable_parallel_restricted...');
        refresh materialized view "callable_parallel_restricted";
        perform progress_report('Materializing callable_security_definer...');
        refresh materialized view "callable_security_definer";
        perform progress_report('Materializing callable_security_invoker...');
        refresh materialized view "callable_security_invoker";
        perform progress_report('Materializing callable_language...');
        refresh materialized view "callable_language";
        perform progress_report('Materializing callable_acl...');
        refresh materialized view "callable_acl";
        perform progress_report('Materializing callable_body...');
        refresh materialized view "callable_body";
        perform progress_report('Materializing trigger...');
        refresh materialized view "trigger";
        perform progress_report('Materializing role...');
        refresh materialized view "role";
        perform progress_report('Materializing role_superuser...');
        refresh materialized view "role_superuser";
        perform progress_report('Materializing role_inherit...');
        refresh materialized view "role_inherit";
        perform progress_report('Materializing role_create_role...');
        refresh materialized view "role_create_role";
        perform progress_report('Materializing role_create_db...');
        refresh materialized view "role_create_db";
        perform progress_report('Materializing role_can_login...');
        refresh materialized view "role_can_login";
        perform progress_report('Materializing role_replication...');
        refresh materialized view "role_replication";
        perform progress_report('Materializing role_connection_limit...');
        refresh materialized view "role_connection_limit";
        perform progress_report('Materializing role_valid_until...');
        refresh materialized view "role_valid_until";
        perform progress_report('Materializing role_setting...');
        refresh materialized view "role_setting";
        perform progress_report('Materializing policy...');
        refresh materialized view "policy";
        perform progress_report('Materializing policy_role...');
        refresh materialized view "policy_role";
        perform progress_report('Materializing constraint_relation_non_null...');
        refresh materialized view "constraint_relation_non_null";
        perform progress_report('Materializing constraint_relation_non_null_column...');
        refresh materialized view "constraint_relation_non_null_column";
        perform progress_report('Materializing constraint_relation_check...');
        refresh materialized view "constraint_relation_check";
        perform progress_report('Materializing constraint_relation_check_expr...');
        refresh materialized view "constraint_relation_check_expr";
        perform progress_report('Materializing constraint_relation_unique...');
        refresh materialized view "constraint_relation_unique";
        perform progress_report('Materializing constraint_relation_unique_column...');
        refresh materialized view "constraint_relation_unique_column";
        perform progress_report('Materializing extension...');
        refresh materialized view "extension";
        perform progress_report('Materializing foreign_data_wrapper...');
        refresh materialized view "foreign_data_wrapper";
        perform progress_report('Materializing foreign_server...');
        refresh materialized view "foreign_server";
        perform progress_report('Materializing foreign_table...');
        refresh materialized view "foreign_table";
        perform progress_report('Materializing foreign_column...');
        refresh materialized view "foreign_column";
        perform progress_report('Materializing relation_foreign_key...');
        refresh materialized view "relation_foreign_key";
        perform progress_report('Materializing relation_foreign_key_constraint...');
        refresh materialized view "relation_foreign_key_constraint";
        perform progress_report('Materializing constraint_relation_foreign_key...');
        refresh materialized view "constraint_relation_foreign_key";
        perform progress_report('Materializing constraint_relation_foreign_key_update...');
        refresh materialized view "constraint_relation_foreign_key_update";
        perform progress_report('Materializing constraint_relation_foreign_key_delete...');
        refresh materialized view "constraint_relation_foreign_key_delete";
        perform progress_report('Materializing constraint_relation_foreign_key_match_full...');
        refresh materialized view "constraint_relation_foreign_key_match_full";
        perform progress_report('Materializing constraint_relation_foreign_key_match_partial...');
        refresh materialized view "constraint_relation_foreign_key_match_partial";
        perform progress_report('Materializing constraint_relation_foreign_key_match_simple...');
        refresh materialized view "constraint_relation_foreign_key_match_simple";
        perform progress_report('Materializing constraint_relation_foreign_key_references...');
        refresh materialized view "constraint_relation_foreign_key_references";
        perform progress_report('Materializing constraint_relation_foreign_key_references_column...');
        refresh materialized view "constraint_relation_foreign_key_references_column";
        perform progress_report('Materializing type...');
        refresh materialized view "type";
        perform progress_report('Materializing type_basic...');
        refresh materialized view "type_basic";
        perform progress_report('Materializing type_array...');
        refresh materialized view "type_array";
        perform progress_report('Materializing type_composite...');
        refresh materialized view "type_composite";
        perform progress_report('Materializing type_composite_attribute...');
        refresh materialized view "type_composite_attribute";
        perform progress_report('Materializing type_composite_attribute_position...');
        refresh materialized view "type_composite_attribute_position";
        perform progress_report('Materializing type_composite_attribute_collation...');
        refresh materialized view "type_composite_attribute_collation";
        perform progress_report('Materializing type_domain...');
        refresh materialized view "type_domain";
        perform progress_report('Materializing type_enum...');
        refresh materialized view "type_enum";
        perform progress_report('Materializing type_enum_label...');
        refresh materialized view "type_enum_label";
        perform progress_report('Materializing type_pseudo...');
        refresh materialized view "type_pseudo";
        perform progress_report('Materializing type_range...');
        refresh materialized view "type_range";
        perform progress_report('Materializing type_multirange...');
        refresh materialized view "type_multirange";
        perform progress_report('Materializing index...');
        refresh materialized view "index";
        perform progress_report('Materializing index_relation...');
        refresh materialized view "index_relation";
        perform progress_report('Materializing index_unique...');
        refresh materialized view "index_unique";
        perform progress_report('Materializing index_unique_null_values_distinct...');
        refresh materialized view "index_unique_null_values_distinct";
        perform progress_report('Materializing index_primary_key...');
        refresh materialized view "index_primary_key";
        perform progress_report('Materializing index_unique_immediate...');
        refresh materialized view "index_unique_immediate";
        perform progress_report('Materializing index_replica_identity...');
        refresh materialized view "index_replica_identity";
        perform progress_report('Materializing index_attribute...');
        refresh materialized view "index_attribute";
        perform progress_report('Materializing index_partial...');
        refresh materialized view "index_partial";
        perform progress_report('Materializing language...');
        refresh materialized view "language";
        perform progress_report('Materializing language_internal...');
        refresh materialized view "language_internal";
        perform progress_report('Materializing language_trusted...');
        refresh materialized view "language_trusted";
        perform progress_report('Materializing language_handler...');
        refresh materialized view "language_handler";
        perform progress_report('Materializing language_inline_handler...');
        refresh materialized view "language_inline_handler";
        perform progress_report('Materializing language_validator...');
        refresh materialized view "language_validator";
        perform progress_report('Materializing language_acl...');
        refresh materialized view "language_acl";
        perform progress_report('Materializing comment...');
        refresh materialized view "comment";
        perform progress_report('Materializing dependency...');
        refresh materialized view "dependency";
        perform progress_report('Materializing acl...');
        refresh materialized view "acl";
    end;
    $refresh_meta$;
    -- TODO: remove hard-coded dependency on omni_schema
    execute format('alter function refresh_meta() set search_path = %s', to_schema || ',' || from_schema || ',omni_schema');

    perform refresh_meta();

    perform set_config('search_path', old_search_path, true);
    return;
end;
$create_remote_meta$;

    create function create_meta_diff(
    baseline regnamespace,
    changed regnamespace,
    schema regnamespace
) returns void
    language plpgsql
as
$create_meta_diff$
declare
    old_search_path text := current_setting('search_path');
    types text[] :=
        array[
            'schema'
            ,
            'cast'
            ,
            'cast_binary_coercible'
            ,
            'cast_function'
            ,
            'cast_implicit'
            ,
            'cast_implicit_in_assignment'
            ,
            'cast_explicit'
            ,
            'operator'
            ,
            'sequence'
            ,
            'sequence_minimum_value'
            ,
            'sequence_start_value'
            ,
            'sequence_maximum_value'
            ,
            'sequence_cycle'
            ,
            'sequence_increment'
            ,
            'sequence_type'
            ,
            'sequence_cache'
            ,
            'sequence_table'
            ,
            'relation_column'
            ,
            'relation_column_type'
            ,
            'relation_column_position'
            ,
            'relation_column_default'
            ,
            'relation_column_identity'
            ,
            'relation_column_generated'
            ,
            'relation_column_nullable'
            ,
            'relation'
            ,
            'table'
            ,
            'table_permanent'
            ,
            'table_temporary'
            ,
            'table_partitioned'
            ,
            'table_rowsecurity'
            ,
            'table_forcerowsecurity'
            ,
            'view'
            ,
            'callable'
            ,
            'callable_function'
            ,
            'callable_procedure'
            ,
            'callable_return_type'
            ,
            'callable_owner'
            ,
            'callable_argument_name'
            ,
            'callable_argument_type'
            ,
            'callable_argument_mode'
            ,
            'callable_argument_default'
            ,
            'callable_immutable'
            ,
            'callable_volatile'
            ,
            'callable_stable'
            ,
            'callable_aggregate'
            ,
            'callable_window'
            ,
            'callable_parallel_safe'
            ,
            'callable_parallel_unsafe'
            ,
            'callable_parallel_restricted'
            ,
            'callable_security_definer'
            ,
            'callable_security_invoker'
            ,
            'callable_language'
            ,
            'callable_acl'
            ,
            'callable_body'
            ,
            'trigger'
            ,
            'role'
            ,
            'role_superuser'
            ,
            'role_inherit'
            ,
            'role_create_role'
            ,
            'role_create_db'
            ,
            'role_can_login'
            ,
            'role_replication'
            ,
            'role_connection_limit'
            ,
            'role_valid_until'
            ,
            'role_setting'
            ,
            'policy'
            ,
            'policy_role'
            ,
            'constraint_relation_non_null'
            ,
            'constraint_relation_non_null_column'
            ,
            'constraint_relation_check'
            ,
            'constraint_relation_check_expr'
            ,
            'constraint_relation_unique'
            ,
            'constraint_relation_unique_column'
            ,
            'extension'
            ,
            'foreign_data_wrapper'
            ,
            'foreign_server'
            ,
            'foreign_table'
            ,
            'foreign_column'
            ,
            'relation_foreign_key'
            ,
            'relation_foreign_key_constraint'
            ,
            'constraint_relation_foreign_key'
            ,
            'constraint_relation_foreign_key_update'
            ,
            'constraint_relation_foreign_key_delete'
            ,
            'constraint_relation_foreign_key_match_full'
            ,
            'constraint_relation_foreign_key_match_partial'
            ,
            'constraint_relation_foreign_key_match_simple'
            ,
            'constraint_relation_foreign_key_references'
            ,
            'constraint_relation_foreign_key_references_column'
            ,
            'type'
            ,
            'type_basic'
            ,
            'type_array'
            ,
            'type_composite'
            ,
            'type_composite_attribute'
            ,
            'type_composite_attribute_position'
            ,
            'type_composite_attribute_collation'
            ,
            'type_domain'
            ,
            'type_enum'
            ,
            'type_enum_label'
            ,
            'type_pseudo'
            ,
            'type_range'
            ,
            'type_multirange'
            ,
            'index'
            ,
            'index_relation'
            ,
            'index_unique'
            ,
            'index_unique_null_values_distinct'
            ,
            'index_primary_key'
            ,
            'index_unique_immediate'
            ,
            'index_replica_identity'
            ,
            'index_attribute'
            ,
            'index_partial'
            ,
            'language'
            ,
            'language_internal'
            ,
            'language_trusted'
            ,
            'language_handler'
            ,
            'language_inline_handler'
            ,
            'language_validator'
            ,
            'language_acl'
            ,
            'comment'
            ,
            'dependency'
            ,
            'acl'
            ];
    type  text;
begin
    -- FIXME: hard-coded omni_schema; not great (instantiation may change this)
    -- We do this so we can get the comparison operator
    perform set_config('search_path', schema::text || ',' || changed::text || ',omni_schema', true);

    foreach type in array types
        loop
            execute format($sql$
    create or replace view added_%3$s as select b.* from %2$s.%3$I b
        natural left outer join %1$s.%3$I a
        where a is not distinct from null
    $sql$, baseline, changed, type);

            execute format($sql$
    create or replace view removed_%3$s as select b.* from %1$s.%3$I b
        natural left outer join %2$s.%3$I a
        where a is not distinct from null
    $sql$, baseline, changed, type);

        end loop;

    perform set_config('search_path', old_search_path, true);

    return;
end;
$create_meta_diff$;


end;
$instantiate_meta$;


select instantiate_meta(schema => 'omni_schema');
;



create function instantiate_evolution(schema name) returns void
    language plpgsql
as
$instantiate_evolution$
declare
    -- Include omni_schema for other functionality
    -- Include omni_polyfill to ensure we have uuidv7
    -- Include omni_vfs to access file systems
    -- Include public for dblink
    search_path text := 'omni_schema,omni_vfs,omni_polyfill,public,pg_catalog';
begin
    if schema != 'omni_schema' then
        search_path := schema || ',' || search_path;
    end if;
    perform set_config('search_path', search_path, true);

    create domain revision_id as uuid check ( uuid_extract_version(value) = 7);

    create function assemble_schema_revision(fs anyelement, revisions_path text, revision revision_id,
                                         db_conninfo text,
                                         rollback bool default true
)
    returns table
            (
                execution_position  int,
                migration_filename  text,
                migration_statement text,
                execution_error     text
            )
    language plpgsql
as
$$
declare
    db_sources jsonb;
    rec        record;
begin
    ---- Install omni_schema in the target schema to get its meta
    perform dblink(db_conninfo, format('create extension if not exists omni_schema version %L cascade',
                                       (select extversion from pg_extension where extname = 'omni_schema')));

    perform dblink(db_conninfo,
                   format(
                           'create table if not exists omni_schema.deployed_revision as select %L::omni_schema.revision_id as revision',
                           revision::text));


    db_sources =
            omni_yaml.to_json(convert_from(
                    omni_vfs.read(fs, revisions_path || '/' || revision || '/sources.yaml'), 'utf8'));

    for rec in with
                   db_fs as materialized (select omni_vfs.table_fs(revision::text) db_fs)
               select
                   db_fs.db_fs,
                   key,
                   decode(value #>> '{}', 'base64') as decoded
               from
                   db_fs
                   inner join (select * from jsonb_each(db_sources) as t(key, value)) files on true
        loop
            perform omni_vfs.write(rec.db_fs, rec.key, rec.decoded, create_file => true);
        end loop;

    perform from pg_tables where tablename = 'assembly_report' and schemaname like 'pg_temp_%';
    if found then
        drop table assembly_report;
    end if;
    create temporary table if not exists assembly_report on commit drop as
        (select *
         from
             assemble_schema(db_conninfo, omni_vfs.table_fs(revision::text), ''));

    for rec in select * from assembly_report
        loop
            execution_position := rec.execution_position;
            migration_filename := rec.migration_filename;
            migration_statement := rec.migration_statement;
            execution_error := rec.execution_error;
            return next;
            execution_position := null;
            migration_filename := null;
            migration_statement := null;
            execution_error := null;
        end loop;

    if rollback then
        raise exception 'assemble_schema_revision_done';
    end if;

    return;

exception
    when others then
        if sqlerrm = 'assemble_schema_revision_done' then
            return;
        else
            ---- re-raise otherwise
            raise;
        end if;
end;
$$;
    execute format('alter function assemble_schema_revision set search_path = %s', search_path);

    create function progress_report(message text) returns text 
language sql
as $$
        select omni_cloudevents.publish(
            omni_cloudevents.cloudevent(
                id => gen_random_uuid(),
                source => format('psql://%s/%s', (select system_identifier from pg_control_system()), current_database()),
                type => 'org.omnigres.omni_schema.progress_report.v1',
                data => message
            )
        );
$$;

create function capture_schema_revision(fs anyelement, source_path text, revisions_path text,
                                        rollback bool default true,
                                        parents revision_id[] default null) returns revision_id
    language plpgsql
as
$$
declare
    revision          revision_id       := uuidv7();
    host              text              := ' host=' ||
                                           current_setting('unix_socket_directories') || ' port=' ||
                                           current_setting('port') ||
                                           ' user=' || current_user;
    self_conn         text              := 'dbname=' || current_database() || host;
    revision_database text              := revision;
    revision_conninfo text              := 'dbname=' || revision_database || host;
    parent_databases  text[];
    parent_conninfo   text[];
    has_parents       bool              := coalesce(cardinality(parents), 0) > 0;
    revision_parents  revision_id[]     := coalesce(parents, '{}'); -- TODO: do we actually need to allow overriding?
    metadata          json;
    sources           jsonb;
    diffs             jsonb             := '{}';
    old_search_path   text;
    rec               record;
    --- We write to this fs first to ensure errors during capture don't leave
    --- us with a partially written capture
    transactional_fs  omni_vfs.table_fs := omni_vfs.table_fs(revision::text);
    captured_revisions jsonb;
begin
    --- 0. Prepare parents

    if parents is null then
        -- Find revisions that are not parents to any other revisions
        select
            coalesce(array_agg(r.revision), '{}')
        into revision_parents
        from
            schema_revisions(fs, revisions_path, leafs_only => true) r;
        has_parents := coalesce(cardinality(revision_parents), 0) > 0;
    end if;

    --- 1. Assemble current revision

    ---- Prepare the database in which we'll be assembling it. Make it a template database
    ---- so that services know it is not an operational database
    perform progress_report(format('Creating revision database %s', revision_database));
    perform dblink(self_conn, format('create database %I', revision_database));
    update pg_database set datistemplate = true where datname = revision_database;

    ---- Install omni_schema in the target schema to get its meta
    perform dblink(revision_conninfo, format('create extension omni_schema version %L cascade',
                                             (select extversion from pg_extension where extname = 'omni_schema')));

    if not has_parents then
        ---- Create a foreign server to connect to the revision
        execute format(
                $sql$create server %1$I foreign data wrapper postgres_fdw options(host %2$L, dbname %4$L, port %3$L)$sql$,
                revision_database || '_empty', current_setting('unix_socket_directories'), current_setting('port'),
                revision_database);
        execute format('create user mapping for %1$I server %2$I options (user %1$L)',
                       current_user, revision_database || '_empty');
        ---- Define remote meta
        perform progress_report('Creating empty revision schema since there are no parents revision');
        execute format('create schema %I', revision || '_empty');
        perform create_remote_meta((revision::text || '_empty')::regnamespace, 'omni_schema'::name,
                                   revision::text || '_empty'::name, materialize => true);

        -- "Freeze" empty meta for further comparison. This is done by issuing a single
        -- query that would put the connection into the repeatable read mode (as per postgres_fdw
        -- implementation), letting us see this snapshot on this connection.
        execute format('select * from %I.relation', revision || '_empty');
    end if;

    perform dblink(revision_conninfo,
                   format(
                           'create table omni_schema.deployed_revision as select %L::omni_schema.revision_id as revision',
                           revision::text));

    perform progress_report('Assembling schema for revision');
    perform from assemble_schema(revision_conninfo, fs, source_path) where execution_error is not null;
    if found then
        raise exception 'New revision cannot be assembled due to errors' using hint = 'Run assemble_schema to see errors';
    end if;

    --- 2. Assemble parents (if any)

    declare
        current_parent  revision_id;
        parent_conninfo text;
        parent_sources  jsonb;
        rec             record;
    begin
        foreach current_parent in array revision_parents
            loop

                perform progress_report(format('Assembling parent revision %s', current_parent::text));
            ---- Prepare the database in which we'll be assembling it. Make it a template database
            ---- so that services know it is not an operational database
                perform dblink(self_conn, format('drop database if exists %I', current_parent::text));
                perform dblink(self_conn, format('create database %I', current_parent::text));
                update pg_database set datistemplate = true where datname = current_parent::text;

                parent_conninfo := 'dbname=' || current_parent || host;

                perform
                from
                    assemble_schema_revision(fs, revisions_path, current_parent, parent_conninfo)
                where
                    execution_error is not null;
                if found then
                    raise exception 'Parent % cannot be assembled due to errors', current_parent
                        using hint = 'Run assemble_schema to see errors';
                end if;
                update pg_database set datistemplate = false where datname = current_parent::text;
            end loop;
    end;

    --- 3. Write metadata
    perform progress_report(format('Writing metadata to %s', revision || '/metadata.yaml'));
    metadata := json_build_object('parents', revision_parents);
    perform omni_vfs.write(transactional_fs, revision || '/metadata.yaml',
                           convert_to(omni_yaml.to_yaml(metadata), 'utf8'), create_file => true);

    --- 4. Capture current source

    ---- FIXME: saving to YAML for now just to make _a_ decision
    perform progress_report(format('Writing sources to %s', revision || '/sources.yaml'));
    select
        jsonb_object_agg(name, encode(omni_vfs.read(fs, source_path || '/' || name), 'base64'))
    into sources
    from
        omni_vfs.list_recursively(fs, source_path)
    where
        kind = 'file';

    perform omni_vfs.write(transactional_fs, revision || '/sources.yaml',
                           convert_to(omni_yaml.to_yaml(sources::json), 'utf8'), create_file => true);

    --- 5. Capture diffs

    ---- Create a foreign server to connect to the revision
    execute format(
            $sql$create server %1$I foreign data wrapper postgres_fdw options(host %2$L, dbname %1$L, port %3$L)$sql$,
            revision_database, current_setting('unix_socket_directories'), current_setting('port'));
    execute format('create user mapping for %1$I server %2$I options (user %1$L)',
                   current_user, revision_database);

    ---- Define remote meta
    perform progress_report(format('Creating %s meta schema for diff', revision));
    execute format('create schema %I', revision);
    perform create_remote_meta(revision::text::regnamespace, 'omni_schema'::name, revision::text::name,
                               materialize => true);

    --- If no parents, compare with the empty version
    if not has_parents then

        execute format('create schema %I', revision || '_diff');
        perform create_meta_diff((revision || '_empty')::regnamespace, revision::text::regnamespace,
                                 (revision || '_diff')::regnamespace);

        old_search_path := current_setting('search_path');
        perform set_config('search_path',
                           (revision::text || '_diff')::regnamespace || ',' || revision::text::regnamespace || ',' ||
                           old_search_path, true);

        perform progress_report('Diffing schema...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_schema}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_schema")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_schema}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_schema")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing cast...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_cast")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_cast")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing cast_binary_coercible...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast_binary_coercible}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_cast_binary_coercible")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast_binary_coercible}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_cast_binary_coercible")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing cast_function...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast_function}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_cast_function")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast_function}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_cast_function")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing cast_implicit...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast_implicit}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_cast_implicit")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast_implicit}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_cast_implicit")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing cast_implicit_in_assignment...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast_implicit_in_assignment}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_cast_implicit_in_assignment")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast_implicit_in_assignment}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_cast_implicit_in_assignment")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing cast_explicit...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast_explicit}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_cast_explicit")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast_explicit}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_cast_explicit")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing operator...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_operator}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_operator")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_operator}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_operator")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing sequence...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_sequence")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_sequence")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing sequence_minimum_value...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_minimum_value}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_sequence_minimum_value")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_minimum_value}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_sequence_minimum_value")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing sequence_start_value...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_start_value}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_sequence_start_value")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_start_value}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_sequence_start_value")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing sequence_maximum_value...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_maximum_value}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_sequence_maximum_value")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_maximum_value}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_sequence_maximum_value")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing sequence_cycle...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_cycle}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_sequence_cycle")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_cycle}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_sequence_cycle")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing sequence_increment...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_increment}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_sequence_increment")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_increment}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_sequence_increment")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing sequence_type...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_type}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_sequence_type")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_type}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_sequence_type")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing sequence_cache...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_cache}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_sequence_cache")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_cache}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_sequence_cache")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing sequence_table...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_table}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_sequence_table")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_table}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_sequence_table")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing relation_column...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_relation_column")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_relation_column")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing relation_column_type...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_type}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_relation_column_type")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_type}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_relation_column_type")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing relation_column_position...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_position}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_relation_column_position")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_position}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_relation_column_position")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing relation_column_default...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_default}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_relation_column_default")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_default}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_relation_column_default")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing relation_column_identity...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_identity}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_relation_column_identity")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_identity}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_relation_column_identity")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing relation_column_generated...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_generated}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_relation_column_generated")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_generated}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_relation_column_generated")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing relation_column_nullable...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_nullable}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_relation_column_nullable")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_nullable}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_relation_column_nullable")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing relation...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_relation")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_relation")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing table...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_table")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_table")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing table_permanent...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table_permanent}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_table_permanent")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table_permanent}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_table_permanent")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing table_temporary...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table_temporary}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_table_temporary")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table_temporary}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_table_temporary")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing table_partitioned...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table_partitioned}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_table_partitioned")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table_partitioned}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_table_partitioned")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing table_rowsecurity...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table_rowsecurity}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_table_rowsecurity")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table_rowsecurity}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_table_rowsecurity")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing table_forcerowsecurity...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table_forcerowsecurity}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_table_forcerowsecurity")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table_forcerowsecurity}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_table_forcerowsecurity")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing view...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_view}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_view")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_view}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_view")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_function...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_function}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_function")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_function}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_function")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_procedure...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_procedure}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_procedure")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_procedure}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_procedure")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_return_type...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_return_type}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_return_type")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_return_type}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_return_type")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_owner...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_owner}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_owner")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_owner}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_owner")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_argument_name...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_argument_name}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_argument_name")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_argument_name}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_argument_name")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_argument_type...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_argument_type}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_argument_type")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_argument_type}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_argument_type")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_argument_mode...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_argument_mode}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_argument_mode")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_argument_mode}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_argument_mode")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_argument_default...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_argument_default}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_argument_default")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_argument_default}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_argument_default")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_immutable...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_immutable}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_immutable")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_immutable}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_immutable")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_volatile...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_volatile}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_volatile")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_volatile}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_volatile")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_stable...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_stable}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_stable")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_stable}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_stable")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_aggregate...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_aggregate}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_aggregate")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_aggregate}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_aggregate")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_window...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_window}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_window")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_window}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_window")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_parallel_safe...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_parallel_safe}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_parallel_safe")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_parallel_safe}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_parallel_safe")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_parallel_unsafe...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_parallel_unsafe}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_parallel_unsafe")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_parallel_unsafe}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_parallel_unsafe")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_parallel_restricted...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_parallel_restricted}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_parallel_restricted")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_parallel_restricted}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_parallel_restricted")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_security_definer...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_security_definer}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_security_definer")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_security_definer}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_security_definer")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_security_invoker...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_security_invoker}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_security_invoker")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_security_invoker}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_security_invoker")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_language...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_language}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_language")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_language}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_language")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_acl...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_acl}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_acl")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_acl}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_acl")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing callable_body...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_body}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_callable_body")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_body}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_callable_body")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing trigger...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_trigger}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_trigger")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_trigger}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_trigger")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing role...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_role")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_role")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing role_superuser...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_superuser}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_role_superuser")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_superuser}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_role_superuser")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing role_inherit...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_inherit}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_role_inherit")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_inherit}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_role_inherit")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing role_create_role...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_create_role}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_role_create_role")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_create_role}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_role_create_role")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing role_create_db...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_create_db}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_role_create_db")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_create_db}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_role_create_db")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing role_can_login...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_can_login}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_role_can_login")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_can_login}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_role_can_login")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing role_replication...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_replication}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_role_replication")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_replication}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_role_replication")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing role_connection_limit...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_connection_limit}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_role_connection_limit")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_connection_limit}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_role_connection_limit")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing role_valid_until...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_valid_until}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_role_valid_until")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_valid_until}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_role_valid_until")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing role_setting...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_setting}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_role_setting")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_setting}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_role_setting")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing policy...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_policy}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_policy")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_policy}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_policy")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing policy_role...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_policy_role}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_policy_role")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_policy_role}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_policy_role")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_non_null...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_non_null}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_non_null")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_non_null}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_non_null")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_non_null_column...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_non_null_column}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_non_null_column")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_non_null_column}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_non_null_column")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_check...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_check}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_check")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_check}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_check")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_check_expr...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_check_expr}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_check_expr")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_check_expr}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_check_expr")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_unique...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_unique}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_unique")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_unique}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_unique")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_unique_column...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_unique_column}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_unique_column")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_unique_column}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_unique_column")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing extension...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_extension}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_extension")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_extension}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_extension")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing foreign_data_wrapper...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_foreign_data_wrapper}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_foreign_data_wrapper")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_foreign_data_wrapper}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_foreign_data_wrapper")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing foreign_server...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_foreign_server}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_foreign_server")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_foreign_server}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_foreign_server")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing foreign_table...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_foreign_table}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_foreign_table")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_foreign_table}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_foreign_table")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing foreign_column...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_foreign_column}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_foreign_column")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_foreign_column}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_foreign_column")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing relation_foreign_key...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_foreign_key}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_relation_foreign_key")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_foreign_key}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_relation_foreign_key")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing relation_foreign_key_constraint...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_foreign_key_constraint}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_relation_foreign_key_constraint")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_foreign_key_constraint}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_relation_foreign_key_constraint")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_foreign_key...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_foreign_key")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_foreign_key")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_foreign_key_update...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_update}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_foreign_key_update")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_update}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_foreign_key_update")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_foreign_key_delete...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_delete}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_foreign_key_delete")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_delete}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_foreign_key_delete")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_foreign_key_match_full...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_match_full}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_foreign_key_match_full")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_match_full}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_foreign_key_match_full")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_foreign_key_match_partial...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_match_partial}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_foreign_key_match_partial")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_match_partial}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_foreign_key_match_partial")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_foreign_key_match_simple...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_match_simple}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_foreign_key_match_simple")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_match_simple}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_foreign_key_match_simple")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_foreign_key_references...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_references}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_foreign_key_references")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_references}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_foreign_key_references")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing constraint_relation_foreign_key_references_column...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_references_column}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_constraint_relation_foreign_key_references_column")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_references_column}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_constraint_relation_foreign_key_references_column")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_basic...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_basic}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_basic")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_basic}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_basic")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_array...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_array}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_array")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_array}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_array")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_composite...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_composite}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_composite")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_composite}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_composite")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_composite_attribute...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_composite_attribute}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_composite_attribute")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_composite_attribute}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_composite_attribute")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_composite_attribute_position...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_composite_attribute_position}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_composite_attribute_position")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_composite_attribute_position}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_composite_attribute_position")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_composite_attribute_collation...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_composite_attribute_collation}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_composite_attribute_collation")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_composite_attribute_collation}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_composite_attribute_collation")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_domain...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_domain}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_domain")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_domain}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_domain")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_enum...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_enum}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_enum")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_enum}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_enum")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_enum_label...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_enum_label}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_enum_label")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_enum_label}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_enum_label")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_pseudo...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_pseudo}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_pseudo")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_pseudo}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_pseudo")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_range...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_range}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_range")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_range}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_range")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing type_multirange...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_multirange}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_type_multirange")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_multirange}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_type_multirange")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing index...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_index")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_index")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing index_relation...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_relation}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_index_relation")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_relation}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_index_relation")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing index_unique...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_unique}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_index_unique")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_unique}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_index_unique")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing index_unique_null_values_distinct...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_unique_null_values_distinct}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_index_unique_null_values_distinct")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_unique_null_values_distinct}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_index_unique_null_values_distinct")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing index_primary_key...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_primary_key}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_index_primary_key")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_primary_key}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_index_primary_key")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing index_unique_immediate...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_unique_immediate}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_index_unique_immediate")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_unique_immediate}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_index_unique_immediate")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing index_replica_identity...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_replica_identity}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_index_replica_identity")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_replica_identity}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_index_replica_identity")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing index_attribute...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_attribute}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_index_attribute")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_attribute}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_index_attribute")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing index_partial...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_partial}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_index_partial")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_partial}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_index_partial")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing language...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_language")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_language")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing language_internal...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_internal}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_language_internal")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_internal}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_language_internal")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing language_trusted...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_trusted}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_language_trusted")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_trusted}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_language_trusted")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing language_handler...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_handler}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_language_handler")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_handler}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_language_handler")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing language_inline_handler...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_inline_handler}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_language_inline_handler")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_inline_handler}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_language_inline_handler")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing language_validator...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_validator}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_language_validator")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_validator}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_language_validator")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing language_acl...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_acl}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_language_acl")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_acl}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_language_acl")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing comment...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_comment}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_comment")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_comment}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_comment")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing dependency...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_dependency}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_dependency")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_dependency}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_dependency")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        perform progress_report('Diffing acl...');
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_acl}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "added_acl")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));
        diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_acl}',
                                             coalesce((with
                                                           t
                                                               as materialized (select * from "removed_acl")
                                                       select
                                                           jsonb_agg(to_jsonb(t))
                                                       from
                                                           t
                                                           left join dependency d on d.id = t.id::object_id and
                                                                                     omni_types.variant(d.dependent_on) =
                                                                                     'extension_id'::regtype
                                                       where
                                                           (d) is not distinct from null), 'null')));

        perform set_config('search_path', old_search_path, true);

        perform omni_vfs.write(transactional_fs, revision || '/diffs.yaml',
                               convert_to(omni_yaml.to_yaml(diffs::json), 'utf8'), create_file => true);

    else
        --- Otherwise, compare with each parent
        declare
            current_parent revision_id;
            diff_schema    text;
            any_diff       bool := false;
        begin

            foreach current_parent in array revision_parents
                loop

                    ---- Create a foreign server to connect to the parent
                    execute format(
                            $sql$create server %1$I foreign data wrapper postgres_fdw options(host %2$L, dbname %1$L, port %3$L)$sql$,
                            current_parent, current_setting('unix_socket_directories'), current_setting('port'));
                    execute format('create user mapping for %1$I server %2$I options (user %1$L)',
                                   current_user, current_parent);

                    ---- Define remote meta
                    execute format('create schema %I', current_parent);
                    perform create_remote_meta(current_parent::text::regnamespace, 'omni_schema'::name,
                                               current_parent::text::name, materialize => true);


                    diff_schema := current_parent::text || '_diff';
                    execute format('create schema %I', diff_schema);
                    perform create_meta_diff(current_parent::text::regnamespace, revision::text::regnamespace,
                                             diff_schema::regnamespace);

                    old_search_path := current_setting('search_path');
                    perform set_config('search_path',
                                       diff_schema::regnamespace || ',' || old_search_path, true);

                    perform progress_report('Diffing schema...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_schema}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_schema")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_schema}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_schema")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing cast...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_cast")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_cast")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing cast_binary_coercible...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast_binary_coercible}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_cast_binary_coercible")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast_binary_coercible}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_cast_binary_coercible")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing cast_function...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast_function}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_cast_function")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast_function}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_cast_function")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing cast_implicit...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast_implicit}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_cast_implicit")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast_implicit}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_cast_implicit")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing cast_implicit_in_assignment...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast_implicit_in_assignment}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_cast_implicit_in_assignment")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast_implicit_in_assignment}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_cast_implicit_in_assignment")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing cast_explicit...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_cast_explicit}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_cast_explicit")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_cast_explicit}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_cast_explicit")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing operator...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_operator}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_operator")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_operator}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_operator")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing sequence...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_sequence")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_sequence")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing sequence_minimum_value...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_minimum_value}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_sequence_minimum_value")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_minimum_value}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_sequence_minimum_value")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing sequence_start_value...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_start_value}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_sequence_start_value")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_start_value}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_sequence_start_value")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing sequence_maximum_value...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_maximum_value}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_sequence_maximum_value")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_maximum_value}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_sequence_maximum_value")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing sequence_cycle...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_cycle}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_sequence_cycle")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_cycle}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_sequence_cycle")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing sequence_increment...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_increment}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_sequence_increment")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_increment}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_sequence_increment")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing sequence_type...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_type}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_sequence_type")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_type}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_sequence_type")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing sequence_cache...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_cache}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_sequence_cache")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_cache}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_sequence_cache")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing sequence_table...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_sequence_table}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_sequence_table")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_sequence_table}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_sequence_table")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing relation_column...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_relation_column")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_relation_column")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing relation_column_type...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_type}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_relation_column_type")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_type}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_relation_column_type")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing relation_column_position...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_position}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_relation_column_position")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_position}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_relation_column_position")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing relation_column_default...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_default}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_relation_column_default")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_default}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_relation_column_default")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing relation_column_identity...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_identity}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_relation_column_identity")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_identity}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_relation_column_identity")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing relation_column_generated...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_generated}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_relation_column_generated")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_generated}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_relation_column_generated")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing relation_column_nullable...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_column_nullable}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_relation_column_nullable")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_column_nullable}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_relation_column_nullable")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing relation...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_relation")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_relation")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing table...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_table")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_table")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing table_permanent...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table_permanent}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_table_permanent")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table_permanent}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_table_permanent")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing table_temporary...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table_temporary}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_table_temporary")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table_temporary}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_table_temporary")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing table_partitioned...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table_partitioned}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_table_partitioned")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table_partitioned}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_table_partitioned")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing table_rowsecurity...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table_rowsecurity}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_table_rowsecurity")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table_rowsecurity}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_table_rowsecurity")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing table_forcerowsecurity...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_table_forcerowsecurity}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_table_forcerowsecurity")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_table_forcerowsecurity}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_table_forcerowsecurity")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing view...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_view}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_view")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_view}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_view")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_function...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_function}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_function")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_function}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_function")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_procedure...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_procedure}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_procedure")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_procedure}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_procedure")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_return_type...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_return_type}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_return_type")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_return_type}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_return_type")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_owner...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_owner}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_owner")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_owner}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_owner")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_argument_name...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_argument_name}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_argument_name")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_argument_name}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_argument_name")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_argument_type...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_argument_type}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_argument_type")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_argument_type}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_argument_type")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_argument_mode...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_argument_mode}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_argument_mode")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_argument_mode}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_argument_mode")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_argument_default...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_argument_default}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_argument_default")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_argument_default}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_argument_default")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_immutable...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_immutable}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_immutable")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_immutable}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_immutable")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_volatile...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_volatile}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_volatile")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_volatile}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_volatile")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_stable...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_stable}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_stable")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_stable}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_stable")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_aggregate...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_aggregate}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_aggregate")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_aggregate}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_aggregate")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_window...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_window}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_window")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_window}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_window")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_parallel_safe...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_parallel_safe}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_parallel_safe")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_parallel_safe}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_parallel_safe")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_parallel_unsafe...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_parallel_unsafe}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_parallel_unsafe")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_parallel_unsafe}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_parallel_unsafe")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_parallel_restricted...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_parallel_restricted}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_parallel_restricted")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_parallel_restricted}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_parallel_restricted")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_security_definer...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_security_definer}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_security_definer")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_security_definer}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_security_definer")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_security_invoker...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_security_invoker}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_security_invoker")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_security_invoker}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_security_invoker")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_language...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_language}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_language")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_language}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_language")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_acl...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_acl}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_acl")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_acl}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_acl")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing callable_body...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_callable_body}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_callable_body")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_callable_body}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_callable_body")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing trigger...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_trigger}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_trigger")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_trigger}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_trigger")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing role...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_role")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_role")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing role_superuser...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_superuser}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_role_superuser")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_superuser}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_role_superuser")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing role_inherit...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_inherit}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_role_inherit")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_inherit}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_role_inherit")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing role_create_role...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_create_role}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_role_create_role")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_create_role}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_role_create_role")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing role_create_db...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_create_db}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_role_create_db")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_create_db}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_role_create_db")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing role_can_login...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_can_login}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_role_can_login")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_can_login}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_role_can_login")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing role_replication...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_replication}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_role_replication")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_replication}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_role_replication")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing role_connection_limit...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_connection_limit}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_role_connection_limit")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_connection_limit}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_role_connection_limit")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing role_valid_until...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_valid_until}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_role_valid_until")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_valid_until}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_role_valid_until")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing role_setting...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_role_setting}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_role_setting")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_role_setting}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_role_setting")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing policy...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_policy}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_policy")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_policy}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_policy")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing policy_role...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_policy_role}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_policy_role")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_policy_role}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_policy_role")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_non_null...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_non_null}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_non_null")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_non_null}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_non_null")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_non_null_column...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_non_null_column}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_non_null_column")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_non_null_column}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_non_null_column")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_check...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_check}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_check")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_check}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_check")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_check_expr...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_check_expr}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_check_expr")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_check_expr}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_check_expr")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_unique...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_unique}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_unique")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_unique}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_unique")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_unique_column...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_unique_column}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_unique_column")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_unique_column}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_unique_column")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing extension...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_extension}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_extension")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_extension}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_extension")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing foreign_data_wrapper...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_foreign_data_wrapper}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_foreign_data_wrapper")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_foreign_data_wrapper}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_foreign_data_wrapper")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing foreign_server...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_foreign_server}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_foreign_server")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_foreign_server}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_foreign_server")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing foreign_table...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_foreign_table}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_foreign_table")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_foreign_table}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_foreign_table")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing foreign_column...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_foreign_column}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_foreign_column")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_foreign_column}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_foreign_column")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing relation_foreign_key...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_foreign_key}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_relation_foreign_key")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_foreign_key}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_relation_foreign_key")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing relation_foreign_key_constraint...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_relation_foreign_key_constraint}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_relation_foreign_key_constraint")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_relation_foreign_key_constraint}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_relation_foreign_key_constraint")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_foreign_key...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_foreign_key")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_foreign_key")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_foreign_key_update...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_update}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_foreign_key_update")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_update}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_foreign_key_update")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_foreign_key_delete...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_delete}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_foreign_key_delete")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_delete}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_foreign_key_delete")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_foreign_key_match_full...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_match_full}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_foreign_key_match_full")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_match_full}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_foreign_key_match_full")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_foreign_key_match_partial...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_match_partial}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_foreign_key_match_partial")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_match_partial}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_foreign_key_match_partial")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_foreign_key_match_simple...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_match_simple}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_foreign_key_match_simple")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_match_simple}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_foreign_key_match_simple")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_foreign_key_references...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_references}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_foreign_key_references")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_references}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_foreign_key_references")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing constraint_relation_foreign_key_references_column...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_constraint_relation_foreign_key_references_column}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_constraint_relation_foreign_key_references_column")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_constraint_relation_foreign_key_references_column}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_constraint_relation_foreign_key_references_column")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_basic...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_basic}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_basic")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_basic}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_basic")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_array...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_array}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_array")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_array}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_array")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_composite...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_composite}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_composite")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_composite}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_composite")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_composite_attribute...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_composite_attribute}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_composite_attribute")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_composite_attribute}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_composite_attribute")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_composite_attribute_position...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_composite_attribute_position}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_composite_attribute_position")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_composite_attribute_position}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_composite_attribute_position")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_composite_attribute_collation...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_composite_attribute_collation}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_composite_attribute_collation")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_composite_attribute_collation}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_composite_attribute_collation")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_domain...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_domain}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_domain")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_domain}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_domain")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_enum...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_enum}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_enum")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_enum}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_enum")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_enum_label...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_enum_label}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_enum_label")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_enum_label}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_enum_label")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_pseudo...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_pseudo}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_pseudo")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_pseudo}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_pseudo")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_range...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_range}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_range")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_range}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_range")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing type_multirange...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_type_multirange}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_type_multirange")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_type_multirange}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_type_multirange")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing index...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_index")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_index")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing index_relation...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_relation}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_index_relation")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_relation}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_index_relation")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing index_unique...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_unique}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_index_unique")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_unique}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_index_unique")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing index_unique_null_values_distinct...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_unique_null_values_distinct}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_index_unique_null_values_distinct")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_unique_null_values_distinct}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_index_unique_null_values_distinct")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing index_primary_key...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_primary_key}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_index_primary_key")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_primary_key}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_index_primary_key")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing index_unique_immediate...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_unique_immediate}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_index_unique_immediate")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_unique_immediate}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_index_unique_immediate")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing index_replica_identity...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_replica_identity}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_index_replica_identity")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_replica_identity}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_index_replica_identity")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing index_attribute...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_attribute}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_index_attribute")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_attribute}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_index_attribute")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing index_partial...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_index_partial}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_index_partial")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_index_partial}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_index_partial")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing language...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_language")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_language")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing language_internal...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_internal}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_language_internal")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_internal}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_language_internal")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing language_trusted...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_trusted}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_language_trusted")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_trusted}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_language_trusted")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing language_handler...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_handler}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_language_handler")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_handler}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_language_handler")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing language_inline_handler...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_inline_handler}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_language_inline_handler")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_inline_handler}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_language_inline_handler")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing language_validator...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_validator}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_language_validator")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_validator}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_language_validator")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing language_acl...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_language_acl}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_language_acl")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_language_acl}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_language_acl")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing comment...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_comment}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_comment")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_comment}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_comment")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing dependency...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_dependency}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_dependency")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_dependency}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_dependency")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    perform progress_report('Diffing acl...');
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{added_acl}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "added_acl")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));
                    diffs := jsonb_strip_nulls(jsonb_set(diffs, '{removed_acl}',
                                                         coalesce((with
                                                                       t
                                                                           as materialized (select * from "removed_acl")
                                                                   select
                                                                       jsonb_agg(to_jsonb(t))
                                                                   from
                                                                       t
                                                                       left join dependency d
                                                                                 on d.id = t.id::object_id and
                                                                                    omni_types.variant(d.dependent_on) =
                                                                                    'extension_id'::regtype
                                                                   where
                                                                       (d) is not distinct from null), 'null')));

                    perform set_config('search_path', old_search_path, true);

                    perform omni_vfs.write(transactional_fs,
                                           revision || '/' || current_parent || '.diffs.yaml',
                                           convert_to(omni_yaml.to_yaml(diffs::json), 'utf8'), create_file => true);

                    any_diff := any_diff or (diffs != '{}');
                end loop;
            if not any_diff then
                raise exception 'There are no differences between revisions';
            end if;
        end;
    end if;


    --- 6. Prepare a migration boilerplate?
    --- 7. Transformation functions?
    --- 8. Capture data

    old_search_path := current_setting('search_path');
    perform set_config('search_path', (revision::text)::regnamespace || ',' || old_search_path, true);

    for rec in
        with
            tables as materialized (select
                                        schema_name,
                                        name
                                    from
                                        "table"                            t
                                        natural inner join table_permanent tp),
            data as materialized (select
                                      schema_name,
                                      name,
                                      result as json
                                  from
                                      tables
                                      inner join lateral (select *
                                                          from
                                                              dblink(revision_conninfo,
                                                                     format('select json_agg(to_json(t)) from %I.%I t', schema_name, name)) t(result json)) t
                                                 on true
                                  where
                                      result is not null)
        select *
        from
            data
        where
            schema_name != 'pg_catalog' and
            schema_name != 'information_schema'
        loop
            perform progress_report(format('Capturing data for %I.%I', rec.schema_name, rec.name));
            perform omni_vfs.write(transactional_fs,
                                   revision || '/data/' || rec.schema_name || '.' || rec.name || '.yaml',
                                   convert_to(omni_yaml.to_yaml(rec.json), 'utf8'), create_file => true);
        end loop;


    perform set_config('search_path', old_search_path, true);

    --- X. Done

    ---- Prepare to commit transactional_fs back to the fs
    select
        jsonb_object_agg(name, encode(coalesce(omni_vfs.read(transactional_fs, name), ''), 'base64'))
    into captured_revisions
    from
        omni_vfs.list_recursively(transactional_fs, '')
    where kind = 'file';

    ---- Clean up
    ---- FIXME: can't really do this, even though the foreign servers are gone,
    ----        but the connections won't be released until the very end of the
    ----        transaction. For now, this has to be cleaned up manually.
    ---- perform dblink(self_conn, format('drop database %I', revision_database));
    perform progress_report('Cleaning up');
    update pg_database set datistemplate = false where datname = revision_database;
    raise notice '%', format('Target database %I remained', revision_database)
        using
            detail =
                    'Due to limitations of postgres_fdw transactional scope, it cannot be dropped from within this function',
            hint = format('Remove it with `drop database %I`', revision_database);
    declare
        current_parent revision_id;
    begin
        foreach current_parent in array revision_parents
            loop
                raise notice '%', format('Target database %I remained', current_parent)
                    using
                        detail =
                                'Due to limitations of postgres_fdw transactional scope, it cannot be dropped from within this function',
                        hint = format('Remove it with `drop database %I`', current_parent);

            end loop;
    end;

    if rollback then
        raise exception 'capture_schema_revision_done';
    end if;

    return revision;

exception
    when others then
        if sqlerrm = 'capture_schema_revision_done' then
            ---- Commit transactional_fs back to the fs
            perform omni_vfs.write(fs, revisions_path || '/' || key,
                                   decode(value #>> '{}', 'base64'),
                                   create_file => true)
            from
                jsonb_each(captured_revisions) t;
            ---- Return the revision ID if this is just the [default] rollback behavior,
            return revision;
        else
            ---- re-raise otherwise
            raise;
        end if;
end;
$$;

    execute format('alter function capture_schema_revision set search_path = %s', search_path);

    create function schema_revisions(fs anyelement, revisions_path text, leafs_only boolean default false)
    returns table
            (
                revision revision_id,
                parents  revision_id[],
                metadata jsonb
            )
    language plpgsql
as
$$
begin
    if leafs_only then
        return query
            with
                revisions as materialized (select * from schema_revisions(fs, revisions_path, leafs_only => false)),
                parent as materialized (select distinct unnest(revisions.parents) as parent_id from revisions)
            select
                r0.*
            from
                revisions r0
                left join parent on parent.parent_id = r0.revision
            where
                parent is not distinct from null;
    else
        return query with
                         recursive
                         revisions as (select
                                           omni_vfs.dirname(name)::uuid::revision_id  as revision,
                                           omni_yaml.to_json(convert_from(
                                                   omni_vfs.read(fs, revisions_path || '/' || name),
                                                   'utf-8'))::jsonb                   as metadata,
                                           array((select
                                                      jsonb_array_elements_text(coalesce(
                                                              (omni_yaml.to_json(convert_from(
                                                                      omni_vfs.read(fs, revisions_path || '/' || name),
                                                                      'utf-8'))::jsonb) ->
                                                              'parents',
                                                              '[]'))))::revision_id[] as parents
                                       from
                                           omni_vfs.list_recursively(fs, revisions_path)
                                       where
                                           omni_vfs.basename(name) = 'metadata.yaml'),
                         all_ancestors as (select
                                               r.revision,
                                               unnest(r.parents) as ancestor
                                           from
                                               revisions r
                                           union all
                                           select
                                               aa.revision,
                                               unnest(r.parents) as ancestor
                                           from
                                               all_ancestors  aa
                                               join revisions r
                                                    on aa.ancestor = r.revision),
                         descendant_counts as (select
                                                   ancestor                    as parent,
                                                   count(distinct aa.revision) as descendant_count
                                               from
                                                   all_ancestors aa
                                               group by ancestor)
                     select
                         r.revision as revision,
                         r.parents  as parents,
                         r.metadata as metadata
                     from
                         revisions                   r
                         left join descendant_counts dc
                                   on dc.parent = r.revision
                     order by coalesce(dc.descendant_count, 0) desc;
    end if;
end;
$$;

    execute format('alter function schema_revisions set search_path = %s', search_path);

    create function migrate_to_schema_revision(fs anyelement, revisions_path text,
                                           target revision_id, source_conn text,
                                           rollback boolean default true) returns jsonb
    language plpgsql
as
$$
declare
    host              text := ' host=' ||
                              current_setting('unix_socket_directories') || ' port=' ||
                              current_setting('port') ||
                              ' user=' || current_user;
    source_db         name;
    revision_metadata jsonb;
    parents           revision_id[];
    diff_schema       name;
    result            jsonb;
    conn              text := gen_random_uuid()::text;
begin
    select db into source_db from dblink(source_conn, 'select current_database()') t(db name);

    perform from dblink(source_conn, $q$select relname from pg_class c
        inner join pg_namespace ns on ns.oid = c.relnamespace and ns.nspname = 'omni_schema' where relname = 'deployed_revision'$q$) t(relname name);

    if found then
        perform from dblink(source_conn, format($q$select true from omni_schema.deployed_revision where revision = %L$q$, target)) t(found bool);
        if found then
            -- Don't try to provision an existing revision
            return null;
        end if;
    end if;

    -- TODO:
    -- For now, we assume the case of direct Source->Target migration but we should be able to find
    -- the path and apply this function for all steps.

    if omni_vfs.file_info(fs, revisions_path || '/' || target || '/metadata.yaml') is distinct from null then
        revision_metadata :=
                omni_yaml.to_json(convert_from(omni_vfs.read(fs, revisions_path || '/' || target || '/metadata.yaml'),
                                               'utf8'))::jsonb;
        parents := array((select jsonb_array_elements_text(revision_metadata -> 'parents')))::revision_id[];
    else
        raise exception 'Metadata file % is not found', revisions_path || '/' || target || '/metadata.yaml';
    end if;

    --- 1. Check for pre-conditions

    --- 1.0. Make sure source database has omni_schema and omni_schema.deployed_revision


    perform progress_report(format('Preparing the database'));
    ---- Install omni_schema in the target schema to get its meta
    perform dblink(source_conn, format('create extension if not exists omni_schema version %L cascade',
                                       (select extversion from pg_extension where extname = 'omni_schema')));

    perform dblink(source_conn,
                   format('create table if not exists omni_schema.deployed_revision (revision omni_schema.revision_id)'));

    --- 1.1. The source database must be a parent of the target revision
    declare
        num_parents int;
    begin
        -- However, there's a case when this is the first revision and then everything is good
        if cardinality(parents) > 0 then
            select
                count
            into num_parents
            from
                dblink(source_conn,
                       format('select distinct count(*) from omni_schema.deployed_revision where revision = any (%L)',
                              parents)) t(count int);
            if num_parents != cardinality(parents) then
                raise exception '% database is not a parent to %', source_db, target;
            end if;
        end if;
    end;

    --- 2. Freeze source's schema meta

    ---- Create a foreign server to connect to the revision
    execute format(
            $sql$create server %1$I foreign data wrapper postgres_fdw options(host %2$L, dbname %1$L, port %3$L)$sql$,
            source_db, current_setting('unix_socket_directories'), current_setting('port'));
    execute format('create user mapping for %1$I server %2$I options (user %1$L)',
                   current_user, source_db);

    ---- Define remote meta
    execute format('create schema %I', 'current_revision');
    perform create_remote_meta('current_revision'::regnamespace, 'omni_schema'::name, source_db::text::name,
                               materialize => true);

    --- 3. Run necessary migrations
    perform dblink_connect(conn, source_conn);
    perform dblink(conn, 'begin');
    if cardinality(parents) = 0 then
        --- 3.0. Just assemble it if there are no parents (first migration)
        perform progress_report(format('Assembling'));
        perform
        from
            assemble_schema_revision(fs, revisions_path, target, source_conn)
        where
            execution_error is not null;
        if found then
            raise exception 'Starting migration cannot be assembled due to errors'
                using hint = 'Run assemble_schema to see errors';
        end if;
    else
        --- 3.1. Run migrate.sql file
        declare
            migration_path text;
            migration      text;
            rec            record;
        begin
            migration_path := revisions_path || '/' || target || '/migrate.sql';
            if omni_vfs.file_info(fs, migration_path) is distinct from null then
                perform progress_report(format('Applying migrations'));
                migration := convert_from(omni_vfs.read(fs, migration_path), 'utf8');
                for rec in select * from omni_sql.raw_statements(migration::cstring, true)
                    loop
                        case
                            when omni_sql.statement_type(rec.source::omni_sql.statement) = 'TransactionStmt'
                                then raise exception 'No transactional statements allowed' using detail =
                                        format('Line %s, Column %s: %s', rec.line, rec.col, rec.source);
                            when omni_sql.statement_type(rec.source::omni_sql.statement) = 'MultiStmt' and
                                 omni_sql.statement_type((select
                                                              array_agg(source order by ordinality)
                                                          from
                                                              omni_sql.raw_statements(rec.source::cstring) with ordinality
                                                          limit 1)[1]::omni_sql.statement) =
                                 'TransactionStmt'
                                then raise exception 'No transactional statements allowed' using detail =
                                        format('Line %s, Column %s: %s', rec.line, rec.col, rec.source);
                            else null;
                            end case;
                        begin
                            perform dblink(conn, rec.source);
                        exception
                            when others then
                                perform dblink(conn, 'rollback');
                                perform dblink_disconnect(conn);
                                raise using hint = rec.source;
                        end;
                    end loop;
            end if;
        end;
    end if;

    ---- Stamp new revision
    perform dblink_exec(conn,
                        format('insert into omni_schema.deployed_revision (revision) values (%L)',
                               target));

    --- 4. Compare diffs between the source database and the target revision, should be the same

    ---- Define remote meta
    execute format('create schema %I', target);
    perform create_remote_meta(target::text::regnamespace, 'omni_schema'::name, conn,
                               materialize => true);

    diff_schema := target::text || '_diff';
    execute format('create schema %I', diff_schema);
    perform create_meta_diff('current_revision'::regnamespace, target::text::regnamespace,
                             diff_schema::regnamespace);

    declare
        current_parent  revision_id;
        diff_path       text;
        diff            jsonb;
        expected_diff   jsonb := '{}';
        old_search_path text;
    begin
        foreach current_parent in array parents
            loop
                diff_path := revisions_path || '/' || target || '/' || current_parent || '.diffs.yaml';
                if omni_vfs.file_info(fs, diff_path) is distinct from null then
                    diff := omni_yaml.to_json(convert_from(omni_vfs.read(fs, diff_path), 'utf8'))::jsonb;


                    old_search_path := current_setting('search_path');
                    perform set_config('search_path',
                                       diff_schema::regnamespace::text || ',' || old_search_path, true);


                    perform progress_report(format('Diffing schema'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_schema}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_schema")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_schema}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_schema")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing cast'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_cast}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_cast")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_cast}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_cast")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing cast_binary_coercible'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_cast_binary_coercible}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_cast_binary_coercible")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_cast_binary_coercible}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_cast_binary_coercible")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing cast_function'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_cast_function}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_cast_function")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_cast_function}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_cast_function")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing cast_implicit'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_cast_implicit}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_cast_implicit")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_cast_implicit}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_cast_implicit")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing cast_implicit_in_assignment'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_cast_implicit_in_assignment}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_cast_implicit_in_assignment")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_cast_implicit_in_assignment}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_cast_implicit_in_assignment")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing cast_explicit'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_cast_explicit}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_cast_explicit")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_cast_explicit}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_cast_explicit")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing operator'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_operator}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_operator")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_operator}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_operator")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing sequence'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_sequence}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_sequence")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_sequence}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_sequence")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing sequence_minimum_value'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_sequence_minimum_value}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_sequence_minimum_value")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_sequence_minimum_value}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_sequence_minimum_value")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing sequence_start_value'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_sequence_start_value}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_sequence_start_value")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_sequence_start_value}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_sequence_start_value")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing sequence_maximum_value'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_sequence_maximum_value}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_sequence_maximum_value")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_sequence_maximum_value}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_sequence_maximum_value")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing sequence_cycle'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_sequence_cycle}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_sequence_cycle")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_sequence_cycle}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_sequence_cycle")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing sequence_increment'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_sequence_increment}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_sequence_increment")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_sequence_increment}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_sequence_increment")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing sequence_type'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_sequence_type}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_sequence_type")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_sequence_type}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_sequence_type")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing sequence_cache'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_sequence_cache}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_sequence_cache")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_sequence_cache}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_sequence_cache")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing sequence_table'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_sequence_table}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_sequence_table")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_sequence_table}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_sequence_table")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing relation_column'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_relation_column}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_relation_column")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_relation_column}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_relation_column")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing relation_column_type'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_relation_column_type}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_relation_column_type")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_relation_column_type}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_relation_column_type")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing relation_column_position'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_relation_column_position}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_relation_column_position")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_relation_column_position}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_relation_column_position")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing relation_column_default'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_relation_column_default}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_relation_column_default")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_relation_column_default}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_relation_column_default")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing relation_column_identity'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_relation_column_identity}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_relation_column_identity")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_relation_column_identity}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_relation_column_identity")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing relation_column_generated'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_relation_column_generated}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_relation_column_generated")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_relation_column_generated}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_relation_column_generated")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing relation_column_nullable'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_relation_column_nullable}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_relation_column_nullable")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_relation_column_nullable}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_relation_column_nullable")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing relation'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_relation}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_relation")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_relation}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_relation")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing table'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_table}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_table")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_table}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_table")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing table_permanent'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_table_permanent}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_table_permanent")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_table_permanent}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_table_permanent")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing table_temporary'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_table_temporary}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_table_temporary")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_table_temporary}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_table_temporary")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing table_partitioned'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_table_partitioned}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_table_partitioned")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_table_partitioned}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_table_partitioned")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing table_rowsecurity'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_table_rowsecurity}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_table_rowsecurity")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_table_rowsecurity}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_table_rowsecurity")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing table_forcerowsecurity'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_table_forcerowsecurity}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_table_forcerowsecurity")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_table_forcerowsecurity}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_table_forcerowsecurity")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing view'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_view}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_view")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_view}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_view")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_function'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_function}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_function")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_function}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_function")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_procedure'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_procedure}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_procedure")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_procedure}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_procedure")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_return_type'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_return_type}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_return_type")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_return_type}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_return_type")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_owner'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_owner}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_owner")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_owner}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_owner")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_argument_name'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_argument_name}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_argument_name")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_argument_name}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_argument_name")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_argument_type'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_argument_type}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_argument_type")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_argument_type}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_argument_type")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_argument_mode'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_argument_mode}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_argument_mode")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_argument_mode}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_argument_mode")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_argument_default'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_argument_default}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_argument_default")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_argument_default}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_argument_default")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_immutable'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_immutable}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_immutable")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_immutable}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_immutable")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_volatile'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_volatile}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_volatile")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_volatile}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_volatile")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_stable'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_stable}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_stable")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_stable}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_stable")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_aggregate'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_aggregate}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_aggregate")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_aggregate}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_aggregate")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_window'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_window}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_window")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_window}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_window")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_parallel_safe'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_parallel_safe}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_parallel_safe")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_parallel_safe}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_parallel_safe")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_parallel_unsafe'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_parallel_unsafe}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_parallel_unsafe")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_parallel_unsafe}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_parallel_unsafe")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_parallel_restricted'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_parallel_restricted}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_parallel_restricted")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_parallel_restricted}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_parallel_restricted")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_security_definer'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_security_definer}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_security_definer")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_security_definer}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_security_definer")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_security_invoker'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_security_invoker}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_security_invoker")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_security_invoker}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_security_invoker")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_language'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_language}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_language")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_language}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_language")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_acl'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_acl}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_acl")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_acl}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_acl")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing callable_body'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_callable_body}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_callable_body")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_callable_body}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_callable_body")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing trigger'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_trigger}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_trigger")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_trigger}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_trigger")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing role'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_role}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_role")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_role}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_role")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing role_superuser'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_role_superuser}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_role_superuser")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_role_superuser}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_role_superuser")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing role_inherit'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_role_inherit}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_role_inherit")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_role_inherit}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_role_inherit")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing role_create_role'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_role_create_role}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_role_create_role")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_role_create_role}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_role_create_role")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing role_create_db'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_role_create_db}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_role_create_db")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_role_create_db}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_role_create_db")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing role_can_login'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_role_can_login}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_role_can_login")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_role_can_login}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_role_can_login")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing role_replication'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_role_replication}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_role_replication")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_role_replication}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_role_replication")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing role_connection_limit'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_role_connection_limit}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_role_connection_limit")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_role_connection_limit}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_role_connection_limit")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing role_valid_until'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_role_valid_until}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_role_valid_until")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_role_valid_until}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_role_valid_until")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing role_setting'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_role_setting}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_role_setting")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_role_setting}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_role_setting")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing policy'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_policy}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_policy")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_policy}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_policy")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing policy_role'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_policy_role}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_policy_role")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_policy_role}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_policy_role")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_non_null'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_non_null}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_non_null")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_non_null}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_non_null")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_non_null_column'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_non_null_column}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_non_null_column")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_non_null_column}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_non_null_column")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_check'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_check}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_check")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_check}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_check")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_check_expr'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_check_expr}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_check_expr")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_check_expr}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_check_expr")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_unique'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_unique}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_unique")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_unique}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_unique")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_unique_column'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_unique_column}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_unique_column")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_unique_column}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_unique_column")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing extension'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_extension}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_extension")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_extension}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_extension")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing foreign_data_wrapper'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_foreign_data_wrapper}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_foreign_data_wrapper")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_foreign_data_wrapper}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_foreign_data_wrapper")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing foreign_server'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_foreign_server}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_foreign_server")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_foreign_server}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_foreign_server")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing foreign_table'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_foreign_table}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_foreign_table")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_foreign_table}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_foreign_table")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing foreign_column'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_foreign_column}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_foreign_column")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_foreign_column}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_foreign_column")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing relation_foreign_key'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_relation_foreign_key}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_relation_foreign_key")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_relation_foreign_key}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_relation_foreign_key")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing relation_foreign_key_constraint'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_relation_foreign_key_constraint}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_relation_foreign_key_constraint")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_relation_foreign_key_constraint}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_relation_foreign_key_constraint")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_foreign_key'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_foreign_key}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_foreign_key")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_foreign_key}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_foreign_key")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_foreign_key_update'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_foreign_key_update}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_foreign_key_update")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_foreign_key_update}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_foreign_key_update")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_foreign_key_delete'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_foreign_key_delete}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_foreign_key_delete")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_foreign_key_delete}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_foreign_key_delete")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_foreign_key_match_full'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_foreign_key_match_full}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_foreign_key_match_full")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_foreign_key_match_full}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_foreign_key_match_full")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_foreign_key_match_partial'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_foreign_key_match_partial}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_foreign_key_match_partial")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_foreign_key_match_partial}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_foreign_key_match_partial")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_foreign_key_match_simple'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_foreign_key_match_simple}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_foreign_key_match_simple")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_foreign_key_match_simple}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_foreign_key_match_simple")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_foreign_key_references'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_foreign_key_references}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_foreign_key_references")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_foreign_key_references}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_foreign_key_references")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing constraint_relation_foreign_key_references_column'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_constraint_relation_foreign_key_references_column}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_constraint_relation_foreign_key_references_column")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_constraint_relation_foreign_key_references_column}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_constraint_relation_foreign_key_references_column")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_basic'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_basic}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_basic")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_basic}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_basic")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_array'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_array}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_array")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_array}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_array")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_composite'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_composite}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_composite")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_composite}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_composite")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_composite_attribute'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_composite_attribute}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_composite_attribute")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_composite_attribute}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_composite_attribute")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_composite_attribute_position'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_composite_attribute_position}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_composite_attribute_position")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_composite_attribute_position}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_composite_attribute_position")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_composite_attribute_collation'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_composite_attribute_collation}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_composite_attribute_collation")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_composite_attribute_collation}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_composite_attribute_collation")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_domain'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_domain}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_domain")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_domain}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_domain")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_enum'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_enum}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_enum")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_enum}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_enum")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_enum_label'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_enum_label}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_enum_label")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_enum_label}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_enum_label")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_pseudo'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_pseudo}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_pseudo")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_pseudo}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_pseudo")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_range'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_range}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_range")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_range}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_range")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing type_multirange'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_type_multirange}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_type_multirange")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_type_multirange}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_type_multirange")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing index'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_index}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_index")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_index}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_index")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing index_relation'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_index_relation}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_index_relation")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_index_relation}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_index_relation")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing index_unique'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_index_unique}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_index_unique")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_index_unique}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_index_unique")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing index_unique_null_values_distinct'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_index_unique_null_values_distinct}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_index_unique_null_values_distinct")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_index_unique_null_values_distinct}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_index_unique_null_values_distinct")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing index_primary_key'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_index_primary_key}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_index_primary_key")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_index_primary_key}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_index_primary_key")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing index_unique_immediate'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_index_unique_immediate}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_index_unique_immediate")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_index_unique_immediate}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_index_unique_immediate")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing index_replica_identity'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_index_replica_identity}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_index_replica_identity")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_index_replica_identity}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_index_replica_identity")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing index_attribute'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_index_attribute}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_index_attribute")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_index_attribute}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_index_attribute")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing index_partial'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_index_partial}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_index_partial")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_index_partial}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_index_partial")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing language'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_language}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_language")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_language}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_language")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing language_internal'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_language_internal}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_language_internal")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_language_internal}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_language_internal")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing language_trusted'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_language_trusted}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_language_trusted")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_language_trusted}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_language_trusted")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing language_handler'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_language_handler}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_language_handler")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_language_handler}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_language_handler")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing language_inline_handler'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_language_inline_handler}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_language_inline_handler")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_language_inline_handler}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_language_inline_handler")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing language_validator'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_language_validator}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_language_validator")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_language_validator}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_language_validator")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing language_acl'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_language_acl}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_language_acl")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_language_acl}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_language_acl")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing comment'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_comment}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_comment")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_comment}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_comment")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing dependency'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_dependency}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_dependency")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_dependency}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_dependency")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                    perform progress_report(format('Diffing acl'));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{added_acl}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "added_acl")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));
                        expected_diff :=
                                jsonb_strip_nulls(jsonb_set(expected_diff, '{removed_acl}',
                                                            coalesce((with
                                                                          t
                                                                              as materialized (select * from "removed_acl")
                                                                      select
                                                                          jsonb_agg(to_jsonb(t))
                                                                      from
                                                                          t
                                                                          left join dependency d
                                                                                    on d.id = t.id::object_id and
                                                                                       omni_types.variant(d.dependent_on) =
                                                                                       'extension_id'::regtype
                                                                      where
                                                                          (d) is not distinct from null), 'null')));

                    perform set_config('search_path', old_search_path, true);

                    ---- Compare diffs. If they are not equal, we need to produce sets of differences for each
                    ---- provided type of diff.
                    if expected_diff != diff then
                        with
                            key_vals as (select
                                             coalesce(a.key, b.key) as key,
                                             a.value                as a_value,
                                             b.value                as b_value
                                         from
                                             jsonb_each(diff)                          a
                                             full outer join jsonb_each(expected_diff) b
                                                             on a.key = b.key)
                        select
                            jsonb_object_agg(key, difference)
                        into result
                        from
                            (select
                                 key,
                                 case
                                     -- If key exists only in j2, return its full array
                                     when a_value is null then b_value
                                     -- If key exists only in j1, return its full array
                                     when b_value is null then a_value
                                     else (
                                         -- For keys existing in both, compute the symmetric difference:
                                         select
                                             jsonb_agg(elem)
                                         from
                                             (
                                                 -- Elements in j1 not in j2
                                                 select
                                                     elem
                                                 from
                                                     jsonb_array_elements(a_value) elem
                                                 except
                                                 select
                                                     elem
                                                 from
                                                     jsonb_array_elements(b_value) elem
                                                 union
                                                 -- Elements in j2 not in j1
                                                 select
                                                     elem
                                                 from
                                                     jsonb_array_elements(b_value) elem
                                                 except
                                                 select
                                                     elem
                                                 from
                                                     jsonb_array_elements(a_value) elem) sub)
                                     end as difference
                             from
                                 key_vals
                             -- Only include keys that differ in some way:
                             where
                                 (a_value is not null and b_value is not null and
                                  a_value is distinct from b_value) or
                                 a_value is null or
                                 b_value is null) t;

                    end if;

                end if;
            end loop;
    end;

    if result is null then
        perform progress_report(format('Migrations applied successfully'));
        perform dblink(conn, 'commit');
        perform dblink_disconnect(conn);
    else
        perform progress_report(format('Schema is not matching expectations, rolled back'));
        raise notice 'Schema is not matching expectation, rolling back';
        perform dblink(conn, 'rollback');
        perform dblink_disconnect(conn);
    end if;

    --- 5. Compare the data with the target revision, revision's data should be present

    if rollback then
        raise exception 'migrate_to_schema_revision_done';
    end if;

    return result;

exception
    when others then
        if sqlerrm = 'migrate_to_schema_revision_done' then
            return result;
        else
            ---- re-raise otherwise
            raise;
        end if;

end;
$$;

    execute format('alter function migrate_to_schema_revision set search_path = %s', search_path);

end;
$instantiate_evolution$;


select instantiate_evolution(schema => 'omni_schema');;
